# Import standard libraries.
from numpy import nonzero


def calculate_ton_kilometer_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate ton kilometer costs.

    The ton kilometers offered (TKO) are the product of the payload and the range. The ton kilometer costs are the DOC
    related to the TKO and allow the analysis of a change in payload capacity and thus the assessment of the payload
    kilometer potential.

    This function adds the 'ton_kilometer_costs' dict to the 'dict_operating_costs' that contains the following
    parameters:
        - array ton_kilometer_costs: Ton kilometer costs (range dependent) in EUR/TKO
        - array ton_kilometer_offered: Ton kilometer offered (range dependent) in EUR/(kg*km)

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module configuration file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary values from 'dict_operating_costs'.
    direct_operating_costs_annual = dict_operating_costs['direct_operating_costs']['direct_operating_costs_annual']
    payloads_for_prd = dict_operating_costs['payload_range_diagram_data']['payloads_for_prd']
    ranges_for_prd = dict_operating_costs['payload_range_diagram_data']['ranges_for_prd']/1000

    # Calculate ton kilometer offered.
    ton_kilometer_offered = payloads_for_prd * ranges_for_prd
    # Replace zeros to eliminate 'division by zero' warning.
    zero_tko = nonzero(ton_kilometer_offered == 0)[0]
    ton_kilometer_offered[zero_tko] = 1e-8
    # Calculate ton kilometer costs.
    ton_kilometer_costs = direct_operating_costs_annual / ton_kilometer_offered

    # Prepare output data.
    dict_operating_costs['ton_kilometer_costs'] = {}
    dict_operating_costs['ton_kilometer_costs']['ton_kilometer_costs'] = ton_kilometer_costs
    dict_operating_costs['ton_kilometer_costs']['ton_kilometer_offered'] = ton_kilometer_offered

    return dict_operating_costs
