/**
 * \file calculateEmissions_schaefer.h
 * \author katrin.bistreck@tuhh.de
 * \brief Strategy to calculate missions according to Schäfer(2011)
 * \version 0.1
 * \date 2023-11-25
 *
 * @copyright Copyright (c) 2023
 *
 */
#include <iostream>
#include <memory>
#include <string>
#include <unitConversion/unitConversion.h>
#include "calculateEmissions_schaefer.h"
#include "schaeferIOData.h"
#include <moduleBasics/plot.h>


void LifeCycleEmissionsSchaefer::run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML) {
    if (rtIO->aircraft_energy_carrier_type() != "kerosene" || rtIO->aircraft_configuration_type() != "tube_and_wing") {
            myRuntimeInfo->warn << "Schaefer method can only be used for kerosene powered TAW aircraft. No LCA calculation will be performed!" << std::endl;
    } else {
        myRuntimeInfo->out << "  Calculation of life cycle emissions" << std::endl;
        schaeferIOData_ = std::make_shared<SchaeferIOData>(rtIO, ecoAssXML);
        /* Production phase */
        myRuntimeInfo->out << "     Calculation of production phase..." << std::endl;
        theProduction.calculateProductionResources(schaeferIOData_, rtIO);
        theProduction.calculateProductionEmissions(schaeferIOData_, rtIO);

        /* Operation phase*/
        myRuntimeInfo->out << "     Calculation of operation phase..." << std::endl;
        theOperation.calculateOperationResources(schaeferIOData_, rtIO, theProduction);
        theOperation.calculateOperationEmissions(schaeferIOData_, rtIO, theProduction);

        /* End-of-Life Phase */
        myRuntimeInfo->out << "     Calculation of end of life phase..." << std::endl;
        theEndOfLife.calculateEoLResources(schaeferIOData_, rtIO);
        theEndOfLife.calculateEoLEmissions(schaeferIOData_, theProduction);

        /* Development Phase (calculation is executed last because of production/EoL of test aircraft)*/
        myRuntimeInfo->out << "     Calculation of development phase..." << std::endl;
        theDevelopment.calculateDevelopmentResources(schaeferIOData_);
        theDevelopment.calculateDevelopmentEmissions(schaeferIOData_, theProduction, theEndOfLife);
    }
}

