/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ECOLOGICAL_ASSESSMENT_SRC_EMISSION_CALCULATION_MISSION_EMISSIONS_FLIGHTPATH_H_
#define ECOLOGICAL_ASSESSMENT_SRC_EMISSION_CALCULATION_MISSION_EMISSIONS_FLIGHTPATH_H_

#include <vector>

/** \class flightPathPoint
 * \brief Class flightPathPoint contains properties to describe the conditions in a certain point of the flight path
 */
class flightPathPoint {
 public:
    double time; /**< Time [s] */
    double range; /**< Range [m] */
    double altitude; /**< Altitude [m] */
    double TAS; /**< True Air Speed [m/s] */
    double Mach; /**< Mach [-] */
    double Tamb; /**< Ambient temperature [K] */
    double Pamb; /**< Ambient pressure [Pa] */
    double fuelConsumed; /**< Consumed fuel [kg] */
    double fuelFlow; /**< Fuel flow [kg/s] */
    double NOxFlow; /**< NOx flow [kg/s]*/
    double N1; /**< Engine speed (N1) compared to nominal speed [-] */
    double St3MassFlow; /**< Mass flow at last compressor exit, cold side heat exchanger inlet [kg/s]*/
    double St3TotalTemperature; /**< Total temperature at last compressor exit, cold side heat exchanger inlet [K]*/
    double St3StaticTemperature;/**< Static temperature at last compressor exit, cold side heat exchanger inlet [K]*/
    double St3TotalPressure; /**< Total pressure at last compressor exit, cold side heat exchanger inlet [kPa]*/
    double St3Mach; /**< Mach at last compressor exit, cold side heat exchanger inlet [K]*/
    double St4FuelAir; /**< Fuel/Air at burner exit [-]*/
    double BPR; /**< By pass ratio [-] */

    /** \brief Constructor of class flightPathPoint
      * \param aTime: Time [s]
      * \param aRange: Range [m]
      * \param anAltitude: Altitude [m]
      * \param aTAS: True air speed [m/s]
      * \param aMach: Mach [-]
      * \param aTamb: Ambient temperature [K]
      * \param aPamb: Ambient pressure [Pa]
      * \param aFuelConsumed: Consumed fuel [kg]
      * \param aFuelFlow: Fuel flow [kg/s]
      * \param aNOxFlow: NOx flow [kg/s]
      * \param aN1: Engine speed (N1) compared to nominal speed [-]
      * \param aSt3MassFlow: Mass flow at last compressor exit, cold side heat exchanger inlet [kg/s]
      * \param aSt3TotalTemperature: Static temperature at last compressor exit, cold side heat exchanger inlet [K]
      * \param aSt3StaticTemperature: Mach at last compressor exit, cold side heat exchanger inlet [K]
      * \param aSt3TotalPressure: Total pressure at last compressor exit, cold side heat exchanger inlet [kPa]
      * \param aSt3Mach: Mach at last compressor exit, cold side heat exchanger inlet [-]
      * \param aSt4FuelAir: Fuel/Air at burner exit [-]
      * \param aBPR: By pass ratio [-]
     */
    flightPathPoint(double aTime,
                    double aRange,
                    double anAltitude,
                    double aTAS,
                    double aMach,
                    double aTamb,
                    double aPamb,
                    double aFuelConsumed,
                    double aFuelFlow,
                    double aNOxFlow,
                    double aN1,
                    double aSt3MassFlow,
                    double aSt3TotalTemperature,
                    double aSt3StaticTemperature,
                    double aSt3TotalPressure,
                    double aSt3Mach,
                    double aSt4FuelAir,
                    double aBPR);
    /** \brief Destructor of class flightPathPoint
     */
    virtual ~flightPathPoint();
};

/** \class flightPath
 * \brief Class flightPath contains the description of all points on the flight path
 */
class flightPath {
 public:
    std::vector<flightPathPoint> pathPoints; /**< Vector containing path objects of pathpoints */
    /** \brief Constructor of class flightPath
     */
    flightPath();
    /** \brief Destructor of class flightPath
     */
    virtual ~flightPath();
};


#endif // ECOLOGICAL_ASSESSMENT_SRC_EMISSION_CALCULATION_MISSION_EMISSIONS_FLIGHTPATH_H_
