/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mission.h"
#include <algorithm>
#include <numeric>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <memory>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>
#include <moduleBasics/runtimeIO.h>


// TODO(franz#1#): emission calculation at derate correct? Check! (Problem: Base curves with EI_ICAO (without Derate))
Mission::Mission()
    :
    missionFileName(""),
    deltaISA(0.),
    myEcoDatabase(ecoDatabase()),
    rangeMission(0.),
    cruiseSteps(0),
    m_fuel_Taxi(0.),
    m_fuel_TO(0.),
    m_fuel_ClimbOut(0.),
    m_fuel_App(0.),
    m_fuel_Taxi_ICAO(0.),
    m_fuel_TO_ICAO(0.),
    m_fuel_ClimbOut_ICAO(0.),
    m_fuel_App_ICAO(0.),
    m_fuel_LTO(0.),
    m_fuel_Cruise(0.),
    taxiTime(0.),
    takeOffTime(0.),
    climbOutTime(0.),
    approachTime(0.),
    LTOtime(0.),
    NOx_TO(0.),
    NOx_ClimbOut(0.),
    NOx_App(0.),
    thrustTaxi(0.),
    thrustTakeoff(0.),
    thrustClimbOut(0.),
    thrustApproach(0.),
    T3_ID(0.),
    T3_TO(0.),
    T3_CL(0.),
    T3_AP(0.),
    P3_ID(0.),
    P3_TO(0.),
    P3_CL(0.),
    P3_AP(0.),
    P3t_ID(0.),
    P3t_TO(0.),
    P3t_CL(0.),
    P3t_AP(0.),
    St3Mach_ID(0.),
    St3Mach_TO(0.),
    St3Mach_CL(0.),
    St3Mach_AP(0.),
    St3MassFlow_ID(0.),
    St3MassFlow_TO(0.),
    St3MassFlow_CL(0.),
    St3MassFlow_AP(0.),
    fuelFlow_ID(0.),
    fuelFlow_TO(0.),
    fuelFlow_CL(0.),
    fuelFlow_AP(0.),
    numberOfSteps(0.),
    timeStep(0.),
    flightsPerLC(0.),
    c_soot_LTO_max(0.) {
}

void Mission::run(const std::string& missionType, const std::shared_ptr<RuntimeIO>& rtIO, std::shared_ptr<node> ecoAssXML) {
    missionIOData_ = std::make_shared<MissionIOData>(rtIO);
    myRuntimeInfo->out << "  Calculation of " << missionType << " mission emissions" << std::endl;
    calculate_mission(missionType, rtIO);
    missionIOData_->updateEcologicalAssessmentXml(ecoAssXML, *this);
}

void Mission::calculate_mission(const std::string& missionType, const std::shared_ptr<RuntimeIO>& rtIO) {
    rtIOPt = rtIO;
    typeOfMission = missionType;
    /*Read mission data*/
    get_mission_data();
    myFlightPath = flightPath();
    for (unsigned int i(0); i < missionInfo.size(); i++) {
        myFlightPath.pathPoints.push_back(flightPathPoint(missionInfo.at(i).time,
                                          missionInfo.at(i).range,
                                          missionInfo.at(i).alt,
                                          missionInfo.at(i).TAS_ms,
                                          missionInfo.at(i).Mach,
                                          missionInfo.at(i).Tamb,
                                          missionInfo.at(i).Pamb,
                                          missionInfo.at(i).fuelConsumed,
                                          missionInfo.at(i).fuelFlow,
                                          missionInfo.at(i).NOxFlow,
                                          missionInfo.at(i).N1,
                                          missionInfo.at(i).St3MassFlow,
                                          missionInfo.at(i).St3TotalTemperature,
                                          missionInfo.at(i).St3StaticTemperature,
                                          missionInfo.at(i).St3TotalPressure,
                                          missionInfo.at(i).St3Mach,
                                          missionInfo.at(i).St4FuelAir,
                                          missionInfo.at(i).BPR));
}
    /* Create emissionspath.csv*/
    write_emissions_path_csv();
    /* Determination of mission segments data*/
    get_mission_segment_data();
    /* Values which are used for ICAO database measurements */
    get_engine_thermodynamics_LTO();
    /* Set mission/aircraft parameter */
    numberOfSteps =  this->myFlightPath.pathPoints.size();
    LTOtime = (taxiTime + takeOffTime + climbOutTime + approachTime) / 60.;
    m_fuel_Taxi_ICAO = missionIOData_->numberOfEngines * (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * myEcoDatabase.taxiTime_ICAO);
    m_fuel_TO_ICAO = missionIOData_->numberOfEngines * (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * myEcoDatabase.takeoffTime_ICAO);
    m_fuel_ClimbOut_ICAO = missionIOData_->numberOfEngines * (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::climb) * myEcoDatabase.climboutTime_ICAO);
    m_fuel_App_ICAO = missionIOData_->numberOfEngines * (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::approach) * myEcoDatabase.appTime_ICAO);
    flightsPerLC = missionIOData_->durationOperation * missionIOData_->flightsPerYear;
    /* Calculate emissions */
    if (rtIO->aircraft_energy_carrier_type() == "kerosene") {
        calculate_kerosene_emissions();
    } else if (rtIO->aircraft_energy_carrier_type() == "liquid_hydrogen") {
        calculate_hydrogen_combustion_emissions();
    } else {
        myRuntimeInfo->err << "No calculation method implemented for fuel type " << rtIO->aircraft_energy_carrier_type() << std::endl;
    }
    /*Compare calculated LTO values with ICAO data*/
    LTO_output_comparison();
}

void Mission::LTO_output_comparison() {
    /* LTO Output Comparison ICAO vs. MissionAnalysis*/
    myRuntimeInfo->info << "Comparison ICAO vs MissionAnalysis:" << std::endl;
    myRuntimeInfo->info << " Taxi" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << m_fuel_Taxi << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << m_fuel_Taxi / taxiTime << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << m_fuel_Taxi_ICAO / myEcoDatabase.taxiTime_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << taxiTime / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.taxiTime_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_ID << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_ID << std::endl;
    myRuntimeInfo->info << " T/O" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << m_fuel_TO << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << m_fuel_TO / takeOffTime << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << m_fuel_TO_ICAO / myEcoDatabase.takeoffTime_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << takeOffTime / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.takeoffTime_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_TO << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_TO << std::endl;
    myRuntimeInfo->info << " Climbinfo" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << m_fuel_ClimbOut << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << m_fuel_ClimbOut / climbOutTime << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << m_fuel_ClimbOut_ICAO / myEcoDatabase.climboutTime_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << climbOutTime / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.climboutTime_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_CL << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_CL << std::endl;
    myRuntimeInfo->info << " Approach" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << m_fuel_App << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << m_fuel_App / approachTime << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << m_fuel_App_ICAO / myEcoDatabase.appTime_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << approachTime / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.appTime_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_AP << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_AP << std::endl;
}

void Mission::get_engine_thermodynamics_LTO() {
    missionIOData_->myEngine.calculate_N1_with_penalties(0., 0., myAtm, 1., "takeoff", 0., 0.);
    //for all engines together; max. takeoff Thrust SL and ISA (not ISA+15!)(see Ruijgrok, p. 298 and ICAO 2008 p. III-2-2
    double maxTakeOffThrustSL = missionIOData_->myEngine.get_thrust_aircraft();
    //Taxi
    thrustTaxi = 0.07 * maxTakeOffThrustSL;//for all engines together!
    //Takeoff
    thrustTakeoff = 1. * maxTakeOffThrustSL;//for all engines together!
    //ClimbOut
    thrustClimbOut = 0.85 * maxTakeOffThrustSL;//for all engines together!
    //Approach
    thrustApproach = 0.3 * maxTakeOffThrustSL;//for all engines!

    /* Set engine parameter for ICAO database measurements*/
    // Taxi (Idle)
    enum thrustLimit{DERATED = 2, LOWERBOUNDLIMITED, UPPERBOUNDLIMITED};
    try {
        missionIOData_->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtm, 1., "takeoff", 0., 0., thrustTaxi); // SI unit
    } catch (const int& engineStatus) {
        switch (engineStatus) {
        case DERATED:
            myRuntimeInfo->warn << "required \"thrustTaxi\" violates thrustLimit from engine deck!" << std::endl;
            myRuntimeInfo->warn << "Engine is derated." << std::endl;
            myRuntimeInfo->warn << "thrustTaxi        = " << thrustTaxi << " N" << std::endl;
            myRuntimeInfo->warn << "engineThrustLimit = " << missionIOData_->myEngine.get_thrust_aircraft() << " N" << std::endl;
            myRuntimeInfo->warn << "Engine rating set to idle." << std::endl;
            missionIOData_->myEngine.calculate_N1_with_penalties(0., 0., myAtm, 1., "idle", 0., 0.);
        case LOWERBOUNDLIMITED:
            myRuntimeInfo->warn << "required \"thrustTaxi\" violates thrustLimit from engine deck!" << std::endl;
            myRuntimeInfo->warn << "Reached lower speed limit." << std::endl;
            myRuntimeInfo->warn << "thrustTaxi        = " << thrustTaxi << " N" << std::endl;
            myRuntimeInfo->warn << "engineThrustLimit = " << missionIOData_->myEngine.get_thrust_aircraft() << " N" << std::endl;
            myRuntimeInfo->warn << "Engine rating set to idle." << std::endl;
            missionIOData_->myEngine.calculate_N1_with_penalties(0., 0., myAtm, 1., "idle", 0., 0.);
        case UPPERBOUNDLIMITED:
            myRuntimeInfo->warn << "required \"thrustTaxi\" violates thrustLimit from engine deck!" << std::endl;
            myRuntimeInfo->warn << "Reached upper speed limit." << std::endl;
            myRuntimeInfo->warn << "thrustTaxi        = " << thrustTaxi << " N" << std::endl;
            myRuntimeInfo->warn << "engineThrustLimit = " << missionIOData_->myEngine.get_thrust_aircraft() << " N" << std::endl;
            myRuntimeInfo->warn << "Engine rating set to idle." << std::endl;
            missionIOData_->myEngine.calculate_N1_with_penalties(0., 0., myAtm, 1., "idle", 0., 0.);
        default:
            break;
        }
    } catch (...) {
        missionIOData_->myEngine.calculate_N1_with_penalties(0., 0., myAtm, 1., "idle", 0., 0.);
        myRuntimeInfo->out << "An unknown error has occured during N1 calculation." << std::endl;
    }
    T3_ID = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_ID = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_ID = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_ID = P3t_ID / (1. + 1.4 / 2.*pow(St3Mach_ID, 2));
    St3MassFlow_ID = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_ID = missionIOData_->myEngine.get_aircraft_fuelflow();
    // Take off
    missionIOData_->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtm, 1., "takeoff", 0., 0., thrustTakeoff);
    T3_TO = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_TO = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_TO = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_TO = P3t_TO / (1. + 1.4 / 2.*pow(St3Mach_TO, 2));
    St3MassFlow_TO = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_TO = missionIOData_->myEngine.get_aircraft_fuelflow();
    // Climb
    missionIOData_->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtm, 1., "takeoff", 0., 0., thrustClimbOut);
    T3_CL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_CL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_CL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_CL = P3t_CL / (1. + 1.4 / 2.*pow(St3Mach_CL, 2));
    St3MassFlow_CL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_CL = missionIOData_->myEngine.get_aircraft_fuelflow();
    // Approach
    missionIOData_->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtm, 1., "takeoff", 0., 0., thrustApproach);
    T3_AP = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_AP = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_AP = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_AP = P3t_AP / (1. + 1.4 / 2.*pow(St3Mach_AP, 2));
    St3MassFlow_AP = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_AP = missionIOData_->myEngine.get_aircraft_fuelflow();
}


/* Functions for mission data*/

void Mission::get_mission_data() {
    /* Read from acxml */
    std::string missionFile = EndnodeReadOnly<std::string>("requirements_and_specifications/mission_files/" + typeOfMission + "_mission_file").read(rtIOPt->acxml).value();
    this->deltaISA = EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + typeOfMission
                        + "_mission/delta_ISA").read(rtIOPt->acxml).value();;
    myAtm.setAtmosphere(0., ISA_TEMPERATURE + this->deltaISA, ISA_PRESSURE); //sets atmosphere to deltaISA conditions
    double cargo_fraction(1.);
    double pax_mass_fraction(1.);
    if (typeOfMission == "study") {
        cargo_fraction = EndnodeReadOnly<double>(std::string("requirements_and_specifications/requirements/top_level_aircraft_requirements/")
                            + std::string("study_mission/payload_fractions/cargo_fraction")).read(rtIOPt->acxml).value();
        pax_mass_fraction = EndnodeReadOnly<double>(std::string("requirements_and_specifications/requirements/top_level_aircraft_requirements/")
                            + std::string("study_mission/payload_fractions/passenger_mass_fraction")).read(rtIOPt->acxml).value();
    }
    double cargo_mass = cargo_fraction * EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/cargo_definition/additional_cargo_mass")).read(rtIOPt->acxml).value();

    double pax_mass = pax_mass_fraction
                        * EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/passenger_definition/total_number_passengers")).read(rtIOPt->acxml).value()
                        * (EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/passenger_definition/mass_per_passenger")).read(rtIOPt->acxml).value()
                            + EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/passenger_definition/luggage_mass_per_passenger")).read(rtIOPt->acxml).value());
    double  missionPayload = cargo_mass + pax_mass;
    cruiseSteps = rtIOPt->acxml.getVector("/aircraft_exchange_file/analysis/mission/" + typeOfMission + "_mission/cruise/cruise_steps/cruise_step").size();
    for (int i = 0; i < cruiseSteps; i++) {
        stepLength.push_back(EndnodeReadOnly<double>("analysis/mission/" + typeOfMission +
                        "_mission/cruise/cruise_steps/cruise_step@" + num2Str(i) + "/relative_end_of_cruise_step").read(rtIOPt->acxml).value());
        flightLevel.push_back(convertUnit(METER, FOOT, (EndnodeReadOnly<double>("analysis/mission/" + typeOfMission +
                        "_mission/cruise/cruise_steps/cruise_step@" + num2Str(i) + "/altitude").read(rtIOPt->acxml).value())/100));
    }
    /* Read from mission.xml */
    const node& missionXML(aixml::openDocument(rtIOPt->getMissionDataDir() + "/" + missionFile));
    taxiTime = EndnodeReadOnly<double>("mission/taxi_time_origin").read(missionXML).value() + EndnodeReadOnly<double>("mission/taxi_time_destination").read(missionXML).value();
    rangeMission = convertUnit(METER, NAUTICALMILE, EndnodeReadOnly<double>("mission/range").read(missionXML).value());

    /* Read from mission.csv */
    std::string outputFilename = rtIOPt->acxmlAccess.filename().string();
    outputFilename.resize(outputFilename.rfind('.')); // cut off any extension, if it exists
    std::stringstream outputName;
    outputName << outputFilename << "_" << missionFile.substr(missionFile.rfind(FILESEPERATOR) + 1, missionFile.rfind('.') - missionFile.rfind(FILESEPERATOR) - 1)
               << "_R" << int(ceil(rangeMission))
               << "_PL" << int(ceil(missionPayload)) << "_out.csv";
    this->missionFileName = rtIOPt->getMissionDataDir() + "/" + outputName.str();
    /* Reading the mission file into a std::vector*/
    std::ifstream missionCSV;
    missionCSV.open(missionFileName.c_str(), std::ios::in);
    if (!missionCSV) {
        myRuntimeInfo->err << missionFileName << " can not be opened." << std::endl;
        exit(1);
    } else {
        /* Read file in std::vector */
        std::vector <std::string> lineVector;
        std::string line;
        std::vector <std::string> headerVector;
        std::vector <std::string> lineData;
        int j = 0; // line number
        while (getline(missionCSV, line)) {
            lineVector.push_back(line);
            split(lineVector.back(), ';', &lineData);
            MissionData tmpMissionData;
            if (j == 0) {
                headerVector = lineData;
            } else if (j > 1) { //exclude header and first line
                for (unsigned int i(0); i < lineData.size(); i++) { // for all columns
                    if (headerVector.at(i).find("Time [s]") != std::string::npos) {
                        tmpMissionData.time = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Range [m]") != std::string::npos) {
                        tmpMissionData.range = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Altitude [m]") != std::string::npos) {
                        tmpMissionData.alt = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("TAS [m/s]") != std::string::npos) {
                        tmpMissionData.TAS_ms = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("CAS [m/s]") != std::string::npos) {
                        tmpMissionData.CAS_ms = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Total mass [kg]") != std::string::npos) {
                        tmpMissionData.mass = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Fuel consumed") != std::string::npos) {
                        // \todo save in struct/class for different propulsors
                        tmpMissionData.fuelConsumed = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Thrust [N]") != std::string::npos) {
                        tmpMissionData.thrust = fabs(atof(lineData.at(i).c_str()));
                    } else if (headerVector.at(i).find("Fuelflow [kg/s]") != std::string::npos) {
                        tmpMissionData.fuelFlow = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("TAS [kts]") != std::string::npos) {
                        tmpMissionData.TAS_kts = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Mach [-]") != std::string::npos) {
                        tmpMissionData.Mach = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("FL [100 ft]") != std::string::npos) {
                        tmpMissionData.flightLevel = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("ROC [fpm]") != std::string::npos) {
                        tmpMissionData.ROC = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("C_L [-]") != std::string::npos) {
                        tmpMissionData.C_L = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("SAR [m/kg]") != std::string::npos) {
                        tmpMissionData.SAR = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Aero Config [-]") != std::string::npos) {
                        lineData.at(i).erase(std::remove_if(lineData.at(i).begin(), lineData.at(i).end(), ::isspace), lineData.at(i).end());
                        tmpMissionData.aeroConfig = lineData.at(i);
                    } else if (headerVector.at(i).find("Bleed [kg/s]") != std::string::npos) {
                        tmpMissionData.bleedAirOfftake = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Shaft power offtake [W]") != std::string::npos) {
                        tmpMissionData.shaftPowerOfftake = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Engine Rating [-]") != std::string::npos) {
                        lineData.at(i).erase(std::remove_if(lineData.at(i).begin(), lineData.at(i).end(), ::isspace), lineData.at(i).end());
                        tmpMissionData.engineRating = lineData.at(i);
                        if (tmpMissionData.engineRating == "0" || tmpMissionData.engineRating == "") tmpMissionData.engineRating = "takeoff";
                    } else if (headerVector.at(i).find("Engine N1") != std::string::npos) {
                        // \todo list for all engines
                        tmpMissionData.N1 = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Angle of attack [deg]") != std::string::npos) {
                        tmpMissionData.AOA = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Spoiler Factor [-]") != std::string::npos) {
                        tmpMissionData.spoilerFactor = atof(lineData.at(i).c_str());
                    } else if (headerVector.at(i).find("Mode name [-]") != std::string::npos) {
                        lineData.at(i).erase(std::remove_if(lineData.at(i).begin(), lineData.at(i).end(), ::isspace), lineData.at(i).end());
                        tmpMissionData.modeName = lineData.at(i);
                    }
                }
                missionIOData_->myEngine.set_operating_point({tmpMissionData.N1, tmpMissionData.alt, tmpMissionData.Mach});
                tmpMissionData.Tamb = myAtm.getTemperature(tmpMissionData.alt);
                tmpMissionData.Pamb = myAtm.getPressure(tmpMissionData.alt);
                tmpMissionData.NOxFlow = missionIOData_->myEngine.get_engine_NOx_emission_index() * missionIOData_->myEngine.get_aircraft_fuelflow();
                tmpMissionData.St3MassFlow = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
                tmpMissionData.St3TotalTemperature = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalTemperature, EngineStage::St3);
                tmpMissionData.St3StaticTemperature = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
                tmpMissionData.St3TotalPressure = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
                tmpMissionData.St3Mach = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
                tmpMissionData.St4FuelAir = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::FuelToAirRatio, EngineStage::St4);
                tmpMissionData.BPR = (missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St22) > 0.) ?
                                     missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St13)
                                     / missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St22) : 0.;
                missionInfo.push_back(MissionData());
                missionInfo.back() = tmpMissionData;
            }
            lineData.clear();
            j = j + 1;
        }
        missionCSV.close();
    }
}

void Mission::get_mission_segment_data() {
    /* Calculate Fuel and distance lengths for different phases */
    /* Taxi (T/O and Landing)*/
    if (typeOfMission == "design") {
        m_fuel_Taxi = EndnodeReadOnly<double>("analysis/mission/design_mission/taxi_energy").read(rtIOPt->acxml).value()/rtIOPt->get_fuel_gravimetric_energy_density(0);
    } else {
        m_fuel_Taxi = EndnodeReadOnly<double>("analysis/mission/study_mission/taxi_energy").read(rtIOPt->acxml).value()/rtIOPt->get_fuel_gravimetric_energy_density(0);
    }

    unsigned int tmpLine = 1;//Line 0 would be first line with description of columns, therefore start at 1
    double tmpHeight = this->missionInfo.at(tmpLine).alt; //Height in meter as 3rd entry in csv file

    /* Takeoff */
    double TOheight = convertUnit(FOOT, METER, 35.); // T/O until 35 ft, afterwards ClimbOut, vgl. "Elements of Aircraft Pollution", von Ruijgrok, Fig. 11.4-1
    double LTOheight = Rounding(convertUnit(FOOT, METER, 3000.), 1); // ICAO: LTO Cycle until 3000 ft.

    /* Start until TO-Height */
    while (tmpHeight <= TOheight) {
        tmpLine = tmpLine + 1;
        tmpHeight = this->missionInfo.at(tmpLine).alt; //Height in meter as 3rd entry in csv file
    }
    tmpLine = tmpLine - 1;//while Schleife einmal zuviel durchgelaufen
    m_fuel_TO = -missionInfo.at(tmpLine).fuelConsumed;//Fuel consumed als 7. in csv-Datei
    takeOffTime = missionInfo.at(tmpLine).time - 0.;

    /* ClimbOut */
    while (tmpHeight <= LTOheight) {
        tmpLine = tmpLine + 1;
        tmpHeight = missionInfo.at(tmpLine).alt; //Height in meter as 3rd entry in csv file
    }
    tmpLine = tmpLine - 1;
    tmpHeight = missionInfo.at(tmpLine).alt;//necessary for query if (tmpHeight == missionInfo.at(tmpLine).alt) below in while loop
    m_fuel_ClimbOut = -missionInfo.at(tmpLine).fuelConsumed - m_fuel_TO; //Fuel consumed als 7. in csv-Datei
    climbOutTime = missionInfo.at(tmpLine).time - takeOffTime;

    /* LTO to 1st Cruise Step (already counts as "Cruise" for NOx calculation) */
    //Attention, gives rounding error and initialCruiseHeight greater than the height in the csv -> below query when 2 times same height comes in a row -> 1st Step
    double initalCruiseHeight = this->flightLevel.front() * convertUnit(FOOT, METER, 100.);
    while (tmpHeight < initalCruiseHeight) {
        tmpLine = tmpLine + 1;
        if (accuracyCheck(tmpHeight, missionInfo.at(tmpLine).alt, ACCURACY_HIGH)) { //To catch the rounding error (see double initalCruiseHeight
            initalCruiseHeight = tmpHeight;
        } else {
            if (tmpLine == missionInfo.size()-1) { //last precautionary measure, but should never come into play due to the previous query
                myRuntimeInfo->err << "Cruise Altitude < beforeInitalCruiseAlt; check if Acft.xml and mission.xml belong together" <<
                                   std::endl; //in case of repeated occurrence of controlled termination
                exit(1);
            }
            tmpHeight = missionInfo.at(tmpLine).alt; //Height in meter as 3. in csv file
        }
    }
    double lastCruiseHeight = this->flightLevel.back() * convertUnit(FOOT, METER, 100.);
    double beforeFinalCruiseAlt = floor(lastCruiseHeight);// instead of asking if height occurs 2 times in the same direction (see InitialCruise Height)
    if (this->flightLevel.size() > 1) {
        /* flown range from Initial Cruise to beginning of last Cruise Step*/
        while (tmpHeight < beforeFinalCruiseAlt) { //
            tmpLine = tmpLine + 1;
            tmpHeight = missionInfo.at(tmpLine).alt; //Height in meter 3rd entry in csv file
       }
    }
    //here last Cruise Step reached, which still must be flown before height is reduced again
    while (tmpHeight >= beforeFinalCruiseAlt) { //falls flightLevel.size()=1 finalCruiseAlt=initalCruiseAlt
        tmpLine = tmpLine + 1;
        tmpHeight = missionInfo.at(tmpLine).alt; //Height in meter as 3rd entry in csv file
    }
    tmpLine = tmpLine - 1;

    /*Descend until LTO Height*/
    while (tmpHeight >= LTOheight) {
        tmpLine = tmpLine + 1;
        tmpHeight = missionInfo.at(tmpLine).alt; //Height in meter as 3. entry in csv file
    }
    tmpLine = tmpLine - 1;//for NOx the FuelFlow up to 1st Cruise Step interesting, not for 1st Cruise Step
    m_fuel_Cruise = -missionInfo.at(tmpLine).fuelConsumed - m_fuel_TO - m_fuel_ClimbOut;
    double timeUntilApproach = missionInfo.at(tmpLine).time;

    /* Approach */
    while (tmpLine < (missionInfo.size() - 2)) { //tmpLine: 1.Linie=1, missionInfo.size: Vektor! 1.Line = 0; and then it goes through again!
        tmpLine = tmpLine + 1;
    }
    m_fuel_App = (-missionInfo.at(tmpLine).fuelConsumed + missionInfo.at(1).fuelConsumed) - m_fuel_Cruise - m_fuel_TO - m_fuel_ClimbOut;
    m_fuel_LTO = m_fuel_Taxi + m_fuel_TO + m_fuel_ClimbOut + m_fuel_App;
    approachTime = missionInfo.at(tmpLine).time - timeUntilApproach;
}

void Mission::write_emissions_path_csv() {
    std::ofstream of2;
    std::stringstream outputName;
    std::string emissionFileName;
    emissionFileName = missionFileName.substr(0, missionFileName.rfind("_out"));
    outputName << emissionFileName;
    std::string fn2 = outputName.str() + "_emissionspath.csv";
    of2.open(fn2.c_str());
    if (!of2) {
        myRuntimeInfo->err << "could not open " << fn2 << std::endl;
        exit(1);
    }
    of2 << "(1) Time [s];(2) Range [m];(3) Altitude [m];(4) TAS m/s];(5) Mach [-];(6) T_amb [K];(7) P_amb [N/m^2];(8) Consumed Fuel [kg];"
           "(9) Fuel Flow [kg/s];(10) NOx [kg/s];(11) N_1 [-];(12) ST3MassFlow [kg/s];(13) St3TotalTemperature [K];(14) St3StaticTemperature [K];"
           "(15) St3TotalPressure [kPa];(16) St3Mach [-];(17) St4FuelAir [-];(18) BPR;(19) modeName;";
    for (unsigned int i = 0; i < missionInfo.size(); i++) {
        of2 << std::endl;
        of2 << std::setprecision(10) << missionInfo.at(i).time << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).range << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).alt << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).TAS_ms <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).Mach <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).Tamb <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).Pamb <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).fuelConsumed <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).fuelFlow <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).NOxFlow <<  ";";
        of2 << std::setprecision(10) << missionInfo.at(i).N1 << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).St3MassFlow << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).St3TotalTemperature << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).St3StaticTemperature << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).St3TotalPressure << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).St3Mach << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).St4FuelAir << ";";
        of2 << std::setprecision(10) << missionInfo.at(i).BPR << ";";
        of2 << missionInfo.at(i).modeName << ";";
    }
    of2.close();
}

/* Functions for interpolation */

// cppcheck-suppress unusedFunction
double Mission::PointToPointFit(double REI_TO, double REI_CL, double REI_AP, double REI_ID,
                                int missionPoint) { //Linear interpolation of the ICAO data from NOx, where REI is the intersection of Wff with the interpolated curve.
    //Wf values from LTO ICAO cycle corrected with Boeing's correction factor r;
    //r = 1.100 for ID; r = 1.010 for takeoff; r = 1.013 for CL; r = 1.020 for AP;
    double Wf_ID = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * 7936 * 1.100; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_TO = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * 7936 * 1.010; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_CL = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::climb) * 7936 * 1.013; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_AP = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::approach) * 7936 * 1.020; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double tempWf = (this->myFlightPath.pathPoints.at(missionPoint).fuelFlow / this->missionIOData_->numberOfEngines) * 7936; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double REI = 0;
    if (tempWf <= Wf_AP) {
        REI = pow(10., (log10(REI_AP) - log10(REI_ID)) / (log10(Wf_AP) - log10(Wf_ID)) * (log10(tempWf) - log10(Wf_ID)) + log10(REI_ID));
    } else if ((tempWf > Wf_AP) && (tempWf <= Wf_CL)) {
        REI = pow(10., (log10(REI_CL) - log10(REI_AP)) / (log10(Wf_CL) - log10(Wf_AP)) * (log10(tempWf) - log10(Wf_AP)) + log10(REI_AP));
    } else {
        REI = pow(10., (log10(REI_TO) - log10(REI_CL)) / (log10(Wf_TO) - log10(Wf_CL)) * (log10(tempWf) - log10(Wf_CL)) + log10(REI_CL));
    }
    return REI;
}

double Mission::LinearLeastSquareFit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int missionPoint) {
    double Wf_ID = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * 7936 * 1.100; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_TO = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * 7936 * 1.010; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_CL = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::climb) * 7936 * 1.013; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_AP = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::approach) * 7936 * 1.020; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double tempWf = (this->myFlightPath.pathPoints.at(missionPoint).fuelFlow / this->missionIOData_->numberOfEngines) * 7936; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double p = log10(Wf_TO) + log10(Wf_CL) + log10(Wf_AP) + log10(Wf_ID);
    double q = log10(REI_TO) + log10(REI_CL) + log10(REI_AP) + log10(REI_ID);
    double r = pow(log10(Wf_TO), 2.) + pow(log10(Wf_CL), 2.) + pow(log10(Wf_AP), 2.) + pow(log10(Wf_ID), 2.);
    double s = log10(Wf_TO) * log10(REI_TO) + log10(Wf_CL) * log10(REI_CL) + log10(Wf_AP) * log10(REI_AP) + log10(Wf_ID) * log10(REI_ID);
    double REI = 0;
    int n = 4;
    REI = pow(10., (n * s - p * q) / (n * r - pow(p, 2.)) * log10(tempWf) + (q * r - p * s) / (n * r - pow(p, 2.)));
    if (REI > REI_TO) {
        REI = REI_TO;
        return REI;
    } else {
        return REI;
    }
}

// Calculation of the EI at reference conditions (REI), linear least square fit

double Mission::biLinearLeastSquareFit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int missionPoint) {
    //Wf values from LTO ICAO cycle corrected with Boeing's correction factor r;
    //r = 1.100 for ID; r = 1.010 for takeoff; r = 1.013 for CL; r = 1.020 for AP;
    double Wf_ID = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * 7936 * 1.100; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_AP = missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::approach) * 7936 * 1.020; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double tempWf = (this->myFlightPath.pathPoints.at(missionPoint).fuelFlow / this->missionIOData_->numberOfEngines) * 7936; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double p = log10(Wf_AP) + log10(Wf_ID);
    double q = log10(REI_AP) + log10(REI_ID);
    double r = pow(log10(Wf_AP), 2.) + pow(log10(Wf_ID), 2.);
    double s = log10(Wf_AP) * log10(REI_AP) + log10(Wf_ID) * log10(REI_ID);
    int n = 2;
    double Wf_intersect = 0; // [lbm/h]
    double REI = 0;
    Wf_intersect = pow(10., ((log10(REI_TO) + log10(REI_CL)) / 2. - (q * r - p * s) / (n * r - pow(p, 2.))) * (n * r - pow(p, 2.)) / (n * s - p * q));
    if (fabs(tempWf) < ACCURACY_HIGH) {
        REI = 0;
        return REI;
    } else {
        if (tempWf <= Wf_intersect) {
            REI = pow(10., (n * s - p * q) / (n * r - pow(p, 2.)) * log10(tempWf) + (q * r - p * s) / (n * r - pow(p, 2.)));
            if (REI > REI_ID) {
                REI = REI_ID;
                return REI;
            } else {
                return REI;
            }
        } else {
            REI = pow(10, (log10(REI_TO) + log10(REI_CL)) / 2.);
            return REI;
        }
    }
}

//potentialFit creates a graphic in the form y=ax^b
//when Y==false "input" should be a value in "Y-axis" and the value returned will be a value from "X-axis"
double Mission::potentialFit(double input, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = log10(x_ID) + log10(x_TO) + log10(x_CL) + log10(x_AP);
    double x_2 = pow(log10(x_ID), 2) + pow(log10(x_TO), 2) + pow(log10(x_CL), 2) + pow(log10(x_AP), 2);
    double y = log10(y_ID) + log10(y_TO) + log10(y_CL) + log10(y_AP);
    double y_2 = log10(x_ID) * log10(y_ID) + log10(x_TO) * log10(y_TO) + log10(x_CL) * log10(y_CL) + log10(x_AP) * log10(y_AP);
    double a = pow(10, (y_2 - y / x * x_2) / (x - 4 / x * x_2));
    double b = (y - 4 * log10(a)) / x;
    if (Y) {
        double x_mom = input;
        double y_mom = a * pow(x_mom, b);
        return y_mom;
    } else {
        double y_mom = input;
        double x_mom = pow(y_mom / a, 1 / b);
        return x_mom;
    }
}

// cppcheck-suppress unusedFunction
double Mission::potentialFit2(double input, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = log10(x_ID) + log10(x_TO) + log10(x_CL) + log10(x_AP);
    double x_2 = pow(log10(x_ID), 2) + pow(log10(x_TO), 2) + pow(log10(x_CL), 2) + pow(log10(x_AP), 2);
    double y = log10(y_ID) + log10(y_TO) + log10(y_CL) + log10(y_AP);
    double y_2 = log10(x_ID) * log10(y_ID) + log10(x_TO) * log10(y_TO) + log10(x_CL) * log10(y_CL) + log10(x_AP) * log10(y_AP);
    double a = pow(10, (y_2 - y / x * x_2) / (x - 4 / x * x_2));
    double b = (y - 4 * log10(a)) / x;
    if (Y) {
        double x_mom = input;
        double y_mom = a * pow(x_mom, b);
        return y_mom;
    } else {
        double y_mom = input;
        double x_mom = pow(y_mom / a, 1 / b);
        return x_mom;
    }
}

// cppcheck-suppress unusedFunction
double Mission::potentialFit3(double input, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = log10(x_ID) + log10(x_TO) + log10(x_CL) + log10(x_AP);
    double x_2 = pow(log10(x_ID), 2) + pow(log10(x_TO), 2) + pow(log10(x_CL), 2) + pow(log10(x_AP), 2);
    double y = log10(y_ID) + log10(y_TO) + log10(y_CL) + log10(y_AP);
    double y_2 = log10(x_ID) * log10(y_ID) + log10(x_TO) * log10(y_TO) + log10(x_CL) * log10(y_CL) + log10(x_AP) * log10(y_AP);
    double a = pow(10, (y_2 - y / x * x_2) / (x - 4 / x * x_2));
    double b = (y - 4 * log10(a)) / x;
    if (Y) {
        double x_mom = input;
        double y_mom = a * pow(x_mom, b);
        return y_mom;
    } else {
        double y_mom = input;
        double x_mom = pow(y_mom / a, 1 / b);
        return x_mom;
    }
}

double Mission::quadraticFit(double x_mom, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = x_ID + x_TO + x_CL + x_AP;
    double x_2 = pow(x_ID, 2) + pow(x_TO, 2) + pow(x_CL, 2) + pow(x_AP, 2);
    double x_3 = pow(x_ID, 3) + pow(x_TO, 3) + pow(x_CL, 3) + pow(x_AP, 3);
    double x_4 = pow(x_ID, 4) + pow(x_TO, 4) + pow(x_CL, 4) + pow(x_AP, 4);
    double y = y_ID + y_TO + y_CL + y_AP;
    double y_2 = x_ID * y_ID + x_TO * y_TO + x_CL * y_CL + x_AP * y_AP;
    double y_3 = pow(x_ID, 2) * y_ID + pow(x_TO, 2) * y_TO + pow(x_CL, 2) * y_CL + pow(x_AP, 2) * y_AP;
    double eigenval = 4 * x_2 * x_4 + 2 * x * x_3 * x_2 - pow(x_2, 3) - 4 * pow(x_3, 2) - pow(x, 2) * x_4;
    double a = (y * x_2 * x_4 + y_2 * x_3 * x_2 + y_3 * x * x_3 - pow(x_2, 2) * y_3 - y_2 * x * x_4 - y * pow(x_3, 2)) / eigenval;
    double b = (4 * y_2 * x_4 + x * y_3 * x_2 + y * x_3 * x_2 - pow(x_2, 2) * y_2 - 4 * x_3 * y_3 - x * x_4 * y) / eigenval;
    double c = (4 * x_2 * y_3 + x * x_3 * y + x * y_2 * x_2 - pow(x_2, 2) * y - 4 * x_3 * y_2 - pow(x, 2) * y_3) / eigenval;
    double y_mom = a + b * x_mom + c * pow(x_mom, 2);
    if (Y) {
        if (y_mom > std::max(y_ID, y_TO)) {
            y_mom = std::max(y_ID, y_TO);
            return y_mom;
        } else {
            return y_mom;
        }
    } else {
        return y_mom;
    }
}

double Mission::linearInterpolation(double x, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP) {
    if (x < x_ID) {
        double y = y_ID * x / y_ID;
        return y;
    } else if ((x_ID <= x) && (x < x_AP)) {
        double y = y_ID + (y_AP - y_ID) * (x - x_ID) / (y_AP - y_ID);
        return y;
    } else if ((x_AP <= x) && (x < x_CL)) {
        double y = y_AP + (y_CL - y_AP) * (x - x_AP) / (y_CL - y_AP);
        return y;
    } else if ((x_CL <= x) && (x < x_TO)) {
        double y = y_CL + (y_TO - y_CL) * (x - x_CL) / (y_TO - y_CL);
        return y;
    } else {
        double y = y_TO;
        return y;
    }
}

/* Constuctors and Destructor*/
Mission::~Mission() {
    //dtor
}

Mission::MissionData::MissionData()
    :
    time(0.),
    range(0.),
    alt(0.),
    TAS_ms(0.),
    CAS_ms(0.),
    mass(0.),
    fuelConsumed(0.),
    thrust(0.),
    fuelFlow(0.),
    TAS_kts(0.),
    Mach(0.),
    flightLevel(0.),
    ROC(0.),
    C_L(0.),
    SAR(0.),
    aeroConfig("TakeOff"),
    bleedAirOfftake(0.),
    shaftPowerOfftake(0.),
    engineRating("takeoff"),
    AOA(0.),
    spoilerFactor(1.),
    Tamb(0.),
    Pamb(0.),
    NOxFlow(0.),
    N1(0.),
    St3MassFlow(0.),
    St3TotalTemperature(0.),
    St3StaticTemperature(0.),
    St3TotalPressure(0.),
    St3Mach(0.),
    St4FuelAir(0.),
    BPR(0.) {
        //ctor
    };

