
/**
 * \file bwbVerticalTailsData.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Blended Wing Body Vertical Tails Data source file
 * \version 0.1
 * \date 2024-04-17
 *
 * @copyright Copyright (c) 2024
 *
 */

#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/geometry/surface.h>
#include <aixml/node.h>

#include "bwbVerticalTailsData.h"

namespace bwb {
namespace low {
VerticalTailsData::VerticalTailsData() :
  wing_reference_position(PositionIO("component_design/wing/position", "wing")),
  maximum_takeoff_mass(MassPropertiesIO("analysis/masses_cg_inertia/maximum_takeoff_mass/","maximum_takeoff_mass")),
  empennage_mass(MassPropertiesIO("component_design/empennage/", "Vertical tails mass")),
  empennage_position(PositionIO("component_design/empennage/position", "Vertical tails position")) {}

void VerticalTailsData::read(const node& xml) {
  wing_reference_position.read(xml);
  maximum_takeoff_mass.read(xml);
}
void VerticalTailsData::read_wing(std::filesystem::path path, std::filesystem::path airfoil_data_dir) {
  std::shared_ptr<node> acxml = aixml::openDocument(path);
  geom2::WingFactory wing_factory{acxml, airfoil_data_dir};
  wing = wing_factory.create("wing/specific/geometry/aerodynamic_surface@0");
}
void VerticalTailsData::update(node& xml) {
  update_vertical_tails(xml);
  empennage_mass.update(xml);
  empennage_position.update(xml);
  
}
void VerticalTailsData::update_vertical_tails(node& xml) {
  for (size_t id = 0; id < tails.size(); ++id) {
    geom2::io::SurfaceType io = geom2::io::Wing(tails[id]);
    std::visit(geom2::io::AixmlConverter(xml.at("component_design/empennage/specific/geometry/"),
                                         {"aerodynamic_surface", std::to_string(id), tails[id].name}),
               io);
    tails_mass[id].update(xml);
  }
}
}  // namespace low
}  // namespace bwb
