/**
 * \file inputs.h
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief
 * \version 0.1
 * \date 2023-11-15
 *
 * @copyright Copyright (c) 2023
 *
 */
#ifndef EMPENNAGEDESIGN_SRC_LIB_IOMETHODS_IO_H_
#define EMPENNAGEDESIGN_SRC_LIB_IOMETHODS_IO_H_

#include <aixml/endnode.h>

class RelativePosition {
 public:
  RelativePosition() = default;
  RelativePosition(const std::string& origin)
      : inner_position_spanwise(EndnodeReadOnly<double>(origin + "position/inner_position/spanwise")),
        inner_position_chord_from(EndnodeReadOnly<double>(origin + "position/inner_position/chord/from")),
        inner_position_chord_to(EndnodeReadOnly<double>(origin + "position/inner_position/chord/to")),
        outer_position_spanwise(EndnodeReadOnly<double>(origin + "position/outer_position/spanwise")),
        outer_position_chord_from(EndnodeReadOnly<double>(origin + "position/outer_position/chord/from")),
        outer_position_chord_to(EndnodeReadOnly<double>(origin + "position/outer_position/chord/to")) {}
  EndnodeReadOnly<double> inner_position_spanwise;
  EndnodeReadOnly<double> inner_position_chord_from;
  EndnodeReadOnly<double> inner_position_chord_to;
  EndnodeReadOnly<double> outer_position_spanwise;
  EndnodeReadOnly<double> outer_position_chord_from;
  EndnodeReadOnly<double> outer_position_chord_to;

  void read(const node& xml) {
    inner_position_spanwise.read(xml);
    inner_position_chord_from.read(xml);
    inner_position_chord_to.read(xml);
    outer_position_spanwise.read(xml);
    outer_position_chord_from.read(xml);
    outer_position_chord_to.read(xml);
  }
};

class Spar {
 public:
  Spar() = default;
  Spar(const std::string& origin)
      : position(RelativePosition(origin)),
        name(EndnodeReadOnly<std::string>(origin + "name")) {}
  EndnodeReadOnly<std::string> name;
  RelativePosition position;


  void read(const node& xml) {
    name.read(xml);
    position.read(xml);
  }
};

class ControlDevice {
 public:
  ControlDevice() = default;
  ControlDevice(const std::string& origin)
      : type(EndnodeReadOnly<std::string>(origin + "type")),
        deflection_full_negative(EndnodeReadOnly<double>(origin + "deflection/full_negative_deflection")),
        deflection_full_positive(EndnodeReadOnly<double>(origin + "deflection/full_positive_deflection")),
        position(RelativePosition(origin)) {}
  EndnodeReadOnly<std::string> type;
  EndnodeReadOnly<double> deflection_full_negative;
  EndnodeReadOnly<double> deflection_full_positive;
  RelativePosition position;

  void read(const node& xml) {
    type.read(xml);
    deflection_full_negative.read(xml);
    deflection_full_positive.read(xml);
    position.read(xml);
  }
};

class TailElement {
 public:
  TailElement(const std::string& origin)
      : origin(origin),
        name(EndnodeReadOnly<std::string>(origin + "name")),
        rear_x_offset(EndnodeReadOnly<double>(origin + "parameter/offset/rear_x_offset")),
        centerline_y_offset(EndnodeReadOnly<double>(origin + "parameter/offset/centerline_y_offset")),
        centerline_z_offset(EndnodeReadOnly<double>(origin + "parameter/offset/centerline_z_offset")),
        volume_coefficient(EndnodeReadOnly<double>(origin + "parameter/volume_coefficient")),
        factor_aspect_ratio(EndnodeReadOnly<double>(origin + "parameter/factor_aspect_ratio")),
        factor_taper_ratio(EndnodeReadOnly<double>(origin + "parameter/factor_taper_ratio")),
        delta_sweep(EndnodeReadOnly<double>(origin + "parameter/delta_sweep")),
        inner_profile(EndnodeReadOnly<std::string>(origin + "profiles/profile@0/name")),
        outer_profile(EndnodeReadOnly<std::string>(origin + "profiles/profile@1/name")) {}
  ~TailElement() = default;

  const std::string origin;
  EndnodeReadOnly<std::string> name;
  EndnodeReadOnly<double> rear_x_offset;
  EndnodeReadOnly<double> centerline_y_offset;
  EndnodeReadOnly<double> centerline_z_offset;
  EndnodeReadOnly<double> volume_coefficient;
  EndnodeReadOnly<double> factor_aspect_ratio;
  EndnodeReadOnly<double> factor_taper_ratio;
  EndnodeReadOnly<double> delta_sweep;
  EndnodeReadOnly<std::string> inner_profile;
  EndnodeReadOnly<std::string> outer_profile;

  std::vector<Spar> spars;
  std::vector<ControlDevice> control_devices;

  void read(const node& xml) {
    name.read(xml);
    rear_x_offset.read(xml);
    centerline_y_offset.read(xml);
    centerline_z_offset.read(xml);
    volume_coefficient.read(xml);
    factor_aspect_ratio.read(xml);
    factor_taper_ratio.read(xml);
    delta_sweep.read(xml);
    inner_profile.read(xml);
    outer_profile.read(xml);

    size_t spars_available = xml.getVector(origin + "spars/spar", 1).size();
    for (size_t i = 0; i < spars_available; ++i) {
      spars.push_back(Spar(origin + "spars/spar@" + std::to_string(i) + "/"));
      spars.back().read(xml);
    }

    size_t control_devices_available = xml.getVector(origin + "control_devices/control_device", 1).size();
    for (size_t i = 0; i < control_devices_available; ++i) {
      control_devices.push_back(ControlDevice(origin + "control_devices/control_device@" + std::to_string(i) + "/"));
      control_devices.back().read(xml);
    }

  }
};

#endif  // EMPENNAGEDESIGN_SRC_LIB_IOMETHODS_IO_H_
