"""Module providing functions for the preparation of user data."""

import pycoordinatesystemconversion as py11csc


def user_method_data_input_preparation(routing_dict):
    """Prepare necessary input data for the user method from aircraft exchange and module configuration files.

    In this function, the user is responsible for preparing the data needed for the user method. Relevant general data
    are obtained from the aircraft exchange file and calculation specific parameter from the module configuration file.
    The user must submit the data in the following format:
        dict = {'parameter_name': [path to parameter node, expected data type], ...}

    :param dict routing_dict: Dictionary containing information on necessary data from module configuration file
    :returns:
        - dict data_to_extract_from_aircraft_exchange_dict: Dictionary containing parameter name, path to parameter,
        and expected data type of parameters to be extracted from aircraft exchange file
        - dict data_to_extract_from_module_configuration_dict: Dictionary containing parameter name, path to parameter,
        and expected data type of parameters to be extracted from module configuration file
    """

    """Aircraft exchange file."""
    # Enter all parameters to be extracted from the aircraft exchange file.
    data_to_extract_from_aircraft_exchange_dict = {
        'additional_cargo_mass':
            ['./requirements_and_specifications/design_specification/transport_task/cargo_definition/additional_cargo_mass', float],  # noPep8 e501
        'additional_fuselage_length':
            ['./component_design/tank/specific/additional_fuselage_length', float],
        'aerodynamic_surface_name':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/name', str],
        'aerodynamic_surface_position_in_x':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/position/x', float],
        'aerodynamic_surface_position_in_y':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/position/y', float],
        'aerodynamic_surface_position_in_z':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/position/z', float],
        'aerodynamic_surface_section_position_in_x':
            ['./component_design/wing/specific/geometry/aerodynamic_surface/parameters/sections/section[@ID="0"]/chord_origin/x', float],  # noPep8 e501
        'aerodynamic_surface_section_position_in_y':
            ['./component_design/wing/specific/geometry/aerodynamic_surface/parameters/sections/section[@ID="0"]/chord_origin/y', float],  # noPep8 e501
        'aerodynamic_surface_section_position_in_z':
            ['./component_design/wing/specific/geometry/aerodynamic_surface/parameters/sections/section[@ID="0"]/chord_origin/z', float],  # noPep8 e501
        'aerodynamic_surface_section_length':
            ['./component_design/wing/specific/geometry/aerodynamic_surface/parameters/sections/section[@ID="0"]/chord_length', float],  # noPep8 e501
        'cargo_density':
            ['./requirements_and_specifications/design_specification/transport_task/cargo_definition/cargo_density', float],  # noPep8 e501
        'class_distribution':
            ['./requirements_and_specifications/design_specification/transport_task/passenger_class_definition/class_distribution', str],  # noPep8 e501
        'configuration_type':
            ['./requirements_and_specifications/design_specification/configuration/configuration_type', str],
        'design_mission_range':
            ['./requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/range', float],  # noPep8 e501
        'fuselage_type':
            ['./requirements_and_specifications/design_specification/configuration/fuselage_definition/fuselage_type', str],  # noPep8 e501
        'fuselage_pressurization':
            ['./requirements_and_specifications/design_specification/configuration/fuselage_definition/fuselage_pressurization', str],  # noPep8 e501
        'global_reference_point_in_x_direction':
            ['./component_design/global_reference_point/x', float],
        'global_reference_point_in_y_direction':
            ['./component_design/global_reference_point/y', float],
        'global_reference_point_in_z_direction':
            ['./component_design/global_reference_point/z', float],
        'luggage_mass_per_passenger':
            ['./requirements_and_specifications/design_specification/transport_task/passenger_definition/luggage_mass_per_passenger', float],  # noPep8 e501
        'main_gear_mounting_position':
            ['./requirements_and_specifications/design_specification/configuration/undercarriage_definition/main_gear_mounting', str],  # noPep8 e501
        'maximum_take_off_mass':
            ['./analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass', float],
        'most_aft_center_of_gravity_in_x_direction':
            ['./analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/x', float],
        'most_aft_center_of_gravity_in_y_direction':
            ['./analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/y', float],
        'most_aft_center_of_gravity_in_z_direction':
            ['./analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/z', float],
        'most_forward_center_of_gravity_in_x_direction':
            ['./analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/x', float],
        'most_forward_center_of_gravity_in_y_direction':
            ['./analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/y', float],
        'most_forward_center_of_gravity_in_z_direction':
            ['./analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/z', float],
        'multi_passenger_deck_layout':
            ['./requirements_and_specifications/design_specification/transport_task/multi_passenger_deck_layout/', bool],  # noPep8 e501
        'propulsor_mounting_position':
            ['./requirements_and_specifications/design_specification/propulsion/propulsor[@ID="0"]/position/parent_component', str],  # noPep8 e501
        'specific_number_of_passenger_decks':
            ['./requirements_and_specifications/design_specification/transport_task/multi_passenger_deck_layout/specific_number_of_passenger_decks', int],  # noPep8 e501
        'total_number_of_passengers':
            ['./requirements_and_specifications/design_specification/transport_task/passenger_definition/total_number_passengers', int],  # noPep8 e501
        'wing_mounting':
            ['./requirements_and_specifications/design_specification/configuration/wing_definition/mounting', str],
        'wing_reference_point_global_in_x':
            ['./component_design/wing/position/x', float],
        'wing_reference_point_global_in_y':
            ['./component_design/wing/position/y', float],
        'wing_reference_point_global_in_z':
            ['./component_design/wing/position/z', float],
        'wing_section_chord_length':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/parameters/sections/section[@ID="0"]/chord_length', float],  # noPep8 e501
        'wing_section_position_in_x':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/parameters/sections/section[@ID="0"]/chord_origin/x', float],  # noPep8 e501
        'wing_section_position_in_y':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/parameters/sections/section[@ID="0"]/chord_origin/z', float],  # noPep8 e501
        'wing_section_position_in_z':
            ['./component_design/wing/specific/geometry/aerodynamic_surface[@ID="0"]/parameters/sections/section[@ID="0"]/chord_origin/y', float],  # noPep8 e501
        }

    """Module configuration file."""
    # Enter all general parameters to be extracted from the module configuration file. 'general parameters' means
    # parameters that do not differ according to the user layer. It should be noted that 'tmp_general' is only used to
    # shorten the path information in the 'general_data_to_extract_from_module_configuration_dict'.
    tmp_general = ('./program_settings/configuration[@ID="tube_and_wing"]/fidelity[@ID="empirical"]/'
                   + 'fuselage_design_tu_berlin/general')
    general_data_to_extract_from_module_configuration_dict = {
        'flight_attendant_mass':
            [tmp_general + '/flight_attendant_mass', float],
        'force_circular_cross_section':
            [tmp_general + '/force_circular_cross_section', bool],
        'high_density_seat_pitch_for_single_aisle':
            [tmp_general + '/high_density_seat_pitch/high_density_seat_pitch_for_single_aisle', float],
        'high_density_seat_pitch_for_wide_body':
            [tmp_general + '/high_density_seat_pitch/high_density_seat_pitch_for_wide_body', float],
        'length_of_flight_deck':
            [tmp_general + '/length_of_flight_deck', float],
        'fuselage_structural_mass_technology_factor':
            [tmp_general + '/mass_technology_factors/fuselage_structural_mass_technology_factor', float],
        'name_of_existing_geometry_file':
            [tmp_general + '/sizing_mode/setting_use_existing_fuselage/name_of_existing_geometry_file', str],
        'name_of_fuselage':
            [tmp_general + '/name_of_fuselage', str],
        'pax_per_full_size_trolley_long_range':
            [tmp_general + '/on_board_service_load/pax_per_full_size_trolley_long_range', float],
        'pax_per_full_size_trolley_medium_range':
            [tmp_general + '/on_board_service_load/pax_per_full_size_trolley_medium_range', float],
        'pax_per_full_size_trolley_short_range':
            [tmp_general + '/on_board_service_load/pax_per_full_size_trolley_short_range', float],
        'pax_per_full_size_trolley_regional_range':
            [tmp_general + '/on_board_service_load/pax_per_full_size_trolley_regional_range', float],
        'sizing_mode':
            [tmp_general + '/sizing_mode', str],
        'system_bay_height_above_cabin':
            [tmp_general + '/system_bay_height_above_cabin', float],
        'use_as_starting_point':
            [tmp_general + '/sizing_mode/setting_use_existing_fuselage/use_as_starting_point', bool]
        }

    # Enter all specific parameters to be extracted from the module configuration file. 'specific parameters' means
    # parameters that differ according to the user layer. It should be noted that 'tmp_specific' is only used to
    # shorten the path information in the 'specific_data_to_extract_from_module_configuration_dict'.
    tmp_specific = ('./program_settings/configuration[@ID="tube_and_wing"]/fidelity[@ID="empirical"]/'
                    + 'fuselage_design_tu_berlin/specific')
    specific_data_to_extract_from_module_configuration_dict = {
        'cabin_floor_offset':
            [tmp_specific + '/cabin_floor_offset', float],
        'uld_type':
            [tmp_specific + '/cargo_accommodation/uld_type', str],
        'container_name':
            [tmp_specific + '/cargo_accommodation/uld_type/container/name', str],
        'use_specific_number_of_ld_container':
            [tmp_specific + '/cargo_accommodation/uld_type/container/use_specific_number_of_ld_container', bool],
        'specific_number_of_ld_container':
            [tmp_specific + '/cargo_accommodation/uld_type/container/use_specific_number_of_ld_container/specific_number_of_ld_container', int],  # noPep8 e501
        'pallet_name':
            [tmp_specific + '/cargo_accommodation/uld_type/pallet/name', str],
        'use_specific_number_of_cargo_pallet':
            [tmp_specific + '/cargo_accommodation/uld_type/pallet/use_specific_number_of_cargo_pallet', bool],
        'specific_number_of_cargo_pallet':
            [tmp_specific + '/cargo_accommodation/uld_type/pallet/use_specific_number_of_cargo_pallet/specific_number_of_cargo_pallet', int],  # noPep8 e501
        'use_internal_bulk_area':
            [tmp_specific + '/cargo_accommodation/use_internal_bulk_area', bool],
        'specific_internal_bulk_area':
            [tmp_specific + '/cargo_accommodation/use_internal_bulk_area/specific_internal_bulk_area', float],
        'dead_end_cabin_arrangement':
            [tmp_specific + '/dead_end_cabin_arrangement', bool],
        'dead_end_cabin_arrangement_in_front':
            [tmp_specific + '/dead_end_cabin_arrangement/dead_end_cabin_arrangement_in_front', bool],
        'dead_end_cabin_arrangement_in_aft':
            [tmp_specific + '/dead_end_cabin_arrangement/dead_end_cabin_arrangement_in_aft', bool],
        'auto_select_exit_types':
            [tmp_specific + '/exit_definition/auto_select_exit_types', bool],
        'select_specific_exit_types':
            [tmp_specific + '/exit_definition/auto_select_exit_types/select_specific_exit_types/exit[@ID="0"]', str],
        'plus_certification_allowed':
            [tmp_specific + '/exit_definition/plus_certification_allowed', bool],
        'fuselage_frame_distance':
            [tmp_specific + '/fuselage_frame_distance', float],
        'set_divider_automatically':
            [tmp_specific + '/passenger_accommodation/divider_definition/set_divider_automatically', bool],
        'use_soft_divider':
            [tmp_specific + '/passenger_accommodation/divider_definition/set_divider_automatically/use_soft_divider', bool],  # noPep8 e501
        'hard_divider_thickness':
            [tmp_specific + '/passenger_accommodation/divider_definition/set_divider_automatically/use_soft_divider/hard_divider_thickness', float],  # noPep8 e501
        'set_divider_width_automatically':
            [tmp_specific + '/passenger_accommodation/divider_definition/set_divider_automatically/divider_width/set_divider_width_automatically', bool],  # noPep8 e501
        'specific_class_divider_width':
            [tmp_specific + '/passenger_accommodation/divider_definition/set_divider_automatically/divider_width/set_divider_width_automatically/specific_class_divider_width', float],  # noPep8 e501
        'use_in_flight_entertainment':
            [tmp_specific + '/passenger_accommodation/in_flight_entertainment/use_in_flight_entertainment', bool],
        'passenger_class_name':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/name', str],
        'passenger_class_abreast':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/abreast', str],
        'passenger_seat_on_left_side_name':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seats/seat_on_left_side/name', str],  # noPep8 e501
        'passenger_seat_on_left_side_pitch':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seats/seat_on_left_side/seat_pitch', int],  # noPep8 e501
        'passenger_seat_on_right_side_name':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seats/seat_on_right_side/name', str],  # noPep8 e501
        'passenger_seat_on_right_side_pitch':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seats/seat_on_right_side/seat_pitch', int],  # noPep8 e501
        'passenger_seat_in_the_center_name':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seats/seat_in_the_center/name', str],  # noPep8 e501
        'passenger_seat_in_the_center_pitch':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seats/seat_in_the_center/seat_pitch',  int],  # noPep8 e501
        'passenger_seat_space':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/seat_space', float],
        'passenger_lavatory_name':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/lavatories/name', str],
        'passenger_per_lavatory':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/lavatories/pax_per_lavatory', int],  # noPep8 e501
        'use_specific_number_of_class_lavatories':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/lavatories/use_specific_number_of_class_lavatories', bool],  # noPep8 e501
        'specific_number_of_class_lavatories':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/lavatories/use_specific_number_of_class_lavatories/specific_number_of_class_lavatories', int],  # noPep8 e501
        'use_wardrobe_for_passenger_class':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/wardrobe/use_wardrobe_for_passenger_class/', bool],  # noPep8 e501
        'wardrobe_space_per_passenger':
            [tmp_specific + '/passenger_accommodation/passenger_classes/class[@ID="0"]/wardrobe/use_wardrobe_for_passenger_class/space_per_passenger', float],  # noPep8 e501
        'use_specific_number_of_full_size_trolleys':
            [tmp_specific + '/passenger_accommodation/galley_definition/use_specific_number_of_full_size_trolleys', bool],  # noPep8 e501
        'specific_number_of_full_size_trolleys':
            [tmp_specific + '/passenger_accommodation/galley_definition/use_specific_number_of_full_size_trolleys/specific_number_of_full_size_trolleys',  int],  # noPep8 e501
        'use_specific_number_of_half_size_trolleys':
            [tmp_specific + '/passenger_accommodation/galley_definition/use_specific_number_of_half_size_trolleys', bool],  # noPep8 e501
        'specific_number_of_half_size_trolleys':
            [tmp_specific + '/passenger_accommodation/galley_definition/use_specific_number_of_half_size_trolleys/specific_number_of_half_size_trolleys', int],  # noPep8 e501
        'use_specific_number_of_standard_units':
            [tmp_specific + '/passenger_accommodation/galley_definition/use_specific_number_of_standard_units', bool],
        'specific_number_of_standard_units':
            [tmp_specific + '/passenger_accommodation/galley_definition/use_specific_number_of_standard_units/specific_number_of_standard_units', int],  # noPep8 e501
        'passenger_aisle_width':
            [tmp_specific + '/passenger_aisle/width', float],
        'passenger_aisle_standing_height':
            [tmp_specific + '/passenger_aisle/standing_height', float],
        'passenger_standing_height_seat_row':
            [tmp_specific + '/passenger_aisle/standing_height_seat_row', float]
        }

    # Merge module configuration dictionaries.
    data_to_extract_from_module_configuration_dict = general_data_to_extract_from_module_configuration_dict | \
        specific_data_to_extract_from_module_configuration_dict

    return data_to_extract_from_aircraft_exchange_dict, data_to_extract_from_module_configuration_dict


def user_method_data_output_preparation(data_dict):
    """Prepare user-specific output data based on the calculation method results.

    This function is responsible for preparing the user-specific output data based on the results of the calculation
    method. The 'data_dict' input parameter contains the results of the module execution.
    The data for the key parameters output must be specified in the following format in order to be written correctly
    to the aircraft exchange file by the 'write_key_data_to_aircraft_exchange_file' function in the following step:
        dict = {'parameter_name': [path to parameter node, value, name (if needed)], ...}
    Important notes:
        (1) It should be noted that only key parameters may be written that have been previously defined by the module
        manager.
        (2) Attention must be paid to the proper path specifications, otherwise warnings may be issued or, in the worst
        case, errors may occur subsequently resulting in the write process and consequently the entire program being
        aborted.
        (3) If the path specifications contain IDs, these must start at '0' and be defined in ascending order without
        gaps.
    For the method-specific output, a path list and a dictionary is necessary to properly write the data to the
    method-specific XML file.
    the dictionary must be specified in the following format:
        dict = ...
    Note: If the user wants to export data from the design and study mission, two path lists and dictionaries are
    necessary.

    :param dict data_dict: Dictionary containing the results of the module execution
    :returns:
        - dict key_output_dict: Output dictionary containing key parameters that are written to aircraft XML file
        - dict method_specific_output_dict: Dictionary containing specific parameters that are written to
        method-specific output XML
    """
    ''' initial local parameter '''
    convert_inch_to_meter = 0.0254
    convert_cubic_inch_to_cubic_meter = 1.63871e-5
    """Key parameters output."""
    total_position_path = './component_design/fuselage/position/'
    total_mass_path = './component_design/fuselage/mass_properties/'
    total_inertia_path = './component_design/fuselage/mass_properties/inertia/'
    total_cog_path = './component_design/fuselage/mass_properties/center_of_gravity/'
    geometry_file_path = './component_design/fuselage/specific/geometry/'

    key_output_dict = {
        # total fuselage parameter
        'x':
            [total_position_path + 'x',
             data_dict['positions']['total_position']['x']],
        'y':
            [total_position_path + 'y',
             data_dict['positions']['total_position']['y']],
        'z':
            [total_position_path + 'z',
             data_dict['positions']['total_position']['z']],
        'mass':
            [total_mass_path + 'mass',
             round(data_dict['fuselage_design_masses']['total_fuselage_mass'], 9)],
        'j_xx':
            [total_inertia_path + 'j_xx',
             0.0],
        'j_yy':
            [total_inertia_path + 'j_yy',
             0.0],
        'j_zz':
            [total_inertia_path + 'j_zz',
             0.0],
        'j_xy':
            [total_inertia_path + 'j_xy',
             0.0],
        'j_xz':
            [total_inertia_path + 'j_xz',
             0.0],
        'j_yx':
            [total_inertia_path + 'j_yx',
             0.0],
        'j_yz':
            [total_inertia_path + 'j_yz',
             0.0],
        'j_zx':
            [total_inertia_path + 'j_zx',
             0.0],
        'j_zy':
            [total_inertia_path + 'j_zy',
             0.0],
        'cog_x':
            [total_cog_path + 'x',
             data_dict['center_of_gravities']['total_cog']['x']],
        'cog_y':
            [total_cog_path + 'y',
             data_dict['center_of_gravities']['total_cog']['y']],
        'cog_z':
            [total_cog_path + 'z',
             data_dict['center_of_gravities']['total_cog']['z']]
    }

    # generate output dictionary for each fuselage
    j = 0
    while j < data_dict['geometry_description']['number_of_fuselages']:
        fuselage_path_id = './component_design/fuselage/specific/geometry/fuselage[@ID="' + str(j) + '"]/'
        fuselage_dict = {
            # parameter for nose gear assemblies
            'name_id' + str(j):
                [fuselage_path_id + 'name',
                 'fuselage_' + str(j)],
            'x_id' + str(j):
                [fuselage_path_id + 'position/x',
                 data_dict['positions']['fuselage_position']['x']],
            'y_id' + str(j):
                [fuselage_path_id + 'position/y',
                 data_dict['positions']['fuselage_position']['y']],
            'z_id' + str(j):
                [fuselage_path_id + 'position/z',
                 data_dict['positions']['fuselage_position']['z']],
            'x_direction' + str(j):
                [fuselage_path_id + 'direction/x',
                 1],
            'y_direction' + str(j):
                [fuselage_path_id + 'direction/y',
                 0],
            'z_direction' + str(j):
                [fuselage_path_id + 'direction/z',
                 0],
            'mass_id' + str(j):
                [fuselage_path_id + 'mass_properties/mass',
                 round(data_dict['fuselage_design_masses']['total_fuselage_mass'], 9)],
            'j_xx_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_xx',
                 0.0],
            'j_yy_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_yy',
                 0.0],
            'j_zz_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_zz',
                 0.0],
            'j_xy_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_xy',
                 0.0],
            'j_xz_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_xz',
                 0.0],
            'j_yx_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_yx',
                 0.0],
            'j_yz_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_yz',
                 0.0],
            'j_zx_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_zx',
                 0.0],
            'j_zy_id' + str(j):
                [fuselage_path_id + 'mass_properties/inertia/j_zy',
                 0.0],
            'cog_x_id' + str(j):
                [fuselage_path_id + 'mass_properties/center_of_gravity/x',
                 data_dict['center_of_gravities']['fuselage_cog']['x']],
            'cog_y_id' + str(j):
                [fuselage_path_id + 'mass_properties/center_of_gravity/y',
                 data_dict['center_of_gravities']['fuselage_cog']['y']],
            'cog_z_id' + str(j):
                [fuselage_path_id + 'mass_properties/center_of_gravity/z',
                 data_dict['center_of_gravities']['fuselage_cog']['z']],
            'number_of_required_cabin_crew_id' + str(j):
                [fuselage_path_id + 'number_of_required_cabin_crew',
                 data_dict['geometry_description']['required_number_of_flight_attendants']],
            'accommodation_x' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/position/x',
                 data_dict['positions']['fuselage_accommodation_position']['x']],
            'accommodation_y' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/position/y',
                 data_dict['positions']['fuselage_accommodation_position']['y']],
            'accommodation_z' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/position/z',
                 data_dict['positions']['fuselage_accommodation_position']['z']],
            'accommodation_mass_id' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/mass_properties/mass',
                 round(data_dict['fuselage_design_masses']['fuselage_furnishing_masses']['total_furnishing_group_mass']
                       + data_dict['fuselage_design_masses']['fuselage_operator_item_masses']['total_group_mass'], 9)],
            'accommodation_cog_x_id' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/mass_properties/center_of_gravity/x',
                 data_dict['center_of_gravities']['fuselage_accommodation_cog']['x']],
            'accommodation_cog_y_id' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/mass_properties/center_of_gravity/y',
                 data_dict['center_of_gravities']['fuselage_accommodation_cog']['y']],
            'accommodation_cog_z_id' + str(j):
                [fuselage_path_id + 'fuselage_accommodation/mass_properties/center_of_gravity/z',
                 data_dict['center_of_gravities']['fuselage_accommodation_cog']['z']],
        }

        # generate payload tube dictionary for current fuselage
        i = 0
        while i < data_dict['geometry_description']['number_of_payload_tubes']:
            payload_tube_path = './component_design/fuselage/specific/geometry/fuselage[@ID="' \
                                + str(j) + '"]/fuselage_accommodation/payload_tube[@ID="' + str(i) + '"]/'
            if i == 0:
                payload_tune_name = 'center_payload_tube'
            else:
                payload_tune_name = 'payload_tube_' + str(i)
            payload_tube_output_dict = {
                # total section parameter
                'tube_name_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'name',
                     payload_tune_name],
                'tube_front_x_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/front_reference_points/x',
                     0.0],
                'tube_front_y_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/front_reference_points/y',
                     0.0],
                'tube_front_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/front_reference_points/z',
                     0.0],
                'tube_front_upper_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/front_reference_points/upper_z',
                     round(data_dict['geometry_description']['cabin_outer_upper_z'] * convert_inch_to_meter, 9)],
                'tube_front_lower_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/front_reference_points/lower_z',
                     - round(data_dict['geometry_description']['cabin_lower_z'] * convert_inch_to_meter, 9)],
                'tube_aft_x_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/aft_reference_points/x',
                     round(data_dict['geometry_description']['total_cabin_length'] * convert_inch_to_meter, 9)],
                'tube_aft_y_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/aft_reference_points/y',
                     0.0],
                'tube_aft_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/aft_reference_points/z',
                     0.0],
                'tube_aft_upper_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/aft_reference_points/upper_z',
                     round(data_dict['geometry_description']['cabin_outer_upper_z'] * convert_inch_to_meter, 9)],
                'tube_aft_lower_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_reference_points/aft_reference_points/lower_z',
                     - data_dict['geometry_description']['cabin_lower_z'] * convert_inch_to_meter],
                'tube_front_wall_x_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/front_reference_points/x',
                     0.0],
                'tube_front_wall_left_y_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/front_reference_points/left_y',
                     - round(data_dict['geometry_description']['total_cabin_width'] / 2 * convert_inch_to_meter, 9)],
                'tube_front_wall_right_y_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/front_reference_points/right_y',
                     round(data_dict['geometry_description']['total_cabin_width'] / 2 * convert_inch_to_meter, 9)],
                'tube_front_wall_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/front_reference_points/z',
                     0.0],
                'tube_aft_wall_x_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/aft_reference_points/x',
                     0.0],
                'tube_aft_wall_left_y_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/aft_reference_points/left_y',
                     - round(data_dict['geometry_description']['total_cabin_width'] / 2 * convert_inch_to_meter, 9)],
                'tube_aft_wall_right_y_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/aft_reference_points/right_y',
                     round(data_dict['geometry_description']['total_cabin_width'] / 2 * convert_inch_to_meter, 9)],
                'tube_aft_wall_z_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_wall_reference_points/aft_reference_points/z',
                     0.0],
                'payload_tube_structural_wall_thickness_id' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_structural_wall_thickness',
                     round(data_dict['geometry_description']['cabin_wall_thickness'] * convert_inch_to_meter, 9)],
                'payload_tube_water_volume' + str(j) + '_id' + str(i):
                    [payload_tube_path + 'payload_tube_water_volume',
                     round(data_dict['geometry_description']['fuselage_dimension']['fuselage_volume']
                           * convert_cubic_inch_to_cubic_meter * 0.5, 9)],
            }

            # generate payload deck dictionary for current fuselage
            k = 0
            sum_of_fst = 0
            payload_deck_name = 'first_deck'
            payload_tube_water_volume = \
                round(data_dict['geometry_description']['fuselage_dimension']['fuselage_volume']
                      * convert_cubic_inch_to_cubic_meter * 0.5, 9)
            for key, value in data_dict['interior_description'].items():
                sum_of_fst += data_dict['interior_description'][key]['number_of_full_size_trolley']

            while k < len(data_dict['positions']['payload_deck_position']):
                payload_deck_path = './component_design/fuselage/specific/geometry/fuselage[@ID="' + str(j) \
                                    + '"]/fuselage_accommodation/payload_tube[@ID="' + str(i) \
                                    + '"]/payload_decks/payload_deck[@ID="' + str(k) + '"]/'
                z = 0
                for key, value in data_dict['positions']['payload_deck_position'].items():
                    payload_deck_name = key
                    if z == k:
                        break
                    else:
                        z += 1
                if k == 0:
                    payload_deck_water_volume = 2/3 * payload_tube_water_volume
                    payload_deck_length = \
                        round(data_dict['geometry_description']['total_cabin_length'] * convert_inch_to_meter, 9)
                    payload_deck_required_galley_power = sum_of_fst * 1000
                else:
                    payload_deck_water_volume = 1/3 * payload_tube_water_volume
                    payload_deck_length = \
                        round(data_dict['geometry_description']['total_cabin_length'] * convert_inch_to_meter * 0.5, 9)
                    payload_deck_required_galley_power = 0
                
                payload_deck_output_dict = {
                    # total section parameter
                    'deck_name_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'name',
                         payload_deck_name],
                    'deck_x_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'position/x',
                         data_dict['positions']['payload_deck_position'][payload_deck_name]['x']],
                    'deck_y_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'position/y',
                         data_dict['positions']['payload_deck_position'][payload_deck_name]['y']],
                    'deck_z_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'position/z',
                         data_dict['positions']['payload_deck_position'][payload_deck_name]['z']],
                    'payload_deck_structural_floor_thickness_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'payload_deck_structural_floor_thickness',
                         round(data_dict['geometry_description']['cabin_floor_thickness'] * convert_inch_to_meter, 9)],
                    'payload_deck_water_volume_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'payload_deck_water_volume',
                         payload_deck_water_volume],
                    'payload_deck_length_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'payload_deck_length',
                         payload_deck_length],
                    'payload_deck_required_galley_power_id' + str(j) + '_id' + str(i) + '_id' + str(k):
                        [payload_deck_path + 'payload_deck_required_galley_power',
                         payload_deck_required_galley_power]
                }
                # merge dictionary to one single output dictionary
                payload_tube_output_dict.update(payload_deck_output_dict)
                k += 1

            # merge dictionary to one single output dictionary
            fuselage_dict.update(payload_tube_output_dict)
            i += 1

            # generate section dictionary for current fuselage
            i = 0
            while i < len(data_dict['dict_fuselage_deign_output']['total_sections']['origin']):
                sections_path = './component_design/fuselage/specific/geometry/fuselage[@ID="' \
                                + str(j) + '"]/sections/section[@ID="' + str(i) + '"]/'
                
                #data_dict['dict_fuselage_deign_output']['total_sections']['origin'][i] = \
                    #coordinate_system_conversion(data_dict['dict_fuselage_deign_output']['total_sections']['origin'][i])
                
                sections_output_dict = {
                    # total section parameter
                    'section_name_id' + str(j) + '_id' + str(i):
                        [sections_path + 'name',
                         'section_' + str(i)],
                    'section_shape_id' + str(j) + '_id' + str(i):
                        [sections_path + 'section_shape',
                         'ellipse'],
                    'section_x_id' + str(j) + '_id' + str(i):
                        [sections_path + 'origin/x',
                         round(data_dict['dict_fuselage_deign_output']['total_sections']['origin'][i]['y'] * -1, 9)],
                    'section_y_id' + str(j) + '_id' + str(i):
                        [sections_path + 'origin/y',
                         round(data_dict['dict_fuselage_deign_output']['total_sections']['origin'][i]['z'], 9)],
                    'section_z_id' + str(j) + '_id' + str(i):
                        [sections_path + 'origin/z',
                         round(data_dict['dict_fuselage_deign_output']['total_sections']['origin'][i]['x'], 9)],
                    'section_upper_height_id' + str(j) + '_id' + str(i):
                        [sections_path + 'upper_height',
                         data_dict['dict_fuselage_deign_output']['total_sections']['upper_height'][i]],
                    'section_lower_height_id' + str(j) + '_id' + str(i):
                        [sections_path + 'lower_height',
                         data_dict['dict_fuselage_deign_output']['total_sections']['lower_height'][i]],
                    'section_width_id' + str(j) + '_id' + str(i):
                        [sections_path + 'width',
                         data_dict['dict_fuselage_deign_output']['total_sections']['width'][i]],
                }
                # merge dictionary to one single output dictionary
                fuselage_dict.update(sections_output_dict)
                i += 1

            # generate fuselage structure mass breakdown dictionary for current fuselage
            i = 0
            for key, value in data_dict['fuselage_design_masses']['fuselage_structural_masses'].items():
                fuselage_structure_path = './component_design/fuselage/specific/geometry/fuselage[@ID="' + str(j) \
                                          + '"]/mass_breakdown/fuselage_structure/component_mass[@ID="' + str(i) \
                                          + '"]/'
                fuselage_structure_name = key
                fuselage_structure_mass = data_dict['fuselage_design_masses']['fuselage_structural_masses'][key]
                fuselage_structure_output_dict = {
                    # fuselage structural mass parameter
                    'fuselage_structure_name_id' + str(j) + '_id' + str(i):
                        [fuselage_structure_path + 'name',
                         fuselage_structure_name],
                    'fuselage_structure_mass_id' + str(j) + '_id' + str(i):
                        [fuselage_structure_path + 'mass',
                         fuselage_structure_mass],
                }
                # merge dictionary to one single output dictionary
                fuselage_dict.update(fuselage_structure_output_dict)
                i += 1

            # generate fuselage furnishing mass breakdown dictionary for current fuselage
            i = 0
            for key, value in data_dict['fuselage_design_masses']['fuselage_furnishing_masses'].items():
                fuselage_furnishing_path = './component_design/fuselage/specific/geometry/fuselage[@ID="' + str(j) \
                                           + '"]/mass_breakdown/fuselage_furnishing/component_mass[@ID="' + str(i) \
                                           + '"]/'
                if key != 'total_furnishing_group_mass':
                    fuselage_furnishing_name = key
                    fuselage_furnishing_mass = data_dict['fuselage_design_masses']['fuselage_furnishing_masses'][key]
                    fuselage_furnishing_output_dict = {
                        # fuselage furnishing mass parameter
                        'fuselage_furnishing_name_id' + str(j) + '_id' + str(i):
                            [fuselage_furnishing_path + 'name',
                             fuselage_furnishing_name],
                        'fuselage_furnishing_mass_id' + str(j) + '_id' + str(i):
                            [fuselage_furnishing_path + 'mass',
                             fuselage_furnishing_mass],
                    }
                    # merge dictionary to one single output dictionary
                    fuselage_dict.update(fuselage_furnishing_output_dict)
                    i += 1

            # generate fuselage operator item mass breakdown dictionary for current fuselage
            i = 0
            for key, value \
                    in data_dict['fuselage_design_masses']['fuselage_operator_item_masses']['group_masses'].items():
                fuselage_operator_item_path = './component_design/fuselage/specific/geometry/fuselage[@ID="' + str(j) \
                                              + '"]/mass_breakdown/fuselage_operator_items/component_mass[@ID="' \
                                              + str(i) + '"]/'
                fuselage_operator_item_name = key
                fuselage_operator_item_mass = \
                    data_dict['fuselage_design_masses']['fuselage_operator_item_masses']['group_masses'][key]
                fuselage_operator_item_output_dict = {
                    # fuselage operator item mass parameter
                    'fuselage_operator_item_name_id' + str(j) + '_id' + str(i):
                        [fuselage_operator_item_path + 'name',
                         fuselage_operator_item_name],
                    'fuselage_operator_item_mass_id' + str(j) + '_id' + str(i):
                        [fuselage_operator_item_path + 'mass',
                         fuselage_operator_item_mass],
                }
                # merge dictionary to one single output dictionary
                fuselage_dict.update(fuselage_operator_item_output_dict)
                i += 1

        key_output_dict.update(fuselage_dict)
        j += 1

    """Method-specific output."""
    method_specific_output_dict = {}
    for key, value in key_output_dict.items():
        if 'mass_properties' not in value[0]:
            method_specific_output_dict[key] = value

    return key_output_dict, method_specific_output_dict


def coordinate_system_conversion(element):
    if len(element.keys()) > 3:
        element_3d = py11csc.Element3D(element["j_xx"], element["j_xy"], element["j_xz"],
                                       element["j_yx"], element["j_yy"], element["j_yz"],
                                       element["j_zx"], element["j_zy"], element["j_zz"])
        element_3d.AC2CGAL()
        for k in element.keys():
            element[k] = getattr(element_3d, k.split("_")[-1])()
        return element
    else:
        element_3d = py11csc.Element3D(element["x"], element["y"], element["z"])
        element_3d.AC2CGAL()
        for k in element.keys():
            element[k] = getattr(element_3d, k)()
        return element
