# imports for python
import math


def new_alpha_factor_from_curve(dict_acn_flex_parameter, runtime_output):
    """
        This function estimates the specific alpha factor for a given number of coverages and landing gear arrangement.
            This function is in accordance to the ICAO COMFAA 3.0 tool from 2013-03-06.

            * Note:
                In December 2006, ICAO accepted revisions to the "alpha factor" multiplier for flexible pavement
                thickness design. The revised alpha factors affect the required pavement thickness and the Aircraft
                Classification Number (ACN) as determined by airplane manufacturers and the COMFAA computer program.
                Until official ICAO documentation incorporates the revisions, users are advised to confirm use of the
                revised '06 alpha factors.
                The FAA will not modify Advisory Circular 150/5320-6D, Airport Pavement Design and Evaluation,
                to incorporate the revised '06 alpha factors as the FAA intends to fully adopt the layered elastic based
                design procedure provided in the FAArfield computer program for flexible pavement design. However,
                the revised alpha factors may be used for pavement evaluations and PCN determination conducted in
                accordance with AC 150/5335-5A."

            * The input dictionary 'dict_acn_flex_parameter' contains all relevant parameter from acn flex method.

            :param: dict_acn_flex_parameter, input dict
            :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
            :return: alpha_factor, output float
        """

    ''' initialize local parameter '''
    a4_define_list = []
    coverages = dict_acn_flex_parameter['coverages']
    number_of_a4_define_list = []
    number_of_wheels = dict_acn_flex_parameter['NW']  # NW

    # set fixed points to define alpha versus log(coverages) curves. log(coverage) values.
    log_coverage_a1 = 1.0
    log_coverage_a4 = 4.0

    # set alphas at log(coverages)
    a0 = 0.1
    a1 = 0.38
    a4 = float()

    # check if the status flag to use pre-defined values for number of wheels and alpha factors from Boeing proposals
    #  in accordance to ICAO code (Alpha.bas line 32ff.)
    if dict_acn_flex_parameter['boeing_proposals_flag']:
        # set incremental damage relative to six-wheels the same as old alphas for higher than 6-wheels
        number_of_a4_define = 13  # NA4define

        # set number of possible wheels per one landing gear strut to list of defined wheels
        number_of_a4_define_list.append(int(1))
        number_of_a4_define_list.append(int(2))
        number_of_a4_define_list.append(int(4))
        number_of_a4_define_list.append(int(6))
        number_of_a4_define_list.append(int(8))
        number_of_a4_define_list.append(int(10))
        number_of_a4_define_list.append(int(12))
        number_of_a4_define_list.append(int(14))
        number_of_a4_define_list.append(int(16))
        number_of_a4_define_list.append(int(18))
        number_of_a4_define_list.append(int(20))
        number_of_a4_define_list.append(int(22))
        number_of_a4_define_list.append(int(24))

        # set defined alpha values to list
        a4_define_list.append(float(0.995))
        a4_define_list.append(float(0.9))
        a4_define_list.append(float(0.8))
        a4_define_list.append(float(0.72))
        a4_define_list.append(float(0.679))
        a4_define_list.append(float(0.65))
        a4_define_list.append(float(0.627))
        a4_define_list.append(float(0.608))
        a4_define_list.append(float(0.591))
        a4_define_list.append(float(0.576))
        a4_define_list.append(float(0.563))
        a4_define_list.append(float(0.551))
        a4_define_list.append(float(0.54))

    # else condition: use pre-defined values from Boeing proposals is not selected -> use standard pre defines from ICAO
    #  in accordance to ICAO code (Alpha.bas line 21ff.)
    else:
        # set incremental damage relative to six-wheels the same as old alphas for higher than 6-wheels
        number_of_a4_define = 8  # NA4define

        # set number of possible wheels per one landing gear strut to list of defined wheels
        number_of_a4_define_list.append(int(1))
        number_of_a4_define_list.append(int(2))
        number_of_a4_define_list.append(int(4))
        number_of_a4_define_list.append(int(6))
        number_of_a4_define_list.append(int(8))
        number_of_a4_define_list.append(int(12))
        number_of_a4_define_list.append(int(18))
        number_of_a4_define_list.append(int(24))

        # set defined alpha values to list
        a4_define_list.append(float(0.995))
        a4_define_list.append(float(0.9))
        a4_define_list.append(float(0.8))
        a4_define_list.append(float(0.72))
        a4_define_list.append(float(0.69))
        a4_define_list.append(float(0.66))
        a4_define_list.append(float(0.64))
        a4_define_list.append(float(0.63))

    #  check if an error is occurred during list generation -> if true: -> terminate function
    if number_of_wheels < number_of_a4_define_list[0]:
        alpha_factor = a4_define_list[0]
        runtime_output.debug('new_alpha_factor_from_curve successfully executed!')
        return alpha_factor

    # check if the given number of coverages is lower than 10
    #  -> if true: -> use regression equation to estimate alpha factor
    if coverages < 10:
        if coverages < 0:
            alpha_factor = 0.15 - 0.23 * math.log10(-coverages)
            runtime_output.debug('new_alpha_factor_from_curve successfully executed!')
            return alpha_factor
        else:
            alpha_factor = 0.15 + 0.23 * math.log10(coverages)
            runtime_output.debug('new_alpha_factor_from_curve successfully executed!')
            return alpha_factor

    # linear interpolation for alpha factors at 10,000 coverages undefined
    for i in range(0, number_of_a4_define-1):
        if number_of_wheels < number_of_a4_define_list[i+1]:
            slope = round((a4_define_list[i + 1] - a4_define_list[i]) /
                          (number_of_a4_define_list[i+1] - number_of_a4_define_list[i]), 10)
            a4 = a4_define_list[i] + slope * (number_of_wheels - number_of_a4_define_list[i])
            break
        else:
            a4 = a4_define_list[number_of_a4_define-1]

    # find the coefficients of the defining exponential association curve
    #  alpha = a * (b - e^(-c * log10(coverages)))
    #   solve for Resid function (below) by Newton's method.
    a = a4 - a0 + 0.1
    delta_a = a / 1000
    resid_m1 = math.log((a0 + a - a1) / a) / log_coverage_a1 - math.log((a0 + a - a4) / a) / log_coverage_a4

    while not abs(resid_m1) < 0.000001:
        a = round(a + delta_a, 10)
        resid = math.log((a0 + a - a1) / a) / log_coverage_a1 - math.log((a0 + a - a4) / a) / log_coverage_a4
        a = a - resid * delta_a / (resid - resid_m1)
        if a < a4 - a0 + 0.001:
            a = a4 - a0 + 0.001
        resid_m1 = math.log((a0 + a - a1) / a) / log_coverage_a1 - math.log((a0 + a - a4) / a) / log_coverage_a4

    b = a0 / a + 1
    c = -math.log((a0 + a - a4) / a) / log_coverage_a4

    runtime_output.debug('new_alpha_factor_from_curve successfully executed!')

    return a * (b - math.exp(-c * math.log10(coverages)))


def alpha_factor_from_curve(dict_acn_flex_parameter, runtime_output):
    """
        This function estimates the specific alpha factor for a given number of coverages and landing gear arrangement.
            This function is in accordance to the ICAO COMFAA 3.0 tool from 2013-03-06.

        * The input dictionary 'dict_acn_flex_parameter' contains all relevant parameter from acn flex method.

        :param: dict_acn_flex_parameter, input dict
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return: alpha_factor, output float
    """

    ''' initialize local parameter '''
    alpha_list = []  # Alpha
    mounted_wheel_list = []  # NWD
    number_of_coverages = dict_acn_flex_parameter['coverages']  # coverages
    number_of_curves = int(7)  # NCurves
    number_of_wheels = dict_acn_flex_parameter['NW']  # NWInterp
    number_of_wheels_max = dict_acn_flex_parameter['NWMax']  # NWMax

    ''' estimate the alpha factor depending on given number of coverages '''
    # call function to generate alpha list according to ICAO code (Alpha.bas line 175)
    alpha_list, mounted_wheel_list = alpha_curves(alpha_list, mounted_wheel_list, number_of_coverages,
                                                  number_of_wheels_max, runtime_output)

    delta_alpha_for_number_of_wheels_1 = (alpha_list[1] - alpha_list[0]) / \
                                         (mounted_wheel_list[1] - mounted_wheel_list[0])

    delta_alpha_for_number_of_wheels_n = (alpha_list[number_of_curves-1] - alpha_list[number_of_curves-2]) / \
                                         (mounted_wheel_list[number_of_curves-1] -
                                          mounted_wheel_list[number_of_curves-2])

    alpha_factor = spline(mounted_wheel_list, alpha_list, number_of_curves, delta_alpha_for_number_of_wheels_1,
                          delta_alpha_for_number_of_wheels_n, number_of_wheels, runtime_output)

    runtime_output.debug('alpha_factor_from_curve successfully executed!')

    return alpha_factor


# sub function to generate alpha list according to ICAO code
def alpha_curves(alpha_list, mounted_wheel_list, number_of_coverages, number_of_wheels_max, runtime_output):
    # set number of possible wheels to mounted wheel list according to ICAO code (Alpha.bas line 122)
    mounted_wheel_list.append(int(1))
    mounted_wheel_list.append(int(2))
    mounted_wheel_list.append(int(4))
    mounted_wheel_list.append(int(6))
    mounted_wheel_list.append(int(12))
    mounted_wheel_list.append(int(16))
    mounted_wheel_list.append(int(number_of_wheels_max))

    # estimate alpha factors depending on the log of coverages
    if number_of_coverages < 1:
        number_of_coverages = 1

    log_of_coverages = math.log10(number_of_coverages)

    # check the value of the log of coverages to estimate the alpha factors depending on
    #  in accordance to ICAO code (Alpha.bas line 128ff.)
    if log_of_coverages < 1:
        alpha_list.append(0.405 + 0.2 * (log_of_coverages - 1))
        alpha_list.append(0.4 + 0.2 * (log_of_coverages - 1))
        alpha_list.append(0.395 + 0.2 * (log_of_coverages - 1))
        alpha_list.append(0.39 + 0.2 * (log_of_coverages - 1))
        alpha_list.append(0.37 + 0.2 * (log_of_coverages - 1))
        alpha_list.append(0.36 + 0.2 * (log_of_coverages - 1))
        alpha_list.append(0.359 + 0.2 * (log_of_coverages - 1))

    elif 1 <= log_of_coverages <= 4:
        alpha_list.append(0.405 + (log_of_coverages - 1) * (0.205 + (log_of_coverages - 2)
                                                            * (-0.0025 - 0.0025 * (log_of_coverages - 3))))
        alpha_list.append(0.4 + (log_of_coverages - 1) * (0.19 + (log_of_coverages - 2)
                                                          * (-0.0125 - 0.000833 * (log_of_coverages - 3))))
        alpha_list.append(0.395 + (log_of_coverages - 1) * (0.18 + (log_of_coverages - 2)
                                                            * (-0.0225 + 0.0041667 * (log_of_coverages - 3))))
        alpha_list.append(0.39 + (log_of_coverages - 1) * (0.175 + (log_of_coverages - 2)
                                                           * (-0.026 + 0.004833 * (log_of_coverages - 3))))
        alpha_list.append(0.37 + (log_of_coverages - 1) * (0.175 + (log_of_coverages - 2)
                                                           * (-0.035 + 0.006667 * (log_of_coverages - 3))))
        alpha_list.append(0.36 + (log_of_coverages - 1) * (0.175 + (log_of_coverages - 2)
                                                           * (-0.035 + 0.005 * (log_of_coverages - 3))))
        alpha_list.append(0.359 + (log_of_coverages - 1) * (0.171 + (log_of_coverages - 2)
                                                            * (-0.0355 + 0.0035 * (log_of_coverages - 3))))

    elif log_of_coverages > 4:
        alpha_list.append(0.99 + 0.17 * (log_of_coverages - 4))
        alpha_list.append(0.89 + 0.125 * (log_of_coverages - 4))
        alpha_list.append(0.825 + 0.0875 * (log_of_coverages - 4))
        alpha_list.append(0.788 + 0.08 * (log_of_coverages - 4))
        alpha_list.append(0.725 + 0.05 * (log_of_coverages - 4))
        alpha_list.append(0.705 + 0.0375 * (log_of_coverages - 4))
        alpha_list.append(0.68 + 0.02 * (log_of_coverages - 4))

    runtime_output.debug('alpha_curves successfully executed!')

    return alpha_list, mounted_wheel_list


# sub function to generate alpha factor from spline curve in accordance to ICAO code (Alpha.bas line 184ff.)
def spline(mounted_wheel_list, alpha_list, number_of_curves, delta_alpha_for_number_of_wheels_1,
           delta_alpha_for_number_of_wheels_n, number_of_wheels, runtime_output):

    # initialize local parameter
    klo = 1
    list_of_u_data_point = []
    list_of_y2_data_points = []

    if delta_alpha_for_number_of_wheels_1 > 9.9E+29:
        list_of_y2_data_points.append(float(0.0))
        list_of_u_data_point.append(float(0.0))
    else:
        list_of_y2_data_points.append(float(-0.5))
        list_of_u_data_point.append(float((3 / (mounted_wheel_list[1] - mounted_wheel_list[0])) *
                                          ((alpha_list[1] - alpha_list[0]) /
                                           (mounted_wheel_list[1] - mounted_wheel_list[0]) -
                                           delta_alpha_for_number_of_wheels_1)))

    for i in range(1, number_of_curves - 1):
        sig = (mounted_wheel_list[i] - mounted_wheel_list[i-1]) / (mounted_wheel_list[i+1] - mounted_wheel_list[i-1])
        p = sig * list_of_y2_data_points[i-1] + 2
        list_of_y2_data_points.append((sig - 1) / p)
        list_of_u_data_point.append((6 * ((alpha_list[i+1] - alpha_list[i]) /
                                          (mounted_wheel_list[i+1] - mounted_wheel_list[i]) -
                                          (alpha_list[i] - alpha_list[i-1]) / (mounted_wheel_list[i] -
                                                                               mounted_wheel_list[i-1])) /
                                     (mounted_wheel_list[i+1] - mounted_wheel_list[i-1]) -
                                     sig * list_of_u_data_point[i-1]) / p)

    if delta_alpha_for_number_of_wheels_n > 9.9E+29:
        qn = 0.0
        un = 0.0
    else:
        qn = 0.5
        un = (3 / (mounted_wheel_list[number_of_curves-1] - mounted_wheel_list[number_of_curves-2])) * \
             (delta_alpha_for_number_of_wheels_n - (alpha_list[number_of_curves-1] - alpha_list[number_of_curves-2]) /
              (mounted_wheel_list[number_of_curves-1] - mounted_wheel_list[number_of_curves-2]))

    list_of_y2_data_points.append((un - qn * list_of_u_data_point[-1]) / (qn * list_of_y2_data_points[-1] + 1))

    for k in range(number_of_curves - 2, -1, -1):
        list_of_y2_data_points[k] = list_of_y2_data_points[k] * list_of_y2_data_points[k+1] + list_of_u_data_point[k]

    while number_of_curves - klo > 1:
        index_k = int((number_of_curves + klo) / 2)
        if mounted_wheel_list[index_k-1] > number_of_wheels:
            number_of_curves = index_k
        else:
            klo = index_k

    h = int(mounted_wheel_list[number_of_curves-1] - mounted_wheel_list[klo-1])
    a = int((mounted_wheel_list[number_of_curves-1] - number_of_wheels) / h)
    b = int((number_of_wheels - mounted_wheel_list[klo-1]) / h)

    runtime_output.debug('spline successfully executed!')

    return a * alpha_list[klo-1] + \
        b * alpha_list[number_of_curves-1] + \
        ((a**3 - a) * list_of_y2_data_points[klo-1] +
         (b**3 - b) * list_of_y2_data_points[number_of_curves-1]) * (h * h) / 6
