def estimate_cog(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ Hier beschreibung hin: Estimation of undercarriage center of gravity.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    ''' initialize local parameter '''
    convert_inch_to_meter = 0.0254
    nose_gear_cog = {'nose_gear_cog_global': {'x': float(), 'y': float(), 'z': float()},
                     'nose_gear_strut_cog_local': {'x': float(), 'y': float(), 'z': float()},
                     'tire_positions_per_strut': {'x': [], 'y': [], 'z': []}
                     }
    main_gear_cog = {'main_gear_cog_global': {'x': float(), 'y': float(), 'z': float()},
                     'main_gear_strut_cog_local': {'x': {'outer_strut': float(), 'inner_strut': None},
                                                   'y': {'outer_strut': float(), 'inner_strut': None},
                                                   'z': {'outer_strut': float(), 'inner_strut': None}
                                                   },
                     'tire_positions_per_outer_strut': {'x': [], 'y': [], 'z': []},
                     'tire_positions_per_inner_strut': {'x': [], 'y': [], 'z': []}
                     }
    total_landing_gear_cog = {'x': float(), 'y': float(), 'z': float()}

    ''' estimate nose gear center of gravity '''
    # estimate global cog for nose gear assembly
    #  global cog in x-direction
    nose_gear_cog['nose_gear_cog_global']['x'] = \
        (landing_gear_description['landing_gear_positions']['nose_gear_x_position']
         - (0.5 * landing_gear_description['nose_gear_description']['nose_gear_arrangement']['nose_gear_strut_length']))

    #  global cog in y-direction
    nose_gear_cog['nose_gear_cog_global']['y'] = 0.0

    #  global cog in z-direction
    fuselage_reference_point_in_z_direction_key, fuselage_reference_point_in_z_direction_value = \
        next(iter(dict_ac_exchange['fuselage_reference_point_in_z_direction'].items()), (None, None))
    if fuselage_reference_point_in_z_direction_value is not None:
        nose_gear_cog['nose_gear_cog_global']['z'] = \
            round(fuselage_reference_point_in_z_direction_value
                  - landing_gear_description['fuselage_parameter']['fuselage_height'] / 2
                  + 0.6, 4)
    else:
        nose_gear_cog['nose_gear_cog_global']['z'] = \
            round(0.0
                  - landing_gear_description['fuselage_parameter']['fuselage_height'] / 2
                  + 0.6, 4)

    ''' estimate main gear center of gravity '''
    # estimate global cog for main gear assembly
    #  global cog in x-direction
    main_gear_cog['main_gear_cog_global']['x'] =\
        landing_gear_description['landing_gear_positions']['main_gear_x_position']
    if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_struts'] > 2:
        main_gear_cog['main_gear_cog_global']['x'] = \
            round(main_gear_cog['main_gear_cog_global']['x']
                  + (0.5 * landing_gear_description['main_gear_description']['main_gear_arrangement']
                                                   ['main_gear_inner_strut_length']), 4)

    #  global cog in y-direction
    main_gear_cog['main_gear_cog_global']['y'] = 0.0

    #  global cog in z-direction
    main_gear_cog['main_gear_cog_global']['z'] = \
        round(-landing_gear_description['fuselage_parameter']['fuselage_height'] / 2
              + abs(landing_gear_description['main_gear_description']['main_gear_arrangement']
                                            ['main_landing_gear_wing_mounting_position_in_z_direction'])/2, 4)

    ''' estimate total landing gear center of gravity '''
    # total cog in x-direction
    total_landing_gear_cog['x'] = \
        round(((nose_gear_cog['nose_gear_cog_global']['x']
               * landing_gear_description['landing_gear_masses']['nose_landing_gear_masses']['nose_landing_gear_mass'])
               + (main_gear_cog['main_gear_cog_global']['x']
               * landing_gear_description['landing_gear_masses']['main_landing_gear_masses']['main_landing_gear_mass']))
              / (landing_gear_description['landing_gear_masses']['nose_landing_gear_masses']['nose_landing_gear_mass']
                 + landing_gear_description['landing_gear_masses']['main_landing_gear_masses']['main_landing_gear_mass']
                 ), 4)

    # total cog in y-direction
    total_landing_gear_cog['y'] = \
        round(((nose_gear_cog['nose_gear_cog_global']['y']
                * landing_gear_description['landing_gear_masses']['nose_landing_gear_masses']['nose_landing_gear_mass'])
               + (main_gear_cog['main_gear_cog_global']['y']
                  * landing_gear_description['landing_gear_masses']['main_landing_gear_masses'][
                      'main_landing_gear_mass']))
              / (landing_gear_description['landing_gear_masses']['nose_landing_gear_masses']['nose_landing_gear_mass']
                 + landing_gear_description['landing_gear_masses']['main_landing_gear_masses']['main_landing_gear_mass']
                 ), 4)

    # total cog in z-direction
    total_landing_gear_cog['z'] = \
        round(((nose_gear_cog['nose_gear_cog_global']['z']
                * landing_gear_description['landing_gear_masses']['nose_landing_gear_masses']['nose_landing_gear_mass'])
               + (main_gear_cog['main_gear_cog_global']['z']
                  * landing_gear_description['landing_gear_masses']['main_landing_gear_masses'][
                      'main_landing_gear_mass']))
              / (landing_gear_description['landing_gear_masses']['nose_landing_gear_masses']['nose_landing_gear_mass']
                 + landing_gear_description['landing_gear_masses']['main_landing_gear_masses']['main_landing_gear_mass']
                 ), 4)

    ''' estimate local cog for each nose gear strut '''
    # local cog in x-direction
    if total_landing_gear_cog['x'] > nose_gear_cog['nose_gear_cog_global']['x']:
        nose_gear_cog['nose_gear_strut_cog_local']['x'] = round(nose_gear_cog['nose_gear_cog_global']['x']
                                                                - total_landing_gear_cog['x'], 4)
    else:
        nose_gear_cog['nose_gear_strut_cog_local']['x'] = round(total_landing_gear_cog['x']
                                                                - nose_gear_cog['nose_gear_cog_global']['x'], 4)

    # local cog in y-direction
    if landing_gear_description['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_struts'] > 1:
        nose_gear_cog['nose_gear_strut_cog_local']['y'] = \
            round(landing_gear_description['landing_gear_positions']['nose_gear_strut_y_position']
                  - nose_gear_cog['nose_gear_cog_global']['y'], 4)
    else:
        nose_gear_cog['nose_gear_strut_cog_local']['y'] = nose_gear_cog['nose_gear_cog_global']['y']

    # local cog in z-direction
    if total_landing_gear_cog['z'] < nose_gear_cog['nose_gear_cog_global']['z']:
        nose_gear_cog['nose_gear_strut_cog_local']['z'] = round(nose_gear_cog['nose_gear_cog_global']['z']
                                                                - total_landing_gear_cog['z'], 4)
    else:
        nose_gear_cog['nose_gear_strut_cog_local']['z'] = round(total_landing_gear_cog['z']
                                                                - nose_gear_cog['nose_gear_cog_global']['z'], 4)

    # tire positions of nose gear strut
    if landing_gear_description['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_tires_per_strut'] <= 2:  # noPep8 e501
        nose_gear_cog['tire_positions_per_strut']['x'].append(0.0)
        nose_gear_cog['tire_positions_per_strut']['y'].append(
            round(nose_gear_cog['nose_gear_strut_cog_local']['y'] -
                  float(landing_gear_description['nose_gear_description']['nose_gear_tires'][
                      next(iter(landing_gear_description['nose_gear_description']['nose_gear_tires']))][
                      'max_section_width_in_inch']) * convert_inch_to_meter * 1.5, 4))
        nose_gear_cog['tire_positions_per_strut']['z'].append(
            -landing_gear_description['nose_gear_description']['nose_gear_arrangement']['nose_gear_strut_length'])
        if landing_gear_description['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_tires_per_strut'] == 2:  # noPep8 e501
            nose_gear_cog['tire_positions_per_strut']['x'].append(nose_gear_cog['tire_positions_per_strut']['x'][0])
            nose_gear_cog['tire_positions_per_strut']['y'].append(-nose_gear_cog['tire_positions_per_strut']['y'][0])
            nose_gear_cog['tire_positions_per_strut']['z'].append(nose_gear_cog['tire_positions_per_strut']['z'][0])
    elif landing_gear_description['nose_gear_description']['nose_gear_arrangement']['number_of_nose_gear_tires_per_strut'] == 4:  # noPep8 e501
        nose_gear_cog['tire_positions_per_strut']['x'] = [0.0 for _ in range(4)]
        nose_gear_cog['tire_positions_per_strut']['y'].append(
            round(nose_gear_cog['nose_gear_strut_cog_local']['y'] -
                  float(landing_gear_description['nose_gear_description']['nose_gear_tires'][
                            next(iter(landing_gear_description['nose_gear_description']['nose_gear_tires']))][
                            'max_section_width_in_inch']) * convert_inch_to_meter * 1.5, 4))
        nose_gear_cog['tire_positions_per_strut']['y'].append(-nose_gear_cog['tire_positions_per_strut']['y'][0])
        nose_gear_cog['tire_positions_per_strut']['y'].append(nose_gear_cog['tire_positions_per_strut']['y'][0]
                                                              + 2 * nose_gear_cog['tire_positions_per_strut']['y'][0])
        nose_gear_cog['tire_positions_per_strut']['y'].append(-nose_gear_cog['tire_positions_per_strut']['y'][2])
        nose_gear_cog['tire_positions_per_strut']['z'] = [
            -landing_gear_description['nose_gear_description']['nose_gear_arrangement']['nose_gear_strut_length']
            for _ in range(4)]

    # tire positions of main gear outer struts
    z_pos = (-landing_gear_description['main_gear_description']['main_gear_arrangement'][
            'main_gear_outer_strut_length'])
    if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_tires_per_outer_strut'] <= 2:  # noPep8 e501
        main_gear_cog['tire_positions_per_outer_strut']['x'].append(0.0)
        main_gear_cog['tire_positions_per_outer_strut']['z'].append(z_pos)
        if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_tires_per_outer_strut'] == 2:  # noPep8 e501
            main_gear_cog['tire_positions_per_outer_strut']['x'].append(0.0)
            main_gear_cog['tire_positions_per_outer_strut']['z'].append(z_pos)
    elif landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_tires_per_outer_strut'] == 4:  # noPep8 e501
        main_gear_cog['tire_positions_per_outer_strut']['x'].append(
            [x_pos - landing_gear_description['main_gear_description'][
                'main_gear_arrangement']['local_tire_position_in_x_for_acn'][2]/2 for x_pos in
             landing_gear_description['main_gear_description'][
                'main_gear_arrangement']['local_tire_position_in_x_for_acn']])
        main_gear_cog['tire_positions_per_outer_strut']['z'].append([z_pos for _ in range(4)])

    elif landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_tires_per_outer_strut'] == 6:  # noPep8 e501
        main_gear_cog['tire_positions_per_outer_strut']['x'].append(
            [x_pos - landing_gear_description['main_gear_description'][
                'main_gear_arrangement']['local_tire_position_in_x_for_acn'][4] / 2 for x_pos in
             landing_gear_description['main_gear_description'][
                 'main_gear_arrangement']['local_tire_position_in_x_for_acn']])
        main_gear_cog['tire_positions_per_outer_strut']['z'].append([z_pos for _ in range(6)])

    main_gear_cog['tire_positions_per_outer_strut']['y'].append(
        landing_gear_description['main_gear_description']['main_gear_arrangement']['local_tire_position_in_y_for_acn'])

    new_dict = {}
    for key, value in main_gear_cog['tire_positions_per_outer_strut'].items():
        if isinstance(value, list):
            if len(value) == 1 and isinstance(value[0], list):
                new_dict[key] = value[0]
            else:
                new_dict[key] = value
        else:
            new_dict[key] = [value]

    main_gear_cog['tire_positions_per_outer_strut'] = new_dict
    del new_dict

    # tire positions of main gear inner struts
    if 'number_of_main_gear_inner_struts' in landing_gear_description['main_gear_description']['main_gear_arrangement']:
        z_pos = (-landing_gear_description['main_gear_description']['main_gear_arrangement'][
            'main_gear_inner_strut_length'])
        main_gear_cog['tire_positions_per_inner_strut']['y'].append(
            landing_gear_description['main_gear_description']['main_gear_arrangement'][
                'local_tire_position_in_y_for_acn'][0])
        main_gear_cog['tire_positions_per_inner_strut']['y'].append(
            -main_gear_cog['tire_positions_per_inner_strut']['y'][0])
        main_gear_cog['tire_positions_per_inner_strut']['z'].append([z_pos for _ in range(2)])
        if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_tires_per_inner_strut'] == 2:  # noPep8 e501
            main_gear_cog['tire_positions_per_inner_strut']['x'].append([0.0 for _ in range(2)])
        elif landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_inner_struts'] == 4:  # noPep8 e501
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                -landing_gear_description['main_gear_description']['main_gear_arrangement']['wheel_base'])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                main_gear_cog['tire_positions_per_inner_strut']['x'][0])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                landing_gear_description['main_gear_description']['main_gear_arrangement']['wheel_base'])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                main_gear_cog['tire_positions_per_inner_strut']['x'][2])
            main_gear_cog['tire_positions_per_inner_strut']['y'].append(
                main_gear_cog['tire_positions_per_inner_strut']['y'])
            main_gear_cog['tire_positions_per_inner_strut']['z'].append(
                main_gear_cog['tire_positions_per_inner_strut']['z'])
        elif landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_inner_struts'] == 6:  # noPep8 e501
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                -landing_gear_description['main_gear_description']['main_gear_arrangement']['wheel_base'])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                main_gear_cog['tire_positions_per_inner_strut']['x'][0])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append([0.0 for _ in range(2)])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                landing_gear_description['main_gear_description']['main_gear_arrangement']['wheel_base'])
            main_gear_cog['tire_positions_per_inner_strut']['x'].append(
                main_gear_cog['tire_positions_per_inner_strut']['x'][4])
            main_gear_cog['tire_positions_per_inner_strut']['y'].append(
                [main_gear_cog['tire_positions_per_inner_strut']['y'] for _ in range(4)])
            main_gear_cog['tire_positions_per_inner_strut']['z'].append(
                [main_gear_cog['tire_positions_per_inner_strut']['z'] for _ in range(4)])

        new_dict = {}
        for key, value in main_gear_cog['tire_positions_per_inner_strut'].items():
            if isinstance(value, list):
                if len(value) == 1 and isinstance(value[0], list):
                    new_dict[key] = value[0]
                else:
                    new_dict[key] = value
            else:
                new_dict[key] = [value]

        main_gear_cog['tire_positions_per_inner_strut'] = new_dict
        del new_dict

    ''' estimate main landing gear strut local position '''
    # local cog in x-direction
    delta_x_outer_strut = round(- 0.5 * landing_gear_description['main_gear_description']['main_gear_arrangement']
                                                                ['main_gear_outer_strut_length'], 4)
    if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_struts'] > 2:
        delta_x_inner_strut = round(0.5 * landing_gear_description['main_gear_description']['main_gear_arrangement']
                                                                  ['main_gear_inner_strut_length'], 4)

        if total_landing_gear_cog['x'] > main_gear_cog['main_gear_cog_global']['x']:
            main_gear_cog['main_gear_strut_cog_local']['x']['outer_strut'] = \
                round(total_landing_gear_cog['x'] - main_gear_cog['main_gear_cog_global']['x']
                      - abs(delta_x_outer_strut), 4)
            main_gear_cog['main_gear_strut_cog_local']['x']['inner_strut'] = \
                round(total_landing_gear_cog['x'] - main_gear_cog['main_gear_cog_global']['x']
                      - abs(delta_x_inner_strut), 4)
        else:
            main_gear_cog['main_gear_strut_cog_local']['x']['outer_strut'] = \
                round(main_gear_cog['main_gear_cog_global']['x'] - total_landing_gear_cog['x']
                      - abs(delta_x_outer_strut), 4)
            main_gear_cog['main_gear_strut_cog_local']['x']['inner_strut'] = \
                round(main_gear_cog['main_gear_cog_global']['x'] - total_landing_gear_cog['x']
                      - abs(delta_x_inner_strut), 4)
    else:
        if total_landing_gear_cog['x'] > main_gear_cog['main_gear_cog_global']['x']:
            main_gear_cog['main_gear_strut_cog_local']['x']['outer_strut'] = \
                round(total_landing_gear_cog['x'] - main_gear_cog['main_gear_cog_global']['x']
                      - abs(delta_x_outer_strut), 4)
        else:
            main_gear_cog['main_gear_strut_cog_local']['x']['outer_strut'] = \
                round(main_gear_cog['main_gear_cog_global']['x'] - total_landing_gear_cog['x']
                      - abs(delta_x_outer_strut), 4)
        main_gear_cog['main_gear_strut_cog_local']['x']['inner_strut'] = None

    # local cog in y-direction
    main_gear_cog['main_gear_strut_cog_local']['y']['outer_strut'] = \
        round(main_gear_cog['main_gear_cog_global']['y']
              + landing_gear_description['landing_gear_positions']['main_gear_outer_strut_y_position'], 4)

    if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_struts'] > 2:
        main_gear_cog['main_gear_strut_cog_local']['y']['inner_strut'] = \
            round(main_gear_cog['main_gear_cog_global']['y']
                  + landing_gear_description['landing_gear_positions']['main_gear_inner_strut_y_position'], 4)
    else:
        main_gear_cog['main_gear_strut_cog_local']['y']['inner_strut'] = None

    # local cog in z-direction
    delta_z_outer_strut = abs(round(landing_gear_description['main_gear_description']['main_gear_arrangement']
                                                            ['main_landing_gear_wing_mounting_position_in_z_direction']
                                    / 4, 4))
    if total_landing_gear_cog['z'] < main_gear_cog['main_gear_cog_global']['z']:
        main_gear_cog['main_gear_strut_cog_local']['z']['outer_strut'] = \
            round(main_gear_cog['main_gear_cog_global']['z'] - total_landing_gear_cog['z'] + delta_z_outer_strut,
                  4)
    else:
        main_gear_cog['main_gear_strut_cog_local']['z']['outer_strut'] = \
            round(total_landing_gear_cog['z'] - main_gear_cog['main_gear_cog_global']['z'] + delta_z_outer_strut,
                  4)

    if landing_gear_description['main_gear_description']['main_gear_arrangement']['number_of_main_gear_struts'] > 2:
        delta_z_inner_strut = \
            abs(round(-1/4 * landing_gear_description['main_gear_description']['main_gear_arrangement']
                                                     ['main_gear_inner_strut_length'], 4))
        if total_landing_gear_cog['z'] < main_gear_cog['main_gear_cog_global']['z']:
            main_gear_cog['main_gear_strut_cog_local']['z']['inner_strut'] = \
                round(main_gear_cog['main_gear_cog_global']['z'] - total_landing_gear_cog['z']
                      - abs(delta_z_inner_strut), 4)
        else:
            main_gear_cog['main_gear_strut_cog_local']['z']['inner_strut'] = \
                round(total_landing_gear_cog['z'] - main_gear_cog['main_gear_cog_global']['z']
                      - abs(delta_z_inner_strut), 4)
    else:
        main_gear_cog['main_gear_strut_cog_local']['z']['inner_strut'] = None

    # save cog results in the landing gear tire dictionary
    landing_gear_cog = {'nose_gear_cog': nose_gear_cog, 'main_gear_cog': main_gear_cog,
                        'total_landing_gear_cog': total_landing_gear_cog}
    landing_gear_description['landing_gear_cog'] = landing_gear_cog
    landing_gear_description['landing_gear_positions']['nose_gear_z_position'] = \
        nose_gear_cog['nose_gear_cog_global']['z']

    runtime_output.debug('estimate_cog successfully executed!')

    return landing_gear_description
