# Import standard modules.
import os
import sys
import math
import importlib

# Import own modules.
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions.convergencecheck import convergence_check  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.plot_functions.generateacnplots import generate_acn_plots  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions.prepareexternaldata import prepare_external_data  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.plot_functions.generatelandinggearplots import generate_landing_gear_plots  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.aicraft_classification_number.estimateaircraftclassificationnumber import estimate_aircraft_classification_number  # noPep8 e501


def landing_gear_design(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """  Module providing calculation functions for landing gear design in accordance CS 25.733 and EASA ETSO-Tire-List
         based on the landing gear load estimation of N. Currey.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """

    ''' Read and prepare external data for fuselage and cabin design '''
    # Call function to prepare external data.
    dict_ac_exchange = prepare_external_data(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config,
                                             runtime_output)

    ''' Initialize local parameter '''
    word_to_insert_after = 'estimate'
    files_to_remove = ['prepareexternaldata.py', 'convergencecheck.py']
    landing_gear_description = {'working_directory': paths_and_names['working_directory']}

    ''' Run landing gear design '''
    # If the script is compiled, sys._MEIPASS gives the temp directory for PyInstaller.
    if getattr(sys, 'frozen', False):    
        # List all entries in the directory.     
        all_entries = os.listdir(sys._MEIPASS + '/' +
                                 routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
        # Filter out only files (ignore directories)
        call_function_list = [entry for entry in all_entries
                            if os.path.isfile(sys._MEIPASS + '/' + routing_dict['module_import_name'].replace('.', '/')
                                              + '/general/call_functions' + '/' + entry)]
    # If running as a regular Python script, use the current directory.
    else:
        # List all entries in the directory.
        all_entries = os.listdir(paths_and_names['working_directory'] + '/' +
                                routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
        # Filter out only files (ignore directories).
        call_function_list = [entry for entry in all_entries
                            if os.path.isfile(paths_and_names['working_directory'] + '/' +
                                                routing_dict['module_import_name'].replace('.', '/')
                                                + '/general/call_functions' + '/' + entry)]

    # Remove unnecessary files from call function list.
    for file in files_to_remove:
        if file in call_function_list:
            call_function_list.remove(file)

    call_function_list = sorted(call_function_list, key=lambda x: (int(x.split('_')[1]), x))

    # initialize convergence criteria
    convergence_flag = False
    convergence_parameter = {'convergence_criteria': 0.001,
                             'nose_gear_position_in_x': math.inf,
                             'nose_gear_strut_length': math.inf,
                             'main_gear_position_in_x': math.inf,
                             'main_gear_strut_length': math.inf,
                             'main_gear_outer_strut_y_position': math.inf,
                             }
    
    i = 1
    while not convergence_flag and i < 100:
        runtime_output.print('Current run of landing gear iteration is: ' + str(i))
        # Loop across all python files from call_function directory to execute them one by one.
        for python_module in call_function_list:
            try:
                # Import own modules from call function directory.
                import_call_function = (routing_dict['module_import_name']
                                        + '.general.call_functions.' + os.path.splitext(python_module)[0])

                if 'estimate' in import_call_function:
                    position_to_insert = python_module.find(word_to_insert_after) + len(word_to_insert_after)
                    python_module = word_to_insert_after + '_' + python_module[position_to_insert:]

                    function_name = os.path.splitext(python_module)[0]
                    # Import calculation module.
                    module = importlib.import_module(import_call_function)

                    # Call function depending on routing parameter.
                    landing_gear_description = getattr(module, function_name)(dict_ac_exchange, dict_mod_config,
                                                                              landing_gear_description, runtime_output)

            # Exception handling for module import error.
            except ModuleNotFoundError as module_import_error:
                runtime_output.critical('Error: ' + str(module_import_error) + ' found in '
                                        + routing_dict['module_name'] + '. Program aborted.')
                sys.exit('Exit information: The own python module to be imported could not be found!')

        ''' convergence check '''
        convergence_parameter, convergence_flag = convergence_check(landing_gear_description, convergence_parameter,
                                                                    runtime_output, convergence_flag, i)
        i += 1

    if not convergence_flag:
        runtime_output.critical('Error: Landing gear design not successful after 100 iterations. '
                                'A valid solution cannot be found.\n '
                                'Program aborted.')
        sys.exit('Landing gear design iteration failed!')

    ''' estimate aircraft classification number (acn) '''
    # initialize parameter to estimate ACN for flex pavement
    dict_acn_flex_parameter, dict_constants = \
        estimate_aircraft_classification_number(paths_and_names, landing_gear_description, dict_ac_exchange,
                                                dict_mod_config, runtime_output)

    # visualisation of results
    dict_landing_gear_output = generate_landing_gear_plots(dict_ac_exchange, landing_gear_description,
                                                           runtime_output)
    dict_acn_flex_parameter = generate_acn_plots(dict_acn_flex_parameter, dict_constants, runtime_output)

    landing_gear_description['acn_estimation'] = dict_acn_flex_parameter
    landing_gear_description['dict_landing_gear_output'] = dict_landing_gear_output

    # Check if the given tool_level from aircrat exchange file is None -> if true: -> initialize with 0.
    if dict_ac_exchange['tool_level'] is None:
        landing_gear_description['current_tool_level'] = 0
    # Else if condition: Check if given tool_level from aircrat exchange is lower than the given own tool level 
    #  from module configuration file -> if true: Add 1 to current tool level from airaft exchange file.
    elif dict_ac_exchange['tool_level'] < int(routing_dict['tool_level']):
        landing_gear_description['current_tool_level'] = dict_ac_exchange['tool_level'] + 1
    # Else condition: The given tool_level from aircrat exchange is equal to the given own tool level
    else: 
        landing_gear_description['current_tool_level'] = int(routing_dict['tool_level'])
    
    # Debug print for function landing_gear_design.        
    runtime_output.debug('landing_gear_design successfully executed!')

    return landing_gear_description
