/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef performance_assessment_SRC_TAW_DEFAULTMETHODS_AIRCRAFT_H_
#define performance_assessment_SRC_TAW_DEFAULTMETHODS_AIRCRAFT_H_

#include <string>
#include <memory>
#include <vector>

#include <moduleBasics/module.h>
#include <aerodynamics/aerodynamics.h>
#include <atmosphere/atmosphere.h>
#include <engine/engine.h>
#include "missionDefinition.h"
#include "../../libs/multi_engine_propulsion.h"

class performance_assessment_settings;

using std::string;

/** \class Aircraft class that contains all properties of the calculated aircraft objects.
*
*/

class aircraft {
 public:
    const std::shared_ptr<RuntimeIO>& rtIO;
    const performance_assessment_settings *mySettingsPt; /**< pointer to the mySettings object */
    node& polarXML; /**< Address of aixml object for the polar file */
    
    string acftName;/**<Name of the aircraft*/

    /** \brief Function to get the mission type
    * \return Type of mission that should be analyzed (DesignMission or StudyMission)(-)
    */
    string getMissionType();

    double banana;
    double MTOM;/**< Maximum Take-off Mass (kg)*/
    double MLM; /**< Maximum Landing Mass (kg)*/
    double OME;/**< Operating Mass Empty(kg) */
    double MFM;/**< Maximum Fuel mass (kg) */
    double maxPayload; /** Maximum payload (kg) */

    double MMO; /**< Maximum operating Mach number (-) */
    double VMO; /**< Maximum operating speed (KCAS) */
    double MDive; /** Diving Mach number (-) */
    double VDive; /** Diving speed (KCAS) */

    double MInitialCruise; /**< Initial cruise Mach number (-) */
    double hInitialCruise; /**< Initial cruise altitude (ft) */
    double hMaxOperating; /**< Maximum operating altitude (ft) */

    double timeToClimb; /**< Design time to climb (from 1500 ft to initial cruise altitude at ISA+deltaISA (s)*/

    double delta_isa; /**< Temperature delta to the ISA-Conditions (K)*/

    double designRange;     /**< Design range (km) */
    double designPayload;   /**< Design payload (kg)*/

    double rangeMaxPayload; /**< Range with maximum payload and fuel till the aircraft reaches MTOM (km) */
    double rangeMaxFuel;    /**< Range with maximum fuel and payload till the aircraft reaches MTOM (km) */
    double payloadRangeMaxfuel; /** Payload with maximum fuel and MTOM (kg) */
    double rangeFerry;      /** Range with maximum fuel and zero payload (km) */

    int numberEngines; /**< Number of engines (-) */

    /** \class Class that contains the required energy carrier IDs for different flight phases
    *
    */
    class energyProvider {
     public:
        int energyCarrierID_ground;
        int energyCarrierID_taxi;
        int energyCarrierID_takeoff;
        int energyCarrierID_climb;
        int energyCarrierID_cruise;
        int energyCarrierID_descent;
        int energyCarrierID_landing;
    };
    energyProvider myEnergyProviders;

    double S_Wing; /**< Wing reference area (m2) */
    double optCL; /**< Aircraft lift coefficient at cruise Mach number and highest lift to drag ratio (-) */

    /* Landing Gear Properties */
    double frictionCoefficient; /**< Friction coefficient between tyres and pavement (-)*/
    double brakingCoefficient; /**< Braking coefficient (deceleration due to braking) (-)*/

    /** \class Requirements Class that contains the required performance characteristics
    *
    */
    class Requirements {
     public:
        /** \brief Function sets the minimum required climb gradients according to user specifications in Aircraft Requirements (compared to CS-25 requirements)
        *   \param climbGradientRequirement const node: constant reference to aircraft node of Requirement/ClimbGradients [-]
        *   \param numberOfEngines const uint16_t: constant reference to aircraft numberEngines [-]
        */
        void setRequiredClimbGradients(const node& climbGradientRequirement, const uint16_t& numberOfEngines);
        // Member variables
        double OEICruiseAltitude;               /**< Maximum operating altitude with one engine inoperative (ft) */
        double TOFL;                            /**< Take-off field length (m) */
        double climbGradientSecondTOSegment;    /**< Required climb gradient in the second climb segment (%) */
        double climbGradientFinalTOSegment;     /**< Required climb gradient in the third (final) climb segment (%) */
        double LDN;                             /**< Design landing field length (m) */
        double climbGradientApproachOEI;        /**< Required climb gradient for go-around during approach and one engine inoperative(%) */
        double climbGradientApproachAEO;        /**< Required climb gradient for go-around during approach (%) */
        double approachSpeed;                   /**< Design approach speed (kts) */
    };
    Requirements myReqs;

    /** \class requirementChecks Class that proves if the requirements are met
    *
    */
    class requirementChecks {
     public:
        bool TOFL; /**< Switch if take-off field length can be maintained (-) */
        bool LDN; /**< Switch if landing field length can be maintained (-) */
        bool vApproach; /**< Switch if approach speed can be maintained (-) */
        bool initCruiseAlt; /**< Switch if landing field length can be maintained (-) */
        bool maxOptAlt; /**< Switch if landing field length can be maintained (-) */
        bool OEICruiseAlt; /**< Switch if landing field length can be maintained (-) */
        bool initCruiseM; /**< Switch if landing field length can be maintained (-) */
        bool climbGradientSecondSegment; /**< Switch if landing field length can be maintained (-) */
        bool climbGradientFinalSegment; /**< Switch if landing field length can be maintained (-) */
        bool climbGradientApproachOEI; /**< Switch if landing field length can be maintained (-) */
        bool climbGradientApproachAEO; /**< Switch if landing field length can be maintained (-) */
    };
    requirementChecks myChecks;

    // Engine myEngine; /**< Engine object */
    aerodynamics aero; /**< Aerodynamics object */
    missionDefinition myMission; /**< MissionDefinition object */
    atmosphere atm; /**<Atmosphere object */
    Multi_engine_propulsion myEngines;

    /** \brief Function to calculate the reference stall speed (1g stalling speed) for a given maximum lift coefficient
    * \param aircraftMass const double: Constant reference to the aircraft mass (kg)
    * \param flightAltitude const double: Constant reference to a reference/flight altitude (m)
    * \param maximumLiftCoefficient const double: Constant reference to a maximum lift coefficient (-)
    * \return reference stalling speed (1g stall speed) (m/s)
    */
    double getReferenceStallSpeed(const double& aircraftMass, const double& flightAltitude, const double& maximumLiftCoefficient);

    /** \brief Function to calculate the 1g stalling speed in clean configuration
    * \details here it is the incompressible stall speed calculated with the CLmax value of the incompressible polar (Mach = 0.)
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return 1g stalling speed in clean configuration (m/s)
    */
    double v1gS_Clean(double Mass, double Altitude);

    /** \brief Function to calculate the 1g stalling speed in take-off configuration
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return 1g stalling speed in take-off configuration (m/s)
    */
    double v1gS_TO(double Mass, double Altitude);

    /** \brief Function to calculate the lift off speed in take-off configuration
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return lift off speed in take-off configuration (m/s)
    */
    double vLOF(double Mass, double Altitude);

    /** \brief Function to calculate take-off safety speed (v2) in take-off configuration
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return take-off safety speed (v2) in take-off configuration (m/s)
    */
    double v2(double Mass, double Altitude);

    /** \brief Function to calculate the final take-off speed in clean configuration
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return take-off speed in clean configuration (m/s)
    */
    double vFTO(double Mass, double Altitude);

    /** \brief Function to calculate the approach speed in landing configuration
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return approach speed in landing configuration (m/s)
    */
    double vAPPR(double Mass, double Altitude);


    /** \brief Function to calculate the rate of climb
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \param Mach Mach number (-)
    * \param fACC Acceleration factor (Scheiderer, 2008: Angewandte Flugleistung - Eine Einfuehrung in die operationelle Flugleistung vom Start bis zur Landung; p.232 table 11.1)
    * \param Configuration Aerodynamic configuration (-)
    * \param Rating Engine rating (-)
    * \param bleedOff Current bleed air offtake (kg/s)
    * \param shaftOff Current shaft power offtake (kW)
    * \param OEI Switch if one engine is inoperative (-)
    * \param thisAERO object of the aerodynamics class
    * \return rate of climb (m/s)
    */
    double ROC(double Mass, double Alt, double Mach, double fACC,
               string Configuration, string Rating, double bleedOff, double shaftOff, bool OEI,
               const aerodynamics& thisAERO);

    /** \brief Function to calculate the wind milling drag
    * \param Mass Aircraft mass (kg)
    * \param Altitude Aircraft altitude (m)
    * \return wind milling drag (N)
    */
    double calcWindMillingDrag(double Mach, double Altitude);

    /** \brief Constructor for an aircraft object
    * \param mySettings Address to the performance_assessment_settings object
    */
    explicit aircraft(const std::shared_ptr<RuntimeIO>& rtIO, const performance_assessment_settings &mySettings);

    /** \brief Destructor for an aircraft object
    */
    virtual ~aircraft();

 private:
    double engineInletD; /**< Width of the engine (m) */

    double factorVLOF; /**< Factor to calculate the take-off safety speed (-) */
    double factorV2; /**< Factor to calculate the final take-off speed (-) */

    /** \brief Function to determine the factor for the final take-off speed
    *   \details Calculate minimum multiple of v_SR (reference stall speed := vS1g) to hold the minimum required climb gradient for one engine inoperative in the 2nd climb segment
    *            (end altitude: 400ft)
    *   \return factorV2 [-]
    */
    double getMinimumFactorForV2();
};

#endif // performance_assessment_SRC_TAW_DEFAULTMETHODS_AIRCRAFT_H_
