/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "flightEnvelopes.h"

namespace low {

flightEnvelopes::flightEnvelopes(aircraft *theAircraft)
    :
    theAcftPt(theAircraft) {
    // ctor
}
flightEnvelopes::~flightEnvelopes() {
    // dtor
}

void flightEnvelopes::doDiveSpeedLimits() {
    theAcftPt->MDive = 1.1 * theAcftPt->MMO;
    theAcftPt->VDive = 1.1 * theAcftPt->VMO;
}

void flightEnvelopes::doFlightEnvelope(std::shared_ptr<abstractCeilingPerformance> ceilingPerf) {
    // Loop over the Gross Mass
    for (unsigned int i(0); i <= ceilingPerf->theCeiling.size() - 1; i ++) {
        myFlightEnvelope.push_back(flightEnvelopeData());
        myFlightEnvelope.back().grossMass = ceilingPerf->theCeiling.at(i).Mass;
        myFlightEnvelope.back().maxAlt = ceilingPerf->theCeiling.at(i).maxCeiling;
        getStallingLimit();
        double crossOverAlt = getCrossOverAlt(convertUnit(KNOTS, METERPERSECOND, theAcftPt->VMO), theAcftPt->MMO);
        getMMOLimit(crossOverAlt);
        crossOverAlt = getCrossOverAlt(convertUnit(KNOTS, METERPERSECOND, theAcftPt->VDive), theAcftPt->MDive);
        getDiveLimit(crossOverAlt);
    }
}

void flightEnvelopes::getStallingLimit() {
    int stepSize(20);
    for (int i = 0; i <= stepSize; i++) {
        myFlightEnvelope.back().vStall.push_back(speedLimit());
        myFlightEnvelope.back().vStall.back().height = static_cast<double>(i) / static_cast<double>(stepSize) * myFlightEnvelope.back().maxAlt;
        myFlightEnvelope.back().vStall.back().speed = theAcftPt->v1gS_Clean(myFlightEnvelope.back().grossMass, myFlightEnvelope.back().vStall.back().height);
    }
}

void flightEnvelopes::getMMOLimit(double theCrossOverAlt) {
    int stepSize(20);
    for (int i = 0; i <= stepSize; i++) {
        myFlightEnvelope.back().vMO.push_back(speedLimit());
        myFlightEnvelope.back().vMO.back().height = static_cast<double>(i) / static_cast<double>(stepSize) * theCrossOverAlt;
        myFlightEnvelope.back().vMO.back().speed = convertUnit(KNOTS, METERPERSECOND, theAcftPt->VMO) /
                                                   theAcftPt->atm.densityRatio(myFlightEnvelope.back().vMO.back().height);
        myFlightEnvelope.back().MMO.push_back(speedLimit());
        myFlightEnvelope.back().MMO.back().height = theCrossOverAlt + static_cast<double>(i) / static_cast<double>(stepSize) * (myFlightEnvelope.back().maxAlt - theCrossOverAlt);
        myFlightEnvelope.back().MMO.back().speed = (theAcftPt->MMO)
                * theAcftPt->atm.getSpeedOfSound(myFlightEnvelope.back().MMO.back().height);
    }
}

void flightEnvelopes::getDiveLimit(double theCrossOverAlt) {
    int stepSize(20);
    for (int i = 0; i <= stepSize; i++) {
        myFlightEnvelope.back().vDive.push_back(speedLimit());
        myFlightEnvelope.back().vDive.back().height = static_cast<double>(i) / static_cast<double>(stepSize) * theCrossOverAlt;
        myFlightEnvelope.back().vDive.back().speed = convertUnit(KNOTS, METERPERSECOND, theAcftPt->VDive) /
                                                     theAcftPt->atm.densityRatio(myFlightEnvelope.back().vDive.back().height);
        myFlightEnvelope.back().MDive.push_back(speedLimit());
        myFlightEnvelope.back().MDive.back().height = theCrossOverAlt + static_cast<double>(i) / static_cast<double>(stepSize) * (myFlightEnvelope.back().maxAlt - theCrossOverAlt);
        myFlightEnvelope.back().MDive.back().speed = (theAcftPt->MDive)
                * theAcftPt->atm.getSpeedOfSound(myFlightEnvelope.back().MDive.back().height);
    }
}

double flightEnvelopes::getCrossOverAlt(double theSpeed, double theMach) {
    double Altitude(0.);
    for (int i(0); i <= 2; i++) {
        int j(0);
        double tempAlt = Altitude;
        while ((theSpeed / theAcftPt->atm.densityRatio(tempAlt)) / theAcftPt->atm.getSpeedOfSound(tempAlt) <= theMach
                && tempAlt <= 50000) {
            tempAlt = Altitude + j * (1. / pow(10., static_cast<double>(i))) * 1000;
            j++;
        }
        j = j - 2;
        Altitude = Altitude + j * (1. / pow(10., static_cast<double>(i))) * 1000;
    }
    return Altitude;
}
}  // namespace low
