/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "optimizeMissionProfile.h"

using std::string;

namespace low {

optimizeMissionProfile::optimizeMissionProfile(aircraft *theAircraft)
    :
    theAcftPt(theAircraft) {
    // ctor
}
optimizeMissionProfile::~optimizeMissionProfile() {
    // dtor
}

void low::optimizeMissionProfile::doOptimizeMission(std::shared_ptr<abstractFuelPlanning>& myFuelPlanning, const std::shared_ptr<RuntimeIO>& rtIO, bool sizingForMTOM) {
    double maxAltitudeForRange = 0.5 * (theAcftPt->designRange - 150.) * 1000. * tan(2.5 * 3.14159 / 180.) / 0.3048 / 100.;
    if (theAcftPt->hMaxOperating / 100. < maxAltitudeForRange) {
        maxAltitudeForRange = theAcftPt->hMaxOperating / 100.;
    }
    for (unsigned int i(0); i <= theAcftPt->myMission.Steps.size() - 1; i++) {
        if (theAcftPt->myMission.Steps.at(i).modeStep == "Cruise") {
            cruiseSteps.push_back(theCruiseSteps());
            cruiseSteps.back().ID = i + 1;
            cruiseSteps.back().oldFL        = theAcftPt->myMission.Steps.at(i).endFL;
            cruiseSteps.back().Mach         = theAcftPt->myMission.Steps.at(i).endMach;
            cruiseSteps.back().grossMass  =  0.5 * (myFuelPlanning->theMissionFuel.at(i).Mass
                                               + myFuelPlanning->theMissionFuel.at(i + 1).Mass);
            /* Calculation of the optimized flight altitude */
            cruiseSteps.back().newFL        = getNewFlightLevel(G_FORCE * cruiseSteps.back().grossMass, cruiseSteps.back().Mach);
            if (cruiseSteps.back().newFL > maxAltitudeForRange) {
                cruiseSteps.back().newFL = static_cast<int>(floor(maxAltitudeForRange));
            }
            /* Write in mission block of AC-XML file */
            std::stringstream tempID;
            tempID.str("");
            tempID << int(cruiseSteps.size());
            // TODO(franz#1#): neue xml-Klasse (mit pointer) verwenden; VORSICHT: gleiches Ergebnis?
            if (sizingForMTOM == true) {
                rtIO->acxml.at("AcftExchangeFile/Performance/MissionAnalysis/DesignMission/CruiseSteps") = static_cast<int>(cruiseSteps.size());
                rtIO->acxml.at("AcftExchangeFile/Performance/MissionAnalysis/DesignMission/CruiseStep@" + tempID.str() + "/FlightLevel") = cruiseSteps.back().newFL;
            } else {
                rtIO->acxml.at("AcftExchangeFile/Performance/MissionAnalysis/StudyMission/CruiseSteps") = static_cast<int>(cruiseSteps.size());
                rtIO->acxml.at("AcftExchangeFile/Performance/MissionAnalysis/StudyMission/CruiseStep@" + tempID.str() + "/FlightLevel") = cruiseSteps.back().newFL;
            }
        }
    }
}

double optimizeMissionProfile::getNewFlightLevel(double theLift, double theMach) {
    double deltaStep = 100.;
    //double nrSteps = floor( (theAcftPt->hMaxOperating - theAcftPt->hInitialCruise)/deltaStep );
    double targetFL(theAcftPt->hInitialCruise / 100.);
    double SAR = 0.;
    for (double i = theAcftPt->hInitialCruise; i <= theAcftPt->hMaxOperating; i = i + deltaStep) {
        double theDrag = theAcftPt->aero.getCruiseDrag(theMach,  i * 0.3048, 0., theLift / G_FORCE, "Clean", theAcftPt->atm);
        theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, i * 0.3048, theMach, theAcftPt->atm, 1., "MaxCont",
                theAcftPt->myMission.bleedOffCr, theAcftPt->myMission.shaftOffCr, theDrag / 1000.);
        if (theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise) < 0.99999 * theDrag / 1000.) {
            break;
        }
        double tempROC = theAcftPt->ROC(theLift / G_FORCE, i * 0.3048, theMach, 1.0, "Clean", "MaxCont",
                                        theAcftPt->myMission.bleedOffCr, theAcftPt->myMission.shaftOffCr,
                                        false, theAcftPt->aero) * (60. / 0.3048);
        if (tempROC < 300.) {
            break;
        }
        double tempFF = theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
        double tempTAS = theMach * theAcftPt->atm.getSpeedOfSoundISA(i * 0.3048);
        double tempSAR = tempTAS / tempFF;
        if (i <= theAcftPt->hInitialCruise) {
            SAR = tempTAS / tempFF;
            targetFL = Rounding(i / 100., 0.);
        } else if (SAR < tempSAR) {
            SAR = tempTAS / tempFF;
            targetFL = Rounding(i / 100., 0.);
        }
    }
    return targetFL;
}
}  // namespace low