/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "performance_assessment_factory.h"

#include "performance_assessment_output.h"



performance_assessment_output::performance_assessment_output(std::shared_ptr<taw::performance_assessment_factory> aFactoryPt, std::shared_ptr<aircraft> anAcftPt)
    :
    theFactoryPt(aFactoryPt),
    myAcftPt(anAcftPt),
    theReport(aFactoryPt->rtIO) {
    // ctor
    if (theFactoryPt->mySettings.mySwitches.doFuelPlanning && theFactoryPt->mySettings.useStudyMissionForAnalysis) {
        theProgramName = theFactoryPt->rtIO->programname + "_study";
    } else {
        theProgramName = theFactoryPt->rtIO->programname;
    }
}

performance_assessment_output::~performance_assessment_output() {
    // dtor
}

void performance_assessment_output::generateHtmlBody() {
    theReport.htmlReportStream() << "<h1>Report " << myAcftPt->acftName << "</h1>" << std::endl;
    if (theFactoryPt->mySettingsPt->mySwitches.doFuelPlanning == true) {
        theReport.htmlReportStream() << "<h2>Fuel Planning</h2>" << std::endl
        << "<table class=\"ceras\">" << std::endl
        << "<tr>" << std::endl
        << "<td>Taxi Fuel: </td><td>" << theFactoryPt->theFuelPlanning->myFuel.taxiFuelTakeoff + theFactoryPt->theFuelPlanning->myFuel.taxiFuelLanding << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>Mission Fuel: </td><td>" << theFactoryPt->theFuelPlanning->myFuel.missionFuel << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>Trip Fuel: </td><td>" << theFactoryPt->theFuelPlanning->myFuel.tripFuel << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>-> TOM: </td><td>" << Rounding(myAcftPt->OME
                + myAcftPt->myMission.missionPayload
                + theFactoryPt->theFuelPlanning->myFuel.missionFuel
                - theFactoryPt->theFuelPlanning->myFuel.taxiFuelTakeoff, 1) << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>-> MTOM: </td><td>" << myAcftPt->MTOM << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl
        << "<img class=\"ceras\" src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir())
                + theProgramName + "_Mission_plot.svg\">" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Payload Range</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doPayloadRange == true) {
        if (theFactoryPt->mySettings.mySwitches.doFuelPlanning == true) {
            theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
                   theProgramName + "_payloadRange.svg\">" << std::endl;
        } else {
            if (theFactoryPt->abstractPayloadRangePt->Range.at(2) > myAcftPt->designRange) {
                theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
                       theProgramName + "_payloadRange.svg\">" << std::endl;
            } else {
                theReport.htmlReportStream() << "<table>" << std::endl
                << "<tr>" << std::endl
                << "<td>... not calculated due to error (Range_maxFuelatMTOM smaller DesignRange)!</td>" << std::endl
                << "</tr>" << std::endl
                << "</table>" << std::endl;
            }
        }
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Flight Envelope</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_FlightEnvelope.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Ceiling</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_Ceiling.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Specific Air Range (SAR)</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_SAR.svg\">" << std::endl
        << "<div id=\"wrapper\">" << std::endl
        << "<div id=\"leftCol\">" << std::endl
        << "<h2>Best Specific Air Range (SAR)</h2>" << std::endl
        << "<img class=\"ceras\" src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_bestSAR.svg\">" << std::endl
        << "</div>" << std::endl
        << "</div>" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Takeoff Performance</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doTOPerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_TOFL.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>BFL Performance</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doBLFLPerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_BLFL.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Landing Performance</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doLandingPerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_LDN.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Engine Performance</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doEnginePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_EnginePerformance.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>V-n-Diagram</h2>" << std::endl;
    if (theFactoryPt->mySettings.mySwitches.doVnDiagram == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(theFactoryPt->rtIO->getPlotDir(), theFactoryPt->rtIO->getReportDir()) +
               theProgramName + "_VnDiagram.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
}

void performance_assessment_output::writeXmlOutput() {
    string subPath;
    /* MTOM */
    if (theFactoryPt->mySettings.designForMTOM == true && theFactoryPt->mySettings.mySwitches.doFuelPlanning == true) {
        theFactoryPt->rtIO->acxml.at("AcftExchangeFile/MassesAndLoadings/MTOM") = myAcftPt->MTOM;
    }
    if (theFactoryPt->mySettings.mySwitches.doFlightEnvelopePerformance == true) {
        /* Speeds */
        subPath = "/aircraft_exchange_file/assessment/performance/speed";
        theFactoryPt->rtIO->acxml.at(subPath + "/maximum_operating_mach_number") = myAcftPt->MMO;
        theFactoryPt->rtIO->acxml.at(subPath + "/maximum_operating_velocity") = myAcftPt->VMO;
        theFactoryPt->rtIO->acxml.at(subPath + "/dive_mach_number") = myAcftPt->MDive;
        theFactoryPt->rtIO->acxml.at(subPath + "/dive_velocity") = myAcftPt->VDive;
        theFactoryPt->rtIO->acxml.at(subPath + "/one_g_stall_speed_velocity") =  myAcftPt->v1gS_Clean(myAcftPt->MTOM, 0.0);
    }
    /* Start */
    if (theFactoryPt->mySettings.mySwitches.doTOPerformance == true) {
        subPath = "/aircraft_exchange_file/assessment/performance/takeoff";
        theFactoryPt->rtIO->acxml.at(subPath + "/takeoff_distance_normal_safety") = theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().TODR_AEO;
        theFactoryPt->rtIO->acxml.at(subPath + "/lift_off_speed_velocity") = myAcftPt->vLOF(myAcftPt->MTOM, 0.0);
        theFactoryPt->rtIO->acxml.at(subPath + "/decision_speed") = theFactoryPt->abstractTakeOffPerformancePt->myBLFL.v1;
        theFactoryPt->rtIO->acxml.at(subPath + "/takeoff_safety_speed") = myAcftPt->v2(myAcftPt->MTOM, 0.0);
        theFactoryPt->rtIO->acxml.at(subPath + "/final_takeoff_speed") = myAcftPt->vFTO(myAcftPt->MTOM, 0.0);
        theFactoryPt->rtIO->acxml.at(subPath + "/time_to_screen_height") = theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().Time;
        theFactoryPt->rtIO->acxml.at(subPath + "/climb_or_descend_segment_climb_gradient") = theFactoryPt->abstractTakeOffPerformancePt->scdSegmentGradient;
        theFactoryPt->rtIO->acxml.at(subPath + "/final_segment_climb_gradient") = theFactoryPt->abstractTakeOffPerformancePt->finalSegmentGradient;
    }
    /* BLFL */
    if (theFactoryPt->mySettings.mySwitches.doBLFLPerformance == true) {
        subPath = "/aircraft_exchange_file/assessment/performance/takeoff";
        theFactoryPt->rtIO->acxml.at(subPath + "/balanced_field_length") = theFactoryPt->abstractTakeOffPerformancePt->myBLFL.BLFL;
    }
    /* Landing */
    if (theFactoryPt->mySettings.mySwitches.doLandingPerformance == true) {
        subPath = "/aircraft_exchange_file/assessment/performance/landing";
        theFactoryPt->rtIO->acxml.at(subPath + "/needed_runway_length") = theFactoryPt->theLandingPerformance->myLDNField.at(0).myLDN.back().ALD_AEO;
        theFactoryPt->rtIO->acxml.at(subPath + "/approach_speed") = myAcftPt->vAPPR(myAcftPt->MLM, 0.0);
    }
    if (theFactoryPt->mySettings.mySwitches.doPayloadRange) {
        /* Ranges */
        subPath = "/aircraft_exchange_file/assessment/performance/range";
        if (theFactoryPt->mySettings.mySwitches.doFuelPlanning == true) {
            theFactoryPt->rtIO->acxml.at(subPath + "/range_max_payload_at_maximum_takeoff_mass") = theFactoryPt->theFuelPlanning->myPayloadRange.at(1).Range;
            theFactoryPt->rtIO->acxml.at(subPath + "/range_max_fuel_at_maximum_takeoff_mass") = theFactoryPt->theFuelPlanning->myPayloadRange.at(2).Range;
            theFactoryPt->rtIO->acxml.at(subPath + "/payload_maximum_fuel_at_maximum_takeoff_mass") = theFactoryPt->theFuelPlanning->myPayloadRange.at(2).Payload;
            theFactoryPt->rtIO->acxml.at(subPath + "/range_maximum_fuel_empty") = theFactoryPt->theFuelPlanning->myPayloadRange.at(3).Range;
        } else {
            if (theFactoryPt->abstractPayloadRangePt->Range.at(2) > myAcftPt->designRange) {
                theFactoryPt->rtIO->acxml.at(subPath + "/range_max_payload_at_maximum_takeoff_mass") = theFactoryPt->abstractPayloadRangePt->Range.at(1);
                theFactoryPt->rtIO->acxml.at(subPath + "/range_max_fuel_at_maximum_takeoff_mass") = theFactoryPt->abstractPayloadRangePt->Range.at(2);
                theFactoryPt->rtIO->acxml.at(subPath + "/payload_maximum_fuel_at_maximum_takeoff_mass") = theFactoryPt->abstractPayloadRangePt->Payload.at(2);
                theFactoryPt->rtIO->acxml.at(subPath + "/range_maximum_fuel_empty") = theFactoryPt->abstractPayloadRangePt->Range.at(3);
            }
        }
    }
    if (theFactoryPt->mySettings.mySwitches.doFuelPlanning == true) {
        /* Mission Fuels */
        if (!theFactoryPt->mySettings.designForMTOM && theFactoryPt->mySettings.useStudyMissionForAnalysis) {
            subPath = "/aircraft_exchange_file/analysis/mission/study_mission";
        } else {
            subPath = "/aircraft_exchange_file/analysis/mission/design_mission";
        }
        theFactoryPt->rtIO->acxml.at(subPath + "/range") = convertUnit(KILO, METER, NAUTICALMILE, myAcftPt->myMission.missionRange);
        // Fuel
        theFactoryPt->rtIO->acxml.at(subPath + "/mission_energy") = std::to_string(theFactoryPt->theFuelPlanning->myFuel.missionFuel) + "<-Fuel[kg]";  // Total Tanked Fuel
        theFactoryPt->rtIO->acxml.at(subPath + "/trip_energy") = theFactoryPt->theFuelPlanning->myFuel.tripFuel;  // Tanked fuel - Reserve Fuel - Taxi Fuel
        theFactoryPt->rtIO->acxml.at(subPath + "/taxi_energy") = theFactoryPt->theFuelPlanning->myFuel.taxiFuelTakeoff;  // TODO(hansmann): xpath&fuel2energyconversion
        theFactoryPt->rtIO->acxml.at(subPath + "/TaxifuelLdg") = theFactoryPt->theFuelPlanning->myFuel.taxiFuelLanding;  // TODO(hansmann): xpath&fuel2energyconversion
        // Rest
        theFactoryPt->rtIO->acxml.at(subPath + "/TOM") = myAcftPt->OME
                                               + myAcftPt->myMission.missionPayload
                                               + theFactoryPt->theFuelPlanning->myFuel.missionFuel
                                               - theFactoryPt->theFuelPlanning->myFuel.taxiFuelTakeoff;
        theFactoryPt->rtIO->acxml.at(subPath + "/payload_mass") = myAcftPt->myMission.missionPayload;
        theFactoryPt->rtIO->acxml.at(subPath + "/number_of_pax") = myAcftPt->myMission.missionPAX;
        theFactoryPt->rtIO->acxml.at(subPath + "/cargo_mass") = myAcftPt->myMission.missionCargo;
        theFactoryPt->rtIO->acxml.at(subPath + "/takeoff_engine_derate") = myAcftPt->myMission.Steps.at(0).derating;
        theFactoryPt->rtIO->acxml.at(subPath + "/block_time") = theFactoryPt->theFuelPlanning->blockTime;
        theFactoryPt->rtIO->acxml.at(subPath + "/flight_time") = theFactoryPt->theFuelPlanning->tripTime;
    }
    /* Requirements Checks */
    // TLARS
    subPath = "/aircraft_exchange_file/requirement_compliance/top_level_aircraft_requirements";
    if (theFactoryPt->mySettings.mySwitches.doTOPerformance == true) {
        theFactoryPt->rtIO->acxml.at(subPath + "/design_takeoff_distance/maintainable") = myAcftPt->myChecks.TOFL;
        theFactoryPt->rtIO->acxml.at(subPath + "/design_takeoff_distance/checked") = true;
    }
    if (theFactoryPt->mySettings.mySwitches.doLandingPerformance == true) {
        theFactoryPt->rtIO->acxml.at(subPath + "/design_landing_field_length/maintainable") = myAcftPt->myChecks.LDN;
        theFactoryPt->rtIO->acxml.at(subPath + "/design_landing_field_length/checked") = true;
        theFactoryPt->rtIO->acxml.at(subPath + "/design_approach_speed/maintainable") = myAcftPt->myChecks.vApproach;
        theFactoryPt->rtIO->acxml.at(subPath + "/design_approach_speed/checked") = true;
    }
    // Certification
    subPath = "AcftExchangeFile/Performance/RequirementsChecks/Certification";
    if (theFactoryPt->mySettings.mySwitches.doTOPerformance == true) {  //TODO
        theFactoryPt->rtIO->acxml.at(subPath + "/climbGradientSecondTOSegment") = myAcftPt->myChecks.climbGradientSecondSegment;
        theFactoryPt->rtIO->acxml.at(subPath + "/climbGradientSecondTOSegment").setAttrib("Checked", 1);
        theFactoryPt->rtIO->acxml.at(subPath + "/climbGradientFinalTOSegment") = myAcftPt->myChecks.climbGradientFinalSegment;
        theFactoryPt->rtIO->acxml.at(subPath + "/climbGradientFinalTOSegment").setAttrib("Checked", 1);
    }
    if (theFactoryPt->mySettings.mySwitches.doLandingPerformance == true) {
        theFactoryPt->rtIO->acxml.at(subPath + "/climbGradientApproachOEI") = myAcftPt->myChecks.climbGradientApproachOEI;
        theFactoryPt->rtIO->acxml.at("/" + subPath + "/climbGradientApproachOEI").setAttrib("Checked", 1);
        theFactoryPt->rtIO->acxml.at(subPath + "/climbGradientApproachAEO") = myAcftPt->myChecks.climbGradientApproachAEO;
        theFactoryPt->rtIO->acxml.at("/" + subPath + "/climbGradientApproachAEO").setAttrib("Checked", 1);
    }
    // aixml::saveDocument(theFactoryPt->rtIO->acxml, mySettingsPt->getToolLevel());  // TODO
    // aixml::closeDocument(theFactoryPt->rtIO->acxml);
}


void performance_assessment_output::generatePlotData() {
    if (theFactoryPt->mySettings.mySwitches.doFuelPlanning == true) {
        generatePlotDataMission();
    }
    if (theFactoryPt->mySettings.mySwitches.doPayloadRange == true) {
        generatePlotDataPayloadRange();
    }
    if (theFactoryPt->mySettings.mySwitches.doEnginePerformance == true) {
        generateEnginePerformanceData();
    }
    if (theFactoryPt->mySettings.mySwitches.doFlightEnvelopePerformance == true) {
        generateCeilingPlotData();
        generateFlightEnvelopeData();
        generateSARPlotData();
        generateBestSARAltPlotData();
    }
    if (theFactoryPt->mySettings.mySwitches.doTOPerformance == true) {
        generateTOFLPlotData();
    }
    if (theFactoryPt->mySettings.mySwitches.doBLFLPerformance == true) {
        generateBLFLPlotData();
    }
    if (theFactoryPt->mySettings.mySwitches.doLandingPerformance == true) {
        generateLDNPlotData();
    }
    if (theFactoryPt->mySettings.mySwitches.doVnDiagram == true) {
        generateVnPlotData();
    }
}

void performance_assessment_output::generatePlotDataMission() {
std::ofstream csv;
    string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Mission.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_Mission.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Mission.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) ID; (2) Time [min]; (3) Range [NM]; (4) Altitude [ft]; (5) TAS [kts]; (6) CAS [kts]; (7) Mach; ";
        csv << "(8) A/C Mass [kg]; (9) consumed Fuel [kg]; (10) total consumed Fuel [kg]; (11) Thrust [kN]; ";
        csv << "(12) Fuel Flow [kg/s]; (13) ROC [m/s]; (14) C_L" << std::endl;
        /*(1)*/ csv << "0; ";
        /*(1)*/ csv << "0; ";
        /*(3)*/ csv << "0; ";
        /*(4)*/ csv << "0; ";
        /*(5)*/ csv << "0; ";
        /*(6)*/ csv << "0; ";
        /*(7)*/ csv << "0; ";
        /*(8)*/ csv << theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(0).Mass << "; ";
        /*(9)*/ csv << "0; ";
        /*(10)*/ csv << "0; ";
        /*(11)*/ csv << "0; ";
        /*(12)*/ csv << "0; ";
        /*(13)*/ csv << "0; ";
        /*(14)*/ csv << "0" << std::endl;
        double time(0.);
        double range(0.);
        double fuel(0.);
        for (unsigned int i = 0; i <= myAcftPt->myMission.Steps.size() - 1; i++) {
            time += theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).Time;
            range += theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).Range;
            fuel += theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).consumedFuel;
            double C_L = (2. * G_FORCE * theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).Mass)
                  / (myAcftPt->atm.getDensity(convertUnit(FOOT, METER, myAcftPt->myMission.Steps.at(i).endAltitude))
                     *  pow(convertUnit(KNOTS, METERPERSECOND, myAcftPt->myMission.Steps.at(i).endTAS), 2.)
                     * myAcftPt->S_Wing);
            /*(1)*/ csv << i + 1 << "; ";
            /*(1)*/ csv << convertUnit(MINUTE, HOUR, time) << "; ";
            /*(3)*/ csv << convertUnit(METER, NAUTICALMILE, range) << "; ";
            /*(4)*/ csv << myAcftPt->myMission.Steps.at(i).endAltitude << "; ";
            /*(5)*/ csv << myAcftPt->myMission.Steps.at(i).endTAS  << "; ";
            /*(6)*/ csv << myAcftPt->myMission.Steps.at(i).endCAS  << "; ";
            /*(7)*/ csv << myAcftPt->myMission.Steps.at(i).endMach  << "; ";
            /*(8)*/ csv << theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).Mass << "; ";
            /*(9)*/ csv << theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).consumedFuel << "; ";
            /*(10)*/ csv << fuel << "; ";
            /*(11)*/ csv << theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).Thrust << "; ";
            /*(12)*/ csv << theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).fuelFlow << "; ";
            /*(13)*/ csv << theFactoryPt->abstractFuelPlanningPt->theMissionFuel.at(i + 1).ROC << "; ";
            /*(14)*/ csv << C_L << std::endl;
        }
        csv.close();
    }
}

void performance_assessment_output::generatePlotDataPayloadRange() {
std::ofstream csv;
    string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_payloadRange.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_payloadRange.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_missionFuel.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Range [NM]; (2) Payload [kg]; (3) Fuel [kg]; (4) TOM [kg]; (5) OME [kg]" << std::endl;
        if (theFactoryPt->mySettingsPt->mySwitches.doFuelPlanning == true) {
            for (unsigned int i(0); i <= theFactoryPt->abstractFuelPlanningPt->myPayloadRange.size() - 1; i++) {
                csv << convertUnit(KILO, METER, NAUTICALMILE, theFactoryPt->abstractFuelPlanningPt->myPayloadRange.at(i).Range) << "; ";
                csv << theFactoryPt->abstractFuelPlanningPt->myPayloadRange.at(i).Payload << "; ";
                // csv << myPerfPt->theFuelPlanning.myPayloadRange.at(i).Fuel << "; ";
                // csv << myPerfPt->theFuelPlanning.myPayloadRange.at(i).TOM << "; ";
                // csv << myAcftPt->OME << "; ";
                csv << std::endl;
            }
        } else {
            for (unsigned int i(0); i <= theFactoryPt->abstractPayloadRangePt->Range.size() - 1; i++) {
                csv << convertUnit(KILO, METER, NAUTICALMILE, theFactoryPt->abstractPayloadRangePt->Range.at(i)) << "; ";
                csv << theFactoryPt->abstractPayloadRangePt->Payload.at(i) << "; ";
                csv << theFactoryPt->abstractPayloadRangePt->Fuel.at(i) << "; ";
                csv << theFactoryPt->abstractPayloadRangePt->TOM.at(i) << "; ";
                // csv << myPerfPt->theFuelPlanning.myPayloadRange.at(i).Fuel << "; ";
                // csv << myPerfPt->theFuelPlanning.myPayloadRange.at(i).TOM << "; ";
                // csv << myAcftPt->OME << "; ";
                csv << std::endl;
            }
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# Design Point " << std::endl;
        csv << convertUnit(KILO, METER, NAUTICALMILE, myAcftPt->designRange) << "; ";
        csv << myAcftPt->designPayload << "; ";
        // csv << "N/A" << "; ";
        // csv << myAcftPt->MTOM << "; ";
        // csv << myAcftPt->OME << "; ";
        csv << std::endl;
        csv << std::endl;
        csv << std::endl;
        csv << "# Mission Point" << std::endl;
        csv << convertUnit(KILO, METER, NAUTICALMILE, myAcftPt->myMission.missionRange) << "; ";
        csv << myAcftPt->myMission.missionPayload << "; ";
        // csv << myPerfPt->theFuelPlanning.myFuel.tripFuel << "; ";
        // csv << myAcftPt->MTOM << "; ";
        // csv << myAcftPt->OME << "; ";
        csv.close();
    }

}

void performance_assessment_output::generateEnginePerformanceData() {
    for (unsigned int i(0); i <= theFactoryPt->abstractEnginePerformancePt->theRatingEngine.size() - 1; i++) {
        std::ofstream csv;
        string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_EnginePerformance_" +
            theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).Rating + ".csv");
        csv.open(csvDataFile.c_str());
        if (!csv) {
            myRuntimeInfo->err << string(theProgramName) + "_EnginePerformance" + theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).Rating
                + ".csv could not be opened!" << std::endl;
            exit(1);
        } else {
            csv << "# " + theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_EnginePerformance" +
                theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).Rating + ".csv" << std::endl;
            csv << "# Rating: " + theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).Rating << std::endl;
            /* Adjust data structure: */
            csv << "# (1) Mach; (2) Altitude [ft]; (3) Thrust per Engine [kN]; (4) SFC [mg/Ns]" << std::endl;
            for (unsigned int j(0); j <= theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.size() - 1; j++) {
                for (unsigned int n(0); n <= theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.size() - 1; n++) {
                    csv << theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.at(n) << "; ";
                    csv << (1. / 0.3048)*theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Altitude << "; ";
                    csv << theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Thrust.at(n) << "; ";
                    csv << 1000.*theFactoryPt->abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).SFC.at(n);
                    csv << std::endl;
                }
                csv << std::endl;
                csv << std::endl;
            }
        }
        csv.close();
    }
}
void performance_assessment_output::generateCeilingPlotData() {
    std::ofstream csv;
    string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Ceiling.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_Ceiling.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Ceiling.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Service Ceiling [ft]; (4) Max Ceiling [ft] ; (5) OEI max Ceiling [ft]; (6) Initial Cruise Altitude [ft]" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractCeilingPerformancePt->theCeiling.size() - 1; i++) {
            csv << theFactoryPt->abstractCeilingPerformancePt->theCeiling.at(i).Mach << "; ";
            csv << theFactoryPt->abstractCeilingPerformancePt->theCeiling.at(i).Mass << "; ";
            csv << theFactoryPt->abstractCeilingPerformancePt->theCeiling.at(i).serviceCeiling * 3.2808 << "; ";
            csv << theFactoryPt->abstractCeilingPerformancePt->theCeiling.at(i).maxCeiling * 3.2808 << "; ";
            csv << theFactoryPt->abstractCeilingPerformancePt->theCeiling.at(i).maxCeilingOEI * 3.2808 << "; ";
            csv << myAcftPt->hInitialCruise << std::endl;
        }
        csv.close();
    }
}
void performance_assessment_output::generateFlightEnvelopeData() {
    std::ofstream csv;
    string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_FlightEnvelope.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_FlightEnvelope.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_FlightEnvelope.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Altitude [ft]; (2) TAS [kts]; (3) Gross Mass [kg]" <<std::endl;
        csv << "# Stall Limit" <<std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.size() - 1; i++) {
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).height / 0.3048  << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).speed / (1.852 / 3.6) << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# VMO Limit" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.size() - 1; i++) {
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).height / 0.3048  << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).speed / (1.852 / 3.6) << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# MMO Limit" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.size() - 1; i++) {
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).height / 0.3048  << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).speed / (1.852 / 3.6) << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv << std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "# VDive Limit" <<std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.size() - 1; i++) {
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).height / 0.3048  << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).speed / (1.852 / 3.6) << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "# MDive Limit" <<std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.size() - 1; i++) {
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).height / 0.3048  << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).speed / (1.852 / 3.6) << "; ";
            csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "#Altitude Limit" <<std::endl;
        csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.back().height / 0.3048  << "; ";
        csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.back().speed / (1.852 / 3.6) << "; ";
        csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
        csv <<std::endl;
        csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.back().height / 0.3048  << "; ";
        csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.back().speed / (1.852 / 3.6) << "; ";
        csv << theFactoryPt->abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
        csv.close();
    }
}
void performance_assessment_output::generateSARPlotData() {
    std::ofstream csv;
    for (unsigned int i(0); i < theFactoryPt->abstractSARPerformancePt->theSAR.size(); i++) {
        std::stringstream Number;
        Number << i;
        string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_SAR_Mass" + Number.str() + ".csv");
        csv.open(csvDataFile.c_str());
        if (!csv) {
            myRuntimeInfo->err << csvDataFile + " could not be opened!" <<std::endl;
            exit(1);
        } else {
            csv << "# " + csvDataFile <<std::endl;
            // Adjust data structure:
            csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Altitude [ft]; (4) SAR [km/kg]" <<std::endl;
            for (unsigned int j(0); j < theFactoryPt->abstractSARPerformancePt->theSAR.at(i).tabSAR.size(); j++) {
                //csv << "# " << floor(myPerfPt->theSAR.at(i).Mass/1000)*1000 << " kg, " << floor(3.2808 * myPerfPt->theSAR.at(i).tabSAR.at(j).Altitude/100)*100<<std::endl;
                for (unsigned int n(0); n < theFactoryPt->abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Mach.size(); n++) {
                    csv << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Mach.at(n) << "; ";
                    csv << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).Mass << "; ";
                    csv << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Altitude / 0.3048 << "; ";
                    csv << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).SAR.at(n) / 1000. <<std::endl;
                }
                csv <<std::endl;
                csv <<std::endl;
            }
        }
        csv.close();
    }
}
void performance_assessment_output::generateBestSARAltPlotData() {
// TODO(franz#1#): kathi (18.11.2013): Plot seltsam -> reparieren
    std::ofstream plot;
    string plotDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_bestSAR.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << string(theProgramName) + "_bestSAR.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + string(theProgramName) + "_bestSAR.csv" << std::endl;
        // Adjust data structure:
        plot << "# (1) Mach; (2) Gross Mass [kg]; (3) Opt. Altitude [ft]; (4) Opt. SAR [km/kg]" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractSARPerformancePt->theSAR.size() - 1; i++) {
            plot << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).bestSARMach << "; ";
            plot << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).Mass << "; ";
            plot << 3.2808 * theFactoryPt->abstractSARPerformancePt->theSAR.at(i).bestSARAlt << "; ";
            plot << theFactoryPt->abstractSARPerformancePt->theSAR.at(i).bestSAR / 1000. << std::endl;
        }
        plot.close();
    }
}
void performance_assessment_output::generateTOFLPlotData() {
std::ofstream csv;
    string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_TOFL.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_TOFL.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_TOFL.csv" << std::endl;
        csv << "# (1) Altitude [ft]; (2) Gross Mass [kg]; (3) TODR_AEO [m]" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.size() - 1; i++) {
            for (unsigned int j(0); j <= theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.size() - 1; j++) {
                csv << theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(i).elevation * 1. / 0.3048 << "; ";
                csv << theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).grossMass << "; ";
                csv << theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).TODR_AEO << std::endl;
            }
            csv << std::endl;
            csv << std::endl;
        }
        // display TOFL
        csv << theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(0).elevation * 1. / 0.3048 << "; ";
        csv << theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().grossMass << "; ";
        csv << theFactoryPt->abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().TODR_AEO << std::endl;
        csv << std::endl;
        csv << std::endl;
        // Req TOFL
        csv << 0.0 << "; ";
        csv << myAcftPt->MTOM << "; ";
        csv << myAcftPt->myReqs.TOFL << std::endl;
        csv.close();
    }
}
void performance_assessment_output::generateBLFLPlotData() {
    std::ofstream csv;
    string csvDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_BLFL.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_BLFL.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_BLFL.csv" << std::endl;
        csv << "# (1) Altitude [ft]; (2) Gross Mass [kg]; (3) TODR_AEO [m]; (4) TODR_OEI [m]; (5) ASDR_OEI [m]; (6) v1 [m/s]" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->abstractTakeOffPerformancePt->myBLFL.vEF.size() - 1; i++) {
            csv << "0.0" << "; ";
            csv << theFactoryPt->abstractTakeOffPerformancePt->myBLFL.grossMass << "; ";
            csv << theFactoryPt->abstractTakeOffPerformancePt->myBLFL.TODR_AEO << "; ";
            csv << theFactoryPt->abstractTakeOffPerformancePt->myBLFL.TODR_OEI.at(i) << "; ";
            csv << theFactoryPt->abstractTakeOffPerformancePt->myBLFL.ASDR_OEI.at(i) << "; ";
            csv << theFactoryPt->abstractTakeOffPerformancePt->myBLFL.vEF.at(i) << std::endl;
        }
        csv.close();
    }
}
void performance_assessment_output::generateLDNPlotData() {
    std::ofstream plot;
    string plotDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_LDN.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << string(theProgramName) + "_LDN.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_LDN.csv" << std::endl;
        plot << "# (1) Altitude [ft]; (2) Gross Mass [kg]; (3) ALD_AEO [m]" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->theLandingPerformance->myLDNField.size() - 1; i++) {
            for (unsigned int j(0); j <= theFactoryPt->theLandingPerformance->myLDNField.at(i).myLDN.size() - 1; j++) {
                plot << theFactoryPt->theLandingPerformance->myLDNField.at(i).elevation * 1. / 0.3048 << "; ";
                plot << theFactoryPt->theLandingPerformance->myLDNField.at(i).myLDN.at(j).grossMass << "; ";
                plot << theFactoryPt->theLandingPerformance->myLDNField.at(i).myLDN.at(j).ALD_AEO << std::endl;
            }
            plot << std::endl;
            plot << std::endl;
        }
        // display TOFL
        plot << theFactoryPt->theLandingPerformance->myLDNField.at(0).elevation * 1. / 0.3048 << "; ";
        plot << theFactoryPt->theLandingPerformance->myLDNField.at(0).myLDN.back().grossMass << "; ";
        plot << theFactoryPt->theLandingPerformance->myLDNField.at(0).myLDN.back().ALD_AEO << std::endl;
        plot << std::endl;
        plot << std::endl;
        // Req TOFL
        plot << 0.0 << "; ";
        plot << myAcftPt->MLM << "; ";
        plot << myAcftPt->myReqs.LDN << std::endl;
        plot.close();
    }
}
void performance_assessment_output::generateVnPlotData() {
    std::ofstream plot;
    string plotDataFile(theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_VnDiagram.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << string(theProgramName) + "_VnDiagram.csv could not be opened!" << std::endl;
        getchar();
        exit(1);
    } else {
        plot << "# " + theFactoryPt->mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_VnDiagram.csv" << std::endl;
        plot << "# (1) v [kts]; (2) max n (pos.) [-]; (3) min n (neg.) [-]" << std::endl;
        plot << "#Clean configuration" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->theVnDiagram->myCleanDiagram.size() - 1; i++) {
            plot << theFactoryPt->theVnDiagram->myCleanDiagram.at(i).speed << "; ";
            plot << theFactoryPt->theVnDiagram->myCleanDiagram.at(i).n_upperBound << "; ";
            plot << theFactoryPt->theVnDiagram->myCleanDiagram.at(i).n_lowerBound << std::endl;
        }
        plot << std::endl;
        plot << std::endl;
        plot << "#Takeoff configuration" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->theVnDiagram->myCleanDiagram.size() - 1; i++) {
            plot << theFactoryPt->theVnDiagram->myTakeoffDiagram.at(i).speed << "; ";
            plot << theFactoryPt->theVnDiagram->myTakeoffDiagram.at(i).n_upperBound << "; ";
            plot << theFactoryPt->theVnDiagram->myTakeoffDiagram.at(i).n_lowerBound << std::endl;
        }
        plot << std::endl;
        plot << std::endl;
        plot << "#Landing configuration" << std::endl;
        for (unsigned int i(0); i <= theFactoryPt->theVnDiagram->myCleanDiagram.size() - 1; i++) {
            plot << theFactoryPt->theVnDiagram->myLandingDiagram.at(i).speed << "; ";
            plot << theFactoryPt->theVnDiagram->myLandingDiagram.at(i).n_upperBound << "; ";
            plot << theFactoryPt->theVnDiagram->myLandingDiagram.at(i).n_lowerBound << std::endl;
        }
        plot.close();
    }


}
