/*  Copyright (C) 2023 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file turbofan.cpp
 * @author Sebastian Oberschwendtner (sebastian.oberschwendtner@tum.de)
 * @brief Design a the nacelle of a turbofan engine using
 * the default nacelle designer.
 * @version 3.0.0
 * @date 2023-12-20
 */

/* === Includes === */
#include "default_nacelle.h"
#include <algorithm>
#include <aircraftGeometry2/geometry/builder.h>

/* === Design implementations === */
namespace geometry
{
    namespace nacelle
    {
        void Default::operator()(Turbofan<EnergyCarrier::Kerosene> &engine)
        {
            /* Initialize the surface and its builder */
            geom2::MultisectionSurface<geom2::PolygonSection> surface{};
            geom2::SectionBuilder<geom2::PolygonSection> builder{};

            /* Get the maximum diameter the nacelle has to fit */
            const double diameter_inner = std::max(
                { engine.dimension().width, engine.dimension().height, engine.fan().diameter });

            /* Create the cross section shape */
            geom2::PolygonSection shape = this->get_section_shape(engine.id());
            shape.set_width(diameter_inner);
            shape.set_height(diameter_inner);

            /* 
             * Transform the scaled polygon coordinates to
             * absolute values, so that the section can be
             * scaled using the scale factor.
             */
            shape.set_contour(shape.get_contour(true));

            /* Section 0 */
            shape.set_scale(1.0);
            builder.insert_back(shape, {0.0, 0.0, 0.0});

            /* Section 1 */
            shape.set_scale(1.25);
            builder.insert_back(shape, {0.0, 0.0, engine.dimension().length * 0.25});

            /* Section 2 */
            shape.set_scale(1.25);
            builder.insert_back(shape, {0.0, 0.0, engine.dimension().length * 0.50});

            /* Section 3 */
            shape.set_scale(1.0);
            builder.insert_back(shape, {0.0, 0.0, engine.dimension().length * 0.25});

            /* Build the surface and insert it into the engine */
            surface.name = "nacelle_" + std::to_string(engine.id());
            surface.normal = {1.0, 0.0, 0.0}; // Orient the nacelle along the global X direction
            surface.sections = builder.get_result();
            engine.set_nacelle(surface);
        }
    } // namespace nacelle
} // namespace geometry
