/*  Copyright (C) 2023 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file turbofan.cpp
 * @author Sebastian Oberschwendtner (sebastian.oberschwendtner@tum.de)
 * @brief Design a the pylon of a turbofan geometry integration using
 * the default pylon designer.
 * @version 3.0.0
 * @date 2023-12-20
 */

/* === Includes === */
#include "default_pylon.h"
#include <aircraftGeometry2/io/dat.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <cmath>
#include <set>
#include <utility>

/* === Design implementations === */
namespace geometry
{
    namespace pylon
    {
        void Default::operator()(Turbofan<EnergyCarrier::Kerosene> &engine)
        {
            /* Check whether a pylon can be designed */
            if (
                std::get<Vertical>(engine.parent_component()) == Vertical::In ||
                !this->aircraft()
            )
            {
                return;
            }

            /* Get the start and end points of pylon */
            const auto attachment = this->get_attachment_points(
                engine.position(), engine.nacelle(), engine.parent_component());

            /* Get profile of each section */
            auto profile = this->get_section_profile(engine.id());

            /* Set the global location and orientation of the pylon */
            geometry::Pylon pylon{};
            pylon.origin = CGAL::ORIGIN;
            pylon.normal = (engine.position() - attachment.first).direction();

            /* Set the pylon name */
            pylon.name = "pylon_" + std::to_string(engine.id());

            /*
             * The offset of the engine to the origin since
             * the pylon origin relative to the engine position.
             */
            const auto offset = engine.position() - CGAL::ORIGIN;

            /* Section at nacelle attachment */
            pylon.sections.emplace_back(profile);
            pylon.sections.back().origin = geom2::transform::to_local(pylon, attachment.first - offset);
            pylon.sections.back().set_chord_length(engine.dimension().length);

            /* Section attaching to the parent component */
            pylon.sections.emplace_back(profile);
            pylon.sections.back().origin = geom2::transform::to_local(pylon, attachment.second - offset);
            pylon.sections.back().set_chord_length(engine.dimension().length);

            /* Add the pylon to the engine */
            engine.set_pylon(pylon);
        }
    } // namespace pylon
} // namespace geometry
