/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/* The following statements import necessary header files */
#include "conventionalATA25.h"                               // Import of header file from src/aircraftSystems folder for conventionalATA25
#include <runtimeInfo/runtimeInfo.h>                         // Import of header file from external runtime library for using runtimeInfo
/* Constructor for conventionalATA25 */
conventionalATA25::conventionalATA25(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-25";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventionalATA25 */
conventionalATA25::~conventionalATA25() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA25::readParameters(const node& config) {
    std::string furnishingConfig = "/module_configuration_file/program_settings/furnishing";
    galleyLoadFractionTakeOff = config.at(furnishingConfig + "/galley_load_fraction_takeoff/value");
    data_->configuration.checkForValidity(galleyLoadFractionTakeOff, 0.0, 1., "galleyLoadFractionTakeOff");
    galleyLoadFractionCruise = config.at(furnishingConfig + "/galley_load_fraction_cruise/value");
    data_->configuration.checkForValidity(galleyLoadFractionCruise, 0.0, 1., "galleyLoadFractionCruise");
    galleyLoadFractionDescent = config.at(furnishingConfig + "/galley_load_fraction_descent/value");
    data_->configuration.checkForValidity(galleyLoadFractionDescent, 0.0, 1., "galleyLoadFractionDescent");
    nonPersonelIFEPower = config.at(furnishingConfig + "/non_personal_ife_power/value");
    data_->configuration.checkForValidity(nonPersonelIFEPower, 0.0, 500., "nonPersonelIFEPower");
    powerPersonelIFE = config.at(furnishingConfig + "/power_personal_ife/value");
    data_->configuration.checkForValidity(powerPersonelIFE, 0.0, 500., "powerPersonelIFE");
    personelIFELoadFractionClimb = config.at(furnishingConfig + "/personal_ife_load_fraction_climb/value");
    data_->configuration.checkForValidity(personelIFELoadFractionClimb, 0.0, 1., "personalIFELoadFractionClimb");
    personelIFELoadFractionCruise = config.at(furnishingConfig + "/personal_ife_load_fraction_cruise/value");
    data_->configuration.checkForValidity(personelIFELoadFractionCruise, 0.0, 1., "personalIFELoadFractionCruise");
    personelIFELoadFractionDescent = config.at(furnishingConfig + "/personal_ife_load_fraction_descent/value");
    data_->configuration.checkForValidity(personelIFELoadFractionDescent, 0.0, 1., "personalIFELoadFractionDescent");
    numberOfPowerSources = config.at(furnishingConfig + "/shaft_power_sources/number_of_power_sources/value");
    double operationFactor_gesamt(0.);
    for (int i(0); i < numberOfPowerSources; i++) {
        powerSources.push_back(PowerSource());
        powerSources.back().type =
            std::string(config.at(furnishingConfig + "/shaft_power_sources/power_source@"
                                     + num2Str(i) + "/type/value"));
        powerSources.back().sourceID =
            config.at(furnishingConfig + "/shaft_power_sources/power_source@"
                         + num2Str(i) + "/source_ID/value");
        powerSources.back().operationFactor =
            config.at(furnishingConfig + "/shaft_power_sources/power_source@"
                         + num2Str(i) + "/operation_factor/value");
        operationFactor_gesamt += Rounding(powerSources.back().operationFactor, 10);
    }
    if (operationFactor_gesamt < 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for the Furnishing results together in less than 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!" << std::endl;
        exit(1);
    } else if (operationFactor_gesamt > 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for the Furnishing results together in more than 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!" << std::endl;
        exit(1);
    }
}
/* Updating system specific parameters in aircraft xml file */
void conventionalATA25::updateXML() {}

/* Calculation of the system power */
void conventionalATA25::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}

/* Calculation of the skin mass */
void conventionalATA25::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    //Attention: ATA33 (Lighting) is subtracted here, because in [Tor13] formula for ATA-25 (Furnishing) already contained
    Mass = getMass();
    //- 0.02 * 0.92 * pow(data_->data.Mass.MTOM, 0.85);//ata33 Masse not substracted because furnishing mass comes from fuselage design
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA25::getPower() {
    /* IFE design power estimation */
    double designIFEPower(0.);
    double designGalleyPower(0.);
    /* Estimate designIFEPower */
    designIFEPower = (powerPersonelIFE + nonPersonelIFEPower) * data_->data.Accommodation.numberIFE;
    /* Estimate designGalleyPower */
    designGalleyPower = data_->data.Accommodation.galleyPower;
    for (size_t n(0); n < Electric.size(); n++) {
        // NOTE(aigner#1#): Temporarily uniform distribution of the designPower to the electrical circuits
        Electric.at(n).designPower = (designIFEPower + designGalleyPower) / static_cast<double>(Electric.size());
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad = 0.0;
        for (size_t n(0); n < Hydraulic.size(); n++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad = 0.0;
        }
        for (size_t n(0); n < Electric.size(); n++) {
            Electric.at(n).missionPower.at(i).baseLoad = 0.0;
            if (data_->mission.Steps.at(i).missionPhase == "Departure") {
                Electric.at(n).missionPower.at(i).baseLoad += personelIFELoadFractionClimb * designIFEPower;
                if (data_->mission.Steps.at(i).modeStep == "Accelerate" ||
                        data_->mission.Steps.at(i).modeStep == "Climb") {
                    Electric.at(n).missionPower.at(i).baseLoad += galleyLoadFractionCruise * designGalleyPower;
                } else {
                    Electric.at(n).missionPower.at(i).baseLoad += galleyLoadFractionTakeOff * designGalleyPower;
                }
            } else if (data_->mission.Steps.at(i).missionPhase == "Approach" || data_->mission.Steps.at(i).modeStep == "Descend_to_approach") {
                Electric.at(n).missionPower.at(i).baseLoad += personelIFELoadFractionDescent * designIFEPower;
                Electric.at(n).missionPower.at(i).baseLoad += galleyLoadFractionDescent * designGalleyPower;
            } else if (data_->mission.Steps.at(i).missionPhase == "Cruise") {
                if (data_->mission.Steps.at(i).modeStep == "change_speed_to_CAS" ||
                        data_->mission.Steps.at(i).modeStep == "Climb") {
                    Electric.at(n).missionPower.at(i).baseLoad += personelIFELoadFractionClimb * designIFEPower;
                    Electric.at(n).missionPower.at(i).baseLoad += galleyLoadFractionCruise * designGalleyPower;
                } else {
                    Electric.at(n).missionPower.at(i).baseLoad += personelIFELoadFractionCruise * designIFEPower;
                    Electric.at(n).missionPower.at(i).baseLoad += galleyLoadFractionCruise * designGalleyPower;
                }
            } else {
                Electric.at(n).missionPower.at(i).baseLoad += designIFEPower;
                Electric.at(n).missionPower.at(i).baseLoad += designGalleyPower;
            }
            // NOTE (aigner#1#): Temporarily uniform distribution of the baseLiad to the electrical circuits
            Electric.at(n).missionPower.at(i).baseLoad *= 1 / static_cast<double>(Electric.size());
        }
        /* Accumulate the electrical power to a total powerProfile*/
        powerProfile myElectricSum(getPowerProfileSum(Electric));
        /* Direct conversion from energy to heat */
        heatLoad.missionPower.at(i).baseLoad = 0.2 * myElectricSum.missionPower.at(i).baseLoad;
    }
}

/* Determination of the mass */
double conventionalATA25::getMass() {
    //calculated in fuselageDesign
    double massATA25 = data_->data.Accommodation.furnishingMass;
    return  massATA25;
}

/* Determination of the reference point */
Vec3 conventionalATA25::getRefPoint() {
    double x_cg(0.);
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    // 0.45*l_fus (arrangement kri&fab), because cockpit+cabin elements are included (and no passenger items)
    x_cg = data_->theFuselage.back().origin.x() + (0.45 * fuselage_length);
    return Vec3(x_cg, data_->theFuselage.back().origin.y(), data_->theFuselage.back().origin.z());
}
