/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA35_H_
#define SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA35_H_

/* The following statements import necessary libraries */
#include <string>                                                       // Import c++ standard library for using string
#include <vector>                                                       // Import c++ standard library for using vector
#include <memory>                                                       // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include <svl/SVL.h>                                                    // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                                          // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                            // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                           // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                        // aircraftSystem.h because that would lead to circular dependencies

/** \class conventionalATA35
 * \brief class that contains all properties of the oxygen system (ATA35)
 */
class conventionalATA35: public aircraftSystem {
 public:
    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    void calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                               const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                               const std::vector<powerProfile>& allPreviousSystemsElectric,
                               const powerProfile& allPreviousSystemsHeat) override;
    void calculateSinkMass(const powerProfile& sourceBleed,
                           const std::vector<powerProfile>& sourceElectric,
                           const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief  calculates the power demand of the oxygen system (ATA35)
     * \return void
     */
    void getPower();

    /** \brief calculates the mass of the oxygen system (ATA35)
     * \return double massATA35: mass of the oxygen system[kg]
     */
    double getMass();

    /** \brief calculates the reference point of the oxygen system (ATA35)
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether system is operating (1) or not(0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     */
    conventionalATA35(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA35();
};

#endif // SYSTEMSDESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA35_H_
