/**
 * \file massPropertiesIO.h
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Mass properties class
 * \version 0.1
 * \date 2023-11-09
 *
 * @copyright Copyright (c) 2023
 *
 */

#include <aixml/endnode.h>
#include <aixml/node.h>

#include <Eigen/Dense>
#include <iostream>
#include <map>

class MassPropertiesIO {
 public:
  explicit MassPropertiesIO(const std::string& subPath, const std::string& component)
      : subPath_(subPath + "/mass_properties") {
    description_ += component;
    data = std::map<std::string, Endnode<double>>{
        {"mass", Endnode<double>(subPath_ + "/mass", "component mass")},
        {"j_xx", Endnode<double>(subPath_ + "/inertia/j_xx", component + " inertia in x")},
        {"j_yy", Endnode<double>(subPath_ + "/inertia/j_yy", component + " inertia in y")},
        {"j_zz", Endnode<double>(subPath_ + "/inertia/j_zz", component + " inertia in z")},
        {"j_xy", Endnode<double>(subPath_ + "/inertia/j_xy", component + " inertia in xy")},
        {"j_xz", Endnode<double>(subPath_ + "/inertia/j_xz", component + " inertia in xz")},
        {"j_yx", Endnode<double>(subPath_ + "/inertia/j_yx", component + " inertia in yx")},
        {"j_yz", Endnode<double>(subPath_ + "/inertia/j_yz", component + " inertia in yz")},
        {"j_zx", Endnode<double>(subPath_ + "/inertia/j_zx", component + " inertia in zx")},
        {"j_zy", Endnode<double>(subPath_ + "/inertia/j_zy", component + " inertia in zy")},
        {"x", Endnode<double>(subPath_ + "/center_of_gravity/x", "x position of center of gravity")},
        {"y", Endnode<double>(subPath_ + "/center_of_gravity/y", "y position of center of gravity")},
        {"z", Endnode<double>(subPath_ + "/center_of_gravity/z", "z position of center of gravity")}};
    data["mass"] = {"kg", 0.0, std::numeric_limits<double>::max()};
    data["j_xx"] = {"kgm^2", 0.0, std::numeric_limits<double>::max()};
    data["j_yy"] = {"kgm^2", 0.0, std::numeric_limits<double>::max()};
    data["j_zz"] = {"kgm^2", 0.0, std::numeric_limits<double>::max()};
    data["j_xy"] = {"kgm^2", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["j_xz"] = {"kgm^2", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["j_yx"] = {"kgm^2", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["j_yz"] = {"kgm^2", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["j_zx"] = {"kgm^2", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["j_zy"] = {"kgm^2", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["x"] = {"m", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["y"] = {"m", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
    data["z"] = {"m", std::numeric_limits<double>::lowest(), std::numeric_limits<double>::max()};
  }
  std::map<std::string, Endnode<double>> data;

  void update(node& xml) {
    for (auto& place : order) {
      data[place.second].update(xml);
    }
    xml[subPath_].setAttrib<std::string>("description", description_);
    xml[subPath_ + "/inertia"].setAttrib<std::string>("description",
                                                      "component inertia referred to local center of gravity");
    xml[subPath_ + "/center_of_gravity"].setAttrib<std::string>(
        "description", "component center of gravity with respect to global coordinate system");
  }

  void read(const node& xml) {
    for (auto& place : order) {
      data[place.second].read(xml);
    }
  }

  void print() {
    for (auto& place : order) {
      data[place.second].print();
    }
  }

  void operator=(const double& mass) { data["mass"] = mass; }
  void operator=(const Eigen::Matrix3d& inertia) {
    data["j_xx"] = inertia(0, 0);
    data["j_yy"] = inertia(1, 1);
    data["j_zz"] = inertia(2, 2);
    data["j_xy"] = inertia(0, 1);
    data["j_xz"] = inertia(0, 2);
    data["j_yx"] = inertia(1, 0);
    data["j_yz"] = inertia(1, 2);
    data["j_zx"] = inertia(2, 0);
    data["j_zy"] = inertia(2, 1);
  }
  void operator=(const std::tuple<Eigen::Matrix3d, Eigen::Matrix3d>& inertia_boundaries) {
    const auto [lb, ub] = inertia_boundaries;
    data["j_xx"] = {lb(0, 0), ub(0, 0)};
    data["j_yy"] = {lb(1, 1), ub(1, 1)};
    data["j_zz"] = {lb(2, 2), ub(2, 2)};
    data["j_xy"] = {lb(0, 1), ub(0, 1)};
    data["j_xz"] = {lb(0, 2), ub(0, 2)};
    data["j_yx"] = {lb(1, 0), ub(1, 0)};
    data["j_yz"] = {lb(1, 2), ub(1, 2)};
    data["j_zx"] = {lb(2, 0), ub(2, 0)};
    data["j_zy"] = {lb(2, 1), ub(2, 1)};
  }

 private:
  std::string subPath_;
  std::string component_;
  std::string description_{"mass_properties of component "};
  std::map<int, std::string> order{
      {0, "mass"}, {1, "j_xx"}, {2, "j_yy"}, {3, "j_zz"}, {4, "j_xy"}, {5, "j_xz"}, {6, "j_yx"},
      {7, "j_yz"}, {8, "j_zx"}, {9, "j_zy"}, {10, "x"},   {11, "y"},   {12, "z"},
  };
};