/**
 * \file missionData.cpp
 * \author 
 * \brief Class for data from mission xml
 * \version 0.1
 * \date 2024-06-27
 * 
 * @copyright Copyright (c) 2024
 * 
 */

#include "missionData.h"
#include <cmath>
#include <aixml/node.h>
#include <aixml/endnode.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>

missionData::missionData(const std::string& fileNameAndDir)
    :
    missionFile(fileNameAndDir),
    missionXML_tmp(aixml::openDocument(missionFile)) {
    myRuntimeInfo->out << "Reading the Mission: " << fileNameAndDir << std::endl;
    missionUtilization = EndnodeReadOnly<double>("mission/number_of_pax").read(missionXML_tmp).value();
    delta_ISA = 0.; //NOTE: All systems are sized and calculated for ISA conditions and not for the respective mission temperature
    numberDepartureSteps = missionXML_tmp.getVector("mission/departure/departure_step").size();
    numberCruiseSteps = missionXML_tmp.getVector("mission/cruise/cruise_step").size();
    numberApproachSteps = missionXML_tmp.getVector("mission/approach/approach_step").size();
    myISA.setAtmosphere(0., ISA_TEMPERATURE + this->delta_ISA, ISA_PRESSURE);
    for (int i = 0; i < numberDepartureSteps; i++) {
        std::string xmlPath = "mission/departure/departure_step@" + num2Str(i);
        Steps.push_back(missionStep());
        Steps.back().missionPhase = "Departure";
        getMissionSegment(missionXML_tmp, xmlPath);
        getMissionPoint();
    }
    for (int i = 0; i < numberCruiseSteps; i++) {
        std::string xmlPath = "mission/cruise/cruise_step@" + num2Str(i);
        Steps.push_back(missionStep());
        Steps.back().missionPhase = "Cruise";
        getMissionSegment(missionXML_tmp, xmlPath);
        getMissionPoint();
    }
    for (int i = 0; i < numberApproachSteps; i++) {
        std::string xmlPath = "mission/approach/approach_step@" + num2Str(i);
        Steps.push_back(missionStep());
        Steps.back().missionPhase = "Approach";
        getMissionSegment(missionXML_tmp, xmlPath);
        getMissionPoint();
    }
    aixml::closeDocument(missionXML_tmp);
}

missionData::~missionData() {
    //dtor
}

void missionData::getMissionSegment(const node& missionXML, std::string XML) {
    /* General parameters that are always available **/
    Steps.back().modeStep = EndnodeReadOnly<std::string>(XML + "/mode").read(missionXML).value();
    Steps.back().configuration = EndnodeReadOnly<std::string>(XML + "/configuration").read(missionXML).value();
    Steps.back().rating = EndnodeReadOnly<std::string>(XML + "/rating").read(missionXML).value();
    Steps.back().derating = EndnodeReadOnly<double>(XML + "/derate").read(missionXML).value();
    //altitude has to be read depending on mission phase, since first departure step does not have defined altitude
    /* Mission phase dependent parameters **/
    if (Steps.back().missionPhase == "Departure") {
        /* Filling the unoccupied entries depending on the step **/
        if (Steps.back().modeStep == "takeoff") {
            Steps.back().endAltitude = 0.0;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endCAS = convertUnit(KNOTS, METERPERSECOND, 160.0);
        } else if (Steps.back().modeStep == "climb") {
            Steps.back().endAltitude = EndnodeReadOnly<double>(XML + "/altitude").read(missionXML).value();
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            if (accuracyCheck(Steps.back().endAltitude, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "The climb phase is not a final flight altitude! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
        } else if (Steps.back().modeStep == "accelerate") {
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endCAS = EndnodeReadOnly<double>(XML + "/calibrated_airspeed").read(missionXML).value();
            if (accuracyCheck(Steps.back().endCAS, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Acceleration flight phase is no end velocity given! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
        } else {
            myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
            myRuntimeInfo->err << "Step in this flight phase is unknown! Abort!" << std::endl;
            exit(1);
        }
        Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
        Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endTAS);
    } else if (Steps.back().missionPhase == "Cruise") {
        /* Filling the unoccupied entries depending on the step **/
        if (Steps.back().modeStep == "change_speed_to_CAS") {
            Steps.back().endCAS = EndnodeReadOnly<double>(XML + "/calibrated_airspeed").read(missionXML).value();
            if (accuracyCheck(Steps.back().endCAS, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Acceleration flight phase is no end velocity given! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
            Steps.back().endMach = convertUnit(CALIBRATEDAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endCAS);
        } else if (Steps.back().modeStep == "climb_to_cruise" || Steps.back().modeStep == "climb_to_ceiling") {
            Steps.back().endAltitude = EndnodeReadOnly<double>(XML + "/altitude").read(missionXML).value();
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endMach = EndnodeReadOnly<double>(XML + "/mach_number").read(missionXML).value();
            Steps.back().endTAS = convertUnit(MACH, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endMach);
            Steps.back().endCAS = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endTAS);
            if (accuracyCheck(Steps.back().endAltitude, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "No end altitude defined for climb phase! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
        } else if (Steps.back().modeStep == "change_speed_to_Mach") {
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endMach = EndnodeReadOnly<double>(XML + "/mach_number").read(missionXML).value();
            Steps.back().endTAS = convertUnit(MACH, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endMach);
            Steps.back().endCAS = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endTAS);
            if (accuracyCheck(Steps.back().endMach, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Acceleration phase is not a final mach number defined! Abort!" << std::endl;
                exit(1);
            }
        } else if (Steps.back().modeStep == "cruise") {
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
            Steps.back().endMach = Steps.at(Steps.size() - 2).endMach;
        } else if (Steps.back().modeStep == "change_flight_level") {
            if (accuracyCheck(Steps.back().endAltitude, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "No end altitude defined for climbing phase! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
            Steps.back().endMach = Steps.at(Steps.size() - 2).endMach;
        } else if (Steps.back().modeStep == "change_flight_level_constant_ROC") {
            Steps.back().endAltitude = EndnodeReadOnly<double>(XML + "/altitude").read(missionXML).value();
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            if (accuracyCheck(Steps.back().endAltitude, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "No end altitude defined for climbing phase! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
            Steps.back().endMach = Steps.at(Steps.size() - 2).endMach;
        } else if (Steps.back().modeStep == "descend_to_approach") {
            Steps.back().endCAS = EndnodeReadOnly<double>(XML + "/calibrated_airspeed").read(missionXML).value();
            Steps.back().endAltitude = EndnodeReadOnly<double>(XML + "/altitude").read(missionXML).value();
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            if (accuracyCheck(Steps.back().endCAS, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Descend is not a final speed defined! Abort!" << std::endl;
                exit(1);
            }
            if (accuracyCheck(Steps.back().endAltitude, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Descend is not to specify a final flight altitude! Abort!" << std::endl;
                exit(1);
            }
            double TAS_ms(convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS));
            Steps.back().endTAS = TAS_ms;
            Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, TAS_ms);
        } else {
            myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
            myRuntimeInfo->err << "Step in this flight phase is unknown! Abort!" << std::endl;
            exit(1);
        }
    } else if (Steps.back().missionPhase == "Approach") {
        /* Filling the unoccupied entries depending on the step **/
        if (Steps.back().modeStep == "change_speed") {
            Steps.back().endCAS = EndnodeReadOnly<double>(XML + "/calibrated_airspeed").read(missionXML).value();
            if (accuracyCheck(Steps.back().endCAS, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Delay flight phase is no final velocity defined! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
        } else if (Steps.back().modeStep == "descend") {
            Steps.back().endAltitude = EndnodeReadOnly<double>(XML + "/altitude").read(missionXML).value();
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            if (accuracyCheck(Steps.back().endAltitude, 0.0, ACCURACY_HIGH)) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
                myRuntimeInfo->err << "Descend is not an end altitude defined! Abort!" << std::endl;
                exit(1);
            }
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
        } else if (Steps.back().modeStep == "landing") {
            Steps.back().endAltitude = 0.0;
            Steps.back().endTemperature = myISA.getTemperature(Steps.back().endAltitude);
            Steps.back().endCAS = 0.0;
        } else if (Steps.back().modeStep == "level_glide_slope_interception") {
            Steps.back() = Steps.at(static_cast<int>(Steps.size() - 2));
        } else {
            myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << std::endl;
            myRuntimeInfo->err << "Step in this flight phase is unknown! Abort!" << std::endl;
            exit(1);
        }
        Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
        Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endTAS);
    } else {
        myRuntimeInfo->err << "The mission segment cannot be determined!" << std::endl;
        myRuntimeInfo->err << "The mission phase is unknown: " << Steps.back().missionPhase << std::endl;
        exit(1);
    }
}

void missionData::getMissionPoint() {
    if (Steps.size() == 1) {
        Steps.back().Point.Altitude = Steps.back().endAltitude;
        Steps.back().Point.Temperature = myISA.getTemperature(Steps.back().Point.Altitude);
        Steps.back().Point.CAS = Steps.back().endCAS;
        Steps.back().Point.TAS = Steps.back().endTAS;
        Steps.back().Point.Mach = Steps.back().endMach;
    } else {
        Steps.back().Point.Altitude = Steps.at(Steps.size() - 2).endAltitude
                                      + 0.5 * (Steps.back().endAltitude - Steps.at(Steps.size() - 2).endAltitude);
        Steps.back().Point.Temperature = myISA.getTemperature(Steps.back().Point.Altitude);
        Steps.back().Point.CAS = Steps.at(Steps.size() - 2).endCAS
                                 + 0.5 * (Steps.back().endCAS - Steps.at(Steps.size() - 2).endCAS);
        Steps.back().Point.TAS = Steps.at(Steps.size() - 2).endTAS
                                 + 0.5 * (Steps.back().endTAS - Steps.at(Steps.size() - 2).endTAS);
        Steps.back().Point.Mach = Steps.at(Steps.size() - 2).endMach
                                  + 0.5 * (Steps.back().endMach - Steps.at(Steps.size() - 2).endMach);
    }
    Steps.back().Point.density = myISA.getDensity(Steps.back().Point.Altitude);
    Steps.back().Point.dynamicPressure = 0.5 * myISA.getDensity(Steps.back().Point.Altitude) *
                                         pow(Steps.back().Point.TAS, 2.);
}
