/**
 * \file missionData.h
 * \author 
 * \brief Class for data from mission xml
 * \version 0.1
 * \date 2024-06-27
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#ifndef SYSTEMSDESIGN_SRC_MISSIONDATA_H_
#define SYSTEMSDESIGN_SRC_MISSIONDATA_H_

#include <string>
#include <vector>

#include <atmosphere/atmosphere.h>

class node;

/** \brief contains all the properties and functions of the flight mission
 */
class missionData {
 public:
    std::string missionFile;/**< name and directory of the mission file*/

    int missionUtilization;/**< number of passengers on the specific mission */
    double delta_ISA;/**< delta T to ISA conditions [K] */

    std::string missionPhase; /**<mission phase ("Departure", "Cruise", "Approach")  */
    atmosphere myISA;/**ISA atmosphere functions*/

    /** \brief nested class that contains the properties of the average mission point in a mission segment
     */
    class averageMissionPoint {
     public:
        double Altitude;/**<[m] altitude at the mission point*/
        double Temperature;/**<[K]  */
        double CAS;/**<[m/s] calibrated air speed*/
        double TAS;/**<[m/s] true air speed*/
        double Mach;/**<[-]  */
        double density;/**<[kg/m^3] air density  */
        double dynamicPressure;/**<[N/m^2] dynamic air pressure  */

        averageMissionPoint() {
            Altitude = 0.0;
            Temperature = 0.0;
            CAS = 0.0;
            TAS = 0.0;
            Mach = 0.0;
            density = 0.0;
            dynamicPressure = 0.0;
        }
    };


    /** \brief nested class that contains all properties of one mission segment
     */
    class missionStep {
     public:
        averageMissionPoint Point;/**< average mission point in this segment */

        std::string missionPhase;/**<mission phase ("Departure", "Cruise", "Approach")  */
        std::string modeStep; /**< mode of the step (e.g. "takeoff", "climb", ...)*/

        std::string rating;/**<current engine rating (e.g. "TO")*/
        double derating;/**<artificial engine derate factor*/
        std::string configuration;/**<current aerodynamic configuration of the aircraft */

        double endAltitude;/**<[m] altitude at the end of the mission segment*/
        double endCAS;/**<[m/s] calibrated air speed at the end of the mission segment*/
        double endTAS;/**<[m/s] true air speed at the end of the mission segment*/
        double endMach;/**<[-] Mach number at the end of the mission segment */
        double endTemperature;/**<[-] Temperature at the end of the mission segment */

        missionStep()
        :
        missionPhase(""),
        modeStep(""),
        rating(""),
        derating(0.0),
        configuration(""),
        endAltitude(0.0),
        endCAS(0.0),
        endTAS(0.0),
        endMach(0.0),
        endTemperature(0.0) {}
    };
    std::vector<missionStep> Steps;//vector containing the mission steps

    explicit missionData(const std::string& fileNameAndDir);
    virtual ~missionData();

 private:
    node& missionXML_tmp;/*mission xml node*/

    int numberDepartureSteps;/*number of steps in the mission segment "departure"*/
    int numberCruiseSteps;/*number of steps in the mission segment "cruise"*/
    int numberApproachSteps;/*number of steps in the mission segment "approach"*/

    /** \brief reads mission segments from the mission xml
    */
    void getMissionSegment(const node& missionXML, std::string XML);

    /** \brief gets altitude, temperature, CAS, TAS, Mach, density, and dyn. pressure for mission point*/
    void getMissionPoint();
};

#endif // SYSTEMSDESIGN_SRC_MISSIONDATA_H_
