#include "tawWingDesignData.h"
/**
 * \file tawWingDesignData.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Tube and wing wing design data
 * \version 0.1
 * \date 2023-12-20
 *
 * @copyright Copyright (c) 2023
 *
 */
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/io/convert.h>

taw::cantilever::WingDesignData::WingDesignData()
    : specification_icao_aerodrome_reference_code(
          EndnodeReadOnly<std::string>("requirements_and_specifications/requirements/top_level_aircraft_requirements/"
                                       "icao_aerodrome_reference_code")),
      specification_wing_mounting(
          EndnodeReadOnly<std::string>("requirements_and_specifications/design_specification/configuration/wing_definition/mounting")),
      specification_landing_gear_mounting(EndnodeReadOnly<std::string>(
          "requirements_and_specifications/design_specification/configuration/undercarriage_definition/main_gear_mounting")),
      specification_design_mach(
          EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/"
                                  "design_mission/initial_cruise_mach_number")),
      specification_design_altitude(
          EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/"
                                  "design_mission/initial_cruise_altitude")),
      specification_maximum_operating_mach(EndnodeReadOnly<double>(
          "requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_mach_number")),
      specification_maximum_load_factor(
          EndnodeReadOnly<double>("requirements_and_specifications/design_specification/design_load_factor")),
      specification_number_of_wing_mounted_engines(0),
      sizing_point_wing_loading(EndnodeReadOnly<double>("sizing_point/wing_loading")),
      mtom_mass_properties(MassPropertiesIO("analysis/masses_cg_inertia/maximum_takeoff_mass","maximum takeoff mass")),
      wing_mass_properties(MassPropertiesIO("component_design/wing", "wing")),
      wing_position(PositionIO("component_design/wing/position", "wing")) {}

void taw::cantilever::WingDesignData::read(const node& xml) {
  specification_icao_aerodrome_reference_code.read(xml);
  specification_wing_mounting.read(xml);
  specification_landing_gear_mounting.read(xml);
  specification_design_mach.read(xml);
  specification_design_altitude.read(xml);
  specification_maximum_operating_mach.read(xml);
  specification_maximum_load_factor.read(xml);
  sizing_point_wing_loading.read(xml);
  mtom_mass_properties.read(xml);
  read_wing_mounted_engine_information(xml);
}

void taw::cantilever::WingDesignData::read_wing_mounted_engine_information(const node& xml) {
  auto propulsors_reqs = xml.getVector("aircraft_exchange_file/requirements_and_specifications/design_specification/propulsion/propulsor",1);

  std::vector<int> wing_mounted_engine_ids;
  for (size_t i = 0; i < propulsors_reqs.size(); ++i) {
    if (EndnodeReadOnly<std::string>("position/parent_component").read(*propulsors_reqs.at(i)).value() == "wing") {
      specification_number_of_wing_mounted_engines++;
      wing_mounted_engine_ids.push_back(i);
    }
  }
  /* floor uneven number of engines -> chiozzotto only uses one side of the wing */
  size_t number_of_wing_mounted_engine_half = static_cast<size_t>(std::floor(specification_number_of_wing_mounted_engines / 2.));
  if (xml.find("aircraft_exchange_file/component_design/propulsion") == 0) {
    myRuntimeInfo->out << "Using default values for engines if chiozzotto mass method is selected" << std::endl;
    myRuntimeInfo->out << "Assume engine + pylon + nacelle ~4000kg and postion between 0.35 and 0.8 eta, starting at 0.35 for first engine" << std::endl;
    engines_exist = false;

    double engine_mass = 4000.0; /*<-- 4000kg per engine assumed each */
    double eta_min = 0.35;
    double eta_max = 0.8;
    for(size_t i = 0; i < number_of_wing_mounted_engine_half; ++i) {
      double eta_engine = std::lerp(eta_min,eta_max,i/number_of_wing_mounted_engine_half);
      wing_mounted_engine_data.push_back({engine_mass, eta_engine});
    }
  }
  else {
    auto propulsors_comp = xml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion",1);
    for (size_t i = 0; i < number_of_wing_mounted_engine_half; ++i) {
      MassPropertiesIO pylon = MassPropertiesIO("nacelle@0","nacelle");
      MassPropertiesIO nacelle = MassPropertiesIO("pylon@0","pylon");
      MassPropertiesIO engine = MassPropertiesIO("engine","engine");
      pylon.read(*propulsors_comp.at(wing_mounted_engine_ids.at(i)));
      nacelle.read(*propulsors_comp.at(wing_mounted_engine_ids.at(i)));
      engine.read(*propulsors_comp.at(wing_mounted_engine_ids.at(i)));
      double propulsion_mass = pylon.data["mass"].value() + nacelle.data["mass"].value() + engine.data["mass"].value();
      double propulsion_position = engine.data["y"].value();
      wing_mounted_engine_data.push_back({propulsion_mass,propulsion_position});

    }
  }

}

void taw::cantilever::WingDesignData::read_fuselage(std::filesystem::path path,
                                                    std::filesystem::path geometry_data_dir) {
  std::shared_ptr<node> acxml = aixml::openDocument(path);
  geom2::FuselageFactory fuselage_factory{acxml, geometry_data_dir};

  fuselage = fuselage_factory.create("fuselage/specific/geometry/fuselage@0");
}

void taw::cantilever::WingDesignData::update_wing(node& xml) {
  geom2::io::SurfaceType io = geom2::io::Wing(wing);
  const std::string path_to_geometry = "component_design/wing/specific/geometry";
  std::visit(geom2::io::AixmlConverter(xml[path_to_geometry],
                                       {"aerodynamic_surface", "0", wing.name}),
             io);
}

void taw::cantilever::WingDesignData::update_spars(node& xml) {
  for (size_t i = 0; i < spars.size(); ++i) {
    geom2::io::SurfaceType io_spar = geom2::io::Spar(spars[i]);
    const std::string path_to_spars = "component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/spars";
    std::visit(geom2::io::AixmlConverter(
                   xml[path_to_spars],
                   {"spar", std::to_string(i), spars[i].name}),
               io_spar);
  }
}

void taw::cantilever::WingDesignData::update_control_devices(node& xml) {
  for (size_t i = 0; i < control_devices.size(); ++i) {
    geom2::io::SurfaceType io_control_device = geom2::io::ControlDevice(control_devices[i]);
    const std::string path_to_control_devices = "component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices";
    std::visit(geom2::io::AixmlConverter(
                   xml[path_to_control_devices],
                   {"control_device", std::to_string(i), control_devices[i].name}),
               io_control_device);
    Endnode<double> min_deflection(path_to_control_devices + "/control_device@" + std::to_string(i) + "/deflection/full_negative_deflection","full negative deflection");
    Endnode<double> max_deflection(path_to_control_devices + "/control_device@" + std::to_string(i) + "/deflection/full_positive_deflection","full positive deflection");
    const auto [min, max] = control_devices_deflections[i];
    min_deflection.set_unit("rad");
    max_deflection.set_unit("rad");
    min_deflection.set_value(min);
    max_deflection.set_value(max);

    min_deflection.update(xml);
    max_deflection.update(xml);
  }
}

void taw::cantilever::WingDesignData::update(node& xml) {
  auto tmp = xml.find("component_design/wing");
  if (tmp != nullptr) {
    tmp->deleteChildren();
  }
  wing_mass_properties.update(xml);
  wing_position.update(xml);
  update_wing(xml);
  update_spars(xml);
  update_control_devices(xml);
}
